#!/usr/bin/perl -W
# Object to handle clustering of residues in global fits of CPMG data
# J. C. Lansing
# 11-01-04

use lib "$ENV{GLOVEDIR}/bin/perl_lib";

# Statistics package that includes F distribution calculations
use Statistics::Distributions;

# Glove I/O packages
use Glove::Input;
use Glove::Output;
use Glove::Cluster;

package Main;

# Duplicate almost anything
# The targets of references are DUPLICATED -- the new reference will point
# to the copy, instead of producing another pointer to the same data
sub duplicate {
        my $a = shift;
        my $i;
        my $b;
        if (!defined($a)) { return $a; }
        my $type = ref($a);
        if (!ref($a)) {
                # Not a reference
                $b = $a;
        } elsif ($type eq "SCALAR") {
                # It's a scalar reference
                $$b = $$a;
        } elsif ($type eq "ARRAY") {
                # It's an array reference
		for ($i = 0; $i <= $#$a; $i++) {
                        $b->[$i] = duplicate($a->[$i]);
                }
        } elsif ($type eq "CODE") {
                print STDERR "ERROR: copyit() can't handle type \'CODE\'\n";
                $b = $a;
        } elsif ($type eq "GLOB") {
                print STDERR "ERROR: copyit() can't handle type \'GLOB\'\n";
                $b = $a;
        } elsif ($type eq "REF") {
                # It's a reference to a reference
                $$b = duplicate($$a);
        } elsif ($type eq "HASH") {
                # It's a hash
                foreach $i (keys(%$a)) {
                        $b->{$i} = duplicate($a->{$i});
                }
        } else {
                # It's an object reference
		$b = {};
		bless($b,$type);
                foreach $i (keys(%$a)) {
                        $b->{$i} = duplicate($a->{$i});
                }
        }
        return $b;
}

# Constructor for any object
sub new { 
	my $class = shift;	# The class name
	my $self = {};		# An anonymous hash reference
	bless($self,$class);	# Bless the reference to an object
	return $self;		# Return the reference to the new object
}

# Subroutine to organize the output files
sub clean_up {
	if ($#_ < 1) { 
		print "ERROR: subroutine clean_up called with insufficient parameters\n";
		return; 
	}
	my $it = $_[0];
	my $infile = $_[1];
	my $outfile = $_[2];
	if (!-e $it) {
		if (system("mkdir $it")) {
			print "# WARNING: Unable to create directory ./$it\n";
			print "# Output files cannot be sorted.\n";
			return;
		}
	} elsif (!-d $it) {
		print "# WARNING: Unable to move output files to ./$it/\n";
		return;
	}
	if (system("mv *.xmgr $infile $outfile $it")) {
		print "# WARNING: Error when moving output files to ./$it/\n";
	}
	return;
}

# Help text
sub help {
	my $file = (caller(0))[1];
	print "Clustering script for GLOVE\n";
	print "Usage:\t$file parfile\n";
	exit(0);
}

# Body of MAIN code
if ($#ARGV < 0) { &help() }
my $file = $ARGV[0];		# Parameter filename
my $gin = new(Glove::Input);
my $gout = new(Glove::Output);
my $clust = new(Glove::Cluster);
my $in;
my $sfcheck = 1.0;		# Scaling factor for measuring flatness
my $filebase;			# Base filename
my $exec;			# Executable filename
my $cleanup;			# Cleanup flag
my $itmax;			# Maximum number of iterations
my $infile;			# Current input filename
my $outfile;			# Current output filename
my $parstring;			# Commandline parameters for GLOVE
my @switch;			#
my $crit;			# Cutoff value for individual X2
my @e;				# Temporary variable
my $val;			# Temporart variable

#Defaults
$clust->set_conf_expand(0.4);
$clust->set_conf_contract(0.1);

# Process parameter file
open(INPUT,$file) || die "ERROR: Unable to open file \'$file\'\n";
while($line = <INPUT>) {
	chomp($line);
	$line =~ /^([^#]*)/;	# Strip comments
	@e = split(/\s+/,$1);	# Split on white space
	if ($#e < 0) {
		# Blank or comment line -- ignore it
	} elsif ($e[0] =~ /^input_file$/i) {
		# Input/Output file base name
		if ($#e < 1) {
			die "ERROR: INPUT_FILE requires a filename\n";
		}
		$filebase = $e[1];
	} elsif ($e[0] =~ /^executable$/i) {
		# Executable file name
		if ($#e < 1) {
			die "ERROR: EXECUTABLE requires a filename\n";
		}
		$exec = $e[1];
	} elsif ($e[0] =~ /^cleanup$/i) {
		# Cleanup flag
		$cleanup = 1;
	} elsif ($e[0] =~ /^tol$/i) {
		# Tolerance used to determine clustering behavior
		if ($#e < 2) {
			die "ERROR: TOL requires a parameter and a tolerance\n";
		}
		$clust->set_tol($e[1],$e[2]);
		if ($#e < 3) { $clust->set_gbl($e[1]); }
		elsif ($#e < 4) { $clust->set_gbl($e[1],$e[3],2.0); }
		else { $clust->set_gbl($e[1],$e[3],$e[4]); }
	} elsif ($e[0] =~ /^maxit$/i) {
		# Maximum number of iterations
		if ($#e < 1) {
			die "ERROR: ITMAX requires a value\n";
		}
		$itmax = $e[1];
	} elsif ($e[0] =~ /^gpar$/i) {
		# Number of steps in grid search for this global parameter
		if ($#e < 1) {
			die "ERROR: GPAR requires a parameter name\n";
		}
		if ($#e < 2) { $clust->set_gbl($e[1]); }
		elsif ($#e < 3)  { $clust->set_gbl($e[1],$e[2],2.0); }
		else  { $clust->set_gbl($e[1],$e[2],$e[3]); }
	} elsif ($e[0] =~ /^conf1$/i) {
		# Confidence limit for contraction tests
		if ($#e < 1) {
			die "ERROR: CONF1 requires a value\n";
		}
		$clust->set_conf_contract($e[1]);
	} elsif ($e[0] =~ /^conf2$/i) {
		# Confidence limit for expansion tests
		if ($#e < 1) {
			die "ERROR: CONF2 requires a value\n";
		}
		$clust->set_conf_expand($e[1]);
	} elsif ($e[0] =~ /^sfcheck$/i) {
		# Scaling factor for linearity check
		if ($#e < 1) {
			die "ERROR: SFCHECK requires a value\n";
		}
		$sfcheck = $e[1];
	} elsif ($e[0] =~ /^do_sims$/i) {
		# Run simulations at the end
		$dosims = 1;
	} elsif ($e[0] =~ /^x2_cutoff$/i) {
		# Don't cluster sets with X2 greater than $crit
		if ($#e < 1) {
			die "ERROR: X2_CUTOFF requires a value\n";
		}
		$crit = $e[1];
	} else { die "ERROR: Instruction \"$line\" not understood.\n"; }
}
close(INPUT);

if ($clust->get_ntol() < 0) { 
	die "ERROR: Keyword TOL missing from parameter file\n";
}
if (!defined($filebase)) {
	die "ERROR: Keyword INPUT_FILE missing from parameter file\n";
}

# Process GLOVE input file template
$gin->read($filebase);
my @setlist = $gin->list_sets();

$clust->set_names(@setlist);
if ($dosims && !($gin->sims_requested())) {
	print "WARNING: DO_SIMS flag used without any simulation instructions.\n";
	print "Simulations will be skipped.\n";
	undef($dosims);
}

# Change all global parameters to local
foreach $k ($gin->list_global_par()) { 
	if ($clust->gbl_exists($k)) {
		$gin->make_local($k);
	} else {
		print STDERR "ERROR: \'$k\' is a global parameter in the template file,\n";
		print STDERR "but \'$k\' is not listed as global in the parameter file!\n";
		exit(1);
	}
}

# Eliminate sets with negligible dispersion
foreach $s (@setlist) {
	if ($gin->range_y($s) < $gin->avg_dy($s) * $sfcheck) {
		$clust->set_bad($s);
	}
}

# Storage space for clustering info
my $it;
for ($it = 0; $it < $itmax; $it++) {
	print "**** Iteration $it ****\n";
	$in = duplicate($gin);
	# Skip simulations
	$in->no_sims();
	# Restrict the fit to the relevant sets
	$in->trim_sets($clust->list_current());
	# Set parameters as global (if necessary)
	foreach $k ($clust->list_gbl()) { 
		# Change the parameter to global
		$in->make_global($k);
		if ($j = $clust->fixed($k)) { 
			# Change the parameter to fixed
			$in->make_fixed($k,$j); }
		else {
			# Restrict the search range
			$val = $clust->guess_range($k);
			if (defined($val)) {
				$in->change_range($k,$val)
			}
		}
	}
	# Write the input file
	$infile = $filebase . "." . $it . ".in";
	$outfile = $filebase . ".$it.out";
	$parstring = " -xmgr -f $infile > $outfile";
	$in->write($infile);
	# Run the fit
	if (system("$exec $parstring")) {
		print "\n\nERROR: Executable did not exit normally!  Aborting.\n";
		print "\n### Executable output:\n";
		open(INPUT,$outfile);
		while($line = <INPUT>) { print $line; }
		close(INPUT);
		exit(1);
	}
	# Read the results
	$gout->read($outfile);

	# Run the cleanup script
	if (defined($cleanup)) { &clean_up($it,$infile,$outfile); }

	# Store the results
	if ($it == 0) { 
		$clust->reference($gout);
		if (defined($crit)) { $clust->filter_x2($crit); }
	} 
	else { $clust->read($gout); }
	

	# Call cluster to make a decision about what to do
	if ($clust->refine() eq "DONE") { last; }
}

if ($it >= $itmax) {
	print "# MAXIMUM ITERATION REACHED\n";
	exit(0);
}

# RUN SIMS
if ($dosims) {
	# Loop over all clusters
	for ($i = 0; $i <= $clust->{CURRENT}; $i++) {
		print "\nRunning simulations for cluster $i\n";
		$in = duplicate($gin);
		# Restrict the fit to the relevant sets
		$in->trim_sets($clust->list_cluster($i));
		# Set global parameters as fixed
		foreach $k ($clust->list_gbl()) { 
			$in->make_global($k); 
			$in->make_fixed($k,$clust->get_par($i,$k));
		}
		# Write the input file
		$infile = "$filebase.sim$i.in";
		$outfile = "$filebase.sim$i.out";
		$parstring = " -xmgr -f $infile > $outfile";
		$in->write($infile);
		# Run the fit
		if (system("$exec $parstring")) {
			print "\n\nERROR: Executable did not exit normally!  Aborting.\n";
			print "\n### Executable output:\n";
			open(INPUT,$outfile);
			while($line = <INPUT>) { print $line; }
			close(INPUT);
			exit(1);
		}
		# Run the cleanup script
		if (defined($cleanup)) { &clean_up("sims",$infile,$outfile); }
	}
}


__END__

