// Calc class
//	Object to keep track of calculation instructions
// Kenji Sugase		11/14/2012

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cmath>
#include "Calc.h"
#include "Alloc.h"
#include "Mathlib.h"

OneCalc::OneCalc(int type, const string &label, const Data &data)
	: calc_type(type), calc_label(label), calc_data(data) {}

void OneCalc::prep(size_t ns)
{
	nsim.resize(ns);
	sum.resize(ns);
	sum2.resize(ns);
	orig.resize(ns);
	zero();
}

void OneCalc::zero()
{
	fill(nsim.begin(), nsim.end(), 0  );
	fill(sum.begin() , sum.end() , 0.0);
	fill(sum2.begin(), sum2.end(), 0.0);
	fill(orig.begin(), orig.end(), 0.0);
}
void OneCalc::addVal(size_t ns, double val)
{
	sum[ns]  += val;
	sum2[ns] += sqr(val);
	++nsim[ns];
}

void OneCalc::addJack(size_t ns, double val)
{
	sum[ns]  += val;
	sum2[ns] += sqr(val - orig[ns]);
	++nsim[ns];
}

double OneCalc::mean(size_t ns)
{
	if (nsim[ns] < 1) { return 0.0; }
	return sum[ns] / nsim[ns];
}

double OneCalc::stdev(size_t ns)
{
	if (nsim[ns] < 2) { return 0.0; }
	return sqrt((sum2[ns] - sqr(sum[ns]) / nsim[ns]) / (nsim[ns] - 1));
}

// Standard deviation estimate for jackknife simulations
// Requires that addVal called with the value (jack_val - mean)
double OneCalc::jackStdev(size_t ns)
{
	if (nsim[ns] < 2) { return 0.0; }
	return sqrt(sum2[ns] * (nsim[ns] - 1) / nsim[ns]);
}

void OneCalc::store(size_t ns, double val)
{
	orig[ns] = val;
}

int OneCalc::type() const
{
	return calc_type;
}

const string & OneCalc::label() const
{
	return calc_label;
}

Calc::Calc() : type_jack(0), type_Monte(0), nMonte(0) {}

void Calc::prep(int ns)
{
	int sz = calc1.size();
	for (int nc = 0; nc < sz; nc++) {
		calc1[nc].prep(ns);
	}
}

void Calc::zero()
{
	int sz = calc1.size();
	for (int nc = 0; nc < sz; nc++) {
		calc1[nc].zero();
	}
}

void Calc::Add(const string &type, const string &label, const dvect &condition, const ivect &index)
{
	int nType;
	if      (!strcasecmp(type.c_str(), "REX"   )) { nType = CALC_REX;    }
	else if (!strcasecmp(type.c_str(), "PB"    )) { nType = CALC_PB;     }
	else if (!strcasecmp(type.c_str(), "PB_RDC")) { nType = CALC_PB_RDC; }
	else if (!strcasecmp(type.c_str(), "PB_MQ" )) { nType = CALC_PB_MQ;  }
	else {
		cerr << "ERROR: CALC \'" << type << "\' is not recognized.\n";
		exit(1);
	}

	Data data;
	data.x = condition;
	data.idx = index;
	calc1.push_back(OneCalc(nType, label, data));
}

int Calc::size() const
{
	return calc1.size();
}

const string & Calc::Label(size_t nc) const
{
	return calc1[nc].label();
}

Data & Calc::getData(size_t nc)
{
	return calc1[nc].calc_data;
}

double Calc::run(size_t nc, const dvect &p, pFunc fx, const i2vect &index)
{
	dvect dy0;
	double *ppar = NULL;
	double ret;
	double papb;

	switch(calc1[nc].type()) {
	case CALC_REX:		// Calculate R2(0) - R2(inf)
		if (checkIndex(calc1[nc].calc_data.idx, index)) {
			int sz = calc1[nc].calc_data.map.size();
			make_new(ppar, sz, "run");
			for (int i = 0; i < sz; i++) {
				ppar[i] = p[calc1[nc].calc_data.map[i]];
			}
			calc1[nc].calc_data.x[0] = 50.0;
			ret = (*fx)(calc1[nc].calc_data, ppar, dy0, bvect());
			calc1[nc].calc_data.x[0] = 100000.0;
			ret -= (*fx)(calc1[nc].calc_data, ppar, dy0, bvect());
			zap(ppar);
			return ret;
		} else {
			return 0;
		}
		break;

	case CALC_PB:		// Calculate population B
		// PAPB stored as parameter # 2 for Richards
		papb = p[calc1[nc].calc_data.map[2]];
		ret = 0.5 * (1 - sqrt(1 - 4*papb));
		if (std::isfinite(ret)) { return ret;      }
		else               { return HUGE_VAL; }
		break;

	case CALC_PB_RDC:	// Calculate population B
		// PAPB stored as parameter # 3 for RDC
		papb = p[calc1[nc].calc_data.map[3]];
		ret = 0.5 * (1 - sqrt(1 - 4*papb));
		if (std::isfinite(ret)) { return ret;      }
		else               { return HUGE_VAL; }
		break;

	case CALC_PB_MQ:	// Calculate population B
		// PAPB stored as parameter # 3 for Richards_mq
		papb = p[calc1[nc].calc_data.map[3]];
		ret = 0.5 * (1 - sqrt(1 - 4*papb));
		if (std::isfinite(ret)) { return ret;      }
		else               { return HUGE_VAL; }
		break;

	default:
		return 0.0;
		break;
	}
	return 0.0;
}

bool Calc::checkIndex(const ivect &calc_index, const i2vect &parm_index)
{
	int szi = parm_index.size();
	for (int ni = 0; ni < szi; ni++) {
		if (parm_index[ni] == calc_index) {
			return true;
		}
	}
	return false;
}

void Calc::addVal(size_t ns, size_t nc, double val)
{
	calc1[nc].addVal(ns, val);
}

void Calc::addJack(size_t ns, size_t nc, double val)
{
	calc1[nc].addJack(ns, val);
}

double Calc::calcMean(size_t ns, size_t nc)
{
	return calc1[nc].mean(ns);
}

double Calc::calcStdev(size_t ns, size_t nc)
{
	return calc1[nc].stdev(ns);
}

double Calc::calcJackStdev(size_t ns, size_t nc)
{
	return calc1[nc].jackStdev(ns);
}

void Calc::storeResult(size_t ns, size_t nc, double val)
{
	calc1[nc].store(ns, val);
}

void Calc::SetJack(int num)
{
	type_jack = num;
}

int Calc::getJackType() const
{
	return type_jack;
}

void Calc::SetMonte(int num, int iter)
{
	type_Monte = num;
	nMonte = iter;
}

int Calc::getMonteType() const
{
	return type_Monte;
}

int Calc::getMonteIter() const
{
	return nMonte;
}
