// ParDef class
//	Class to hold temporary info about global variables
// Jonathan C. Lansing	3/6/2005
// Kenji Sugase		7/26/2012

#ifndef __PARMDEF_H__
#define __PARMDEF_H__

#include <iostream>
#include <vector>
#include <string>
#include <cmath>
#include "Data.h"
#include "Rule.h"
#include "Type.h"

struct Global
{
	int num;			// parameter number
	string name;			// parameter name
	bool restraint;			// flag for restraint (GPAR = false, GREST = true)

	Global(int n, const string &GPAR, bool flag) : num(n), name(GPAR), restraint(flag) {}
};

struct Local
{
	int num;			// parameter number
	string name;			// parameter name (base parameter name + index)
	int nBasePar;			// base parameter number
	ivect index;			// index of each paramter defined in the input file
					// -1 if index is not defined

	Local(int n, const string &PAR, int nb, const ivect &idx) :
		num(n), name(PAR), nBasePar(nb), index(idx) {}

	bool isMatch(int nb, const ivect &idx)
	{
		if (nBasePar != nb) { return false; }
		int sz = index.size();
		for (int i = 0; i < sz; i++) {
			if (index[i] >= 0 && index[i] != idx[i]) { return false; }
		}
		return true;
	}
};

namespace {
	const int UNDEF = 0;
	const int GRID  = 1;
	const int FIXED = 2;

	const int GRID_NORMAL = 0;
	const int GRID_EXP    = 1;
	const int GRID_GAUSS  = 2;
	const int GRID_TENTH  = 3;
	const int GRID_iEXP   = 4;
	const int GRID_iGAUSS = 5;
	const int GRID_iTENTH = 6;
}

// Class to hold info about a single parameter
class Elem
{
public:
	int parm_type;			// UNDEF, GRID or FIXED
	int nBasePar;			// base parameter number
	string parm_name;		// PAR + index
	string parm_fullname;		// SET + PAR + index
	double grid_start;
	double grid_end;
	double grid_range;		// end - start
	double grid_step;
	int grid_size;
	int grid_interval;		// flag for the method to calculate a grid interval
	double grid_initial;		// initial value
	double grid_lower;		// lower bound
	double grid_upper;		// upper bound
	ivect nSetByPAR;		// dataset numbers used by each parameter

	Elem();
	void setName(const string &PAR, const string &set);
	void setGrid(double start, double end, int size, double interval, double initial);
	void setFixed(double value);
	void Restraint(double start, double end);
	void Add_nSetByPAR(int num);
	bool IsLocalNonFIXED(int ns) const;
	bool isLocal() const;
	bool isGlobal() const;
};



// Class to hold info about all parameters in all sets
class ParDef
{
private:
	vector<Elem> pv;		// parameter values
	Rules rule;			// pointer to Rules object
	svect title;			// dataset title
	vector<vector<Local> > lPar;	// local parameters
	vector<Global> gPar;		// global parameters
	int nSet;			// number of dataset
	int gdof;			// grobal degrees of freedom
	ivect ldof;			// local degrees of freedom
	void Store_SET();
	void Remove_SET();
	void Add_ParmMap(int num, const string &name);
	int Find_GPAR(const string &name, const string &GPAR);
	bool Calc_lDoF();
	void Calc_gDoF();

public:
	bvect mask_global;		// whether a global parameter or not
	b2vect mask_local;		// whether a local parameter or not for each SET
	vector<Data> data;		// data points
	i2vect nDataBySet;		// data point numbers used in each dataset
	i2vect nParBySet;		// parameter numbers used in each dataset
	s2vect ParNameBySet;		// parameter names used in each dataset
	ivect allPAR;			// list of all parameter numbers
	ivect allGPAR;			// list of all global parameters
	ivect allData;			// list of all data numbers
	bvect fixed;			// flags for FIXED parameters
	i3vect indexBySet;		// index used in each dataset

	void setRule(const Rules &ptr);
	void Add_GPAR(const string &name, double start, double end, int size, double interval, double initial);
	void Add_GREST(const string &name, double start, double end, int size, double interval, double initial);
	void Add_GFIXED(const string &name, double value);
	void Add_PAR(const string &name, double start, double end, int size, double interval, double initial);
	void Add_PAR(const string &name, string &GPAR);
	void Add_REST(const string &name, double start, double end, int size, double interval, double initial);
	void Add_FIXED(const string &name, double min);
	void Add_Set(const string &name);
	void Add_DATA(dvect &condition, ivect &index, double value, double error);
	ivect Get_nPARs(const ivect &index);
	void prep();
	void Get_datamap(Data &data, size_t ns);
	int match_local(int bn, const ivect &idx);
	int match_local(int bn, size_t ns, const ivect &idx);
	void setgDoF(int val);
	void setlDoF(size_t n, int val);
	int SETsize() const;
	int PAR_size() const;
	int gDoF() const;
	int lDoF(size_t n) const;
	const string & TITLE(size_t n) const;
	const string & Name(size_t n) const;
	const string & FullName(size_t n) const;
	double grd_start(size_t n) const;
	double grd_end(size_t n) const;
	double grd_range(size_t n) const;
	double grd_initial(size_t n) const;
	double grd_step(size_t n) const;
	int grd_size(size_t n) const;
	int grd_interval(size_t n) const;
	bool isGRID(size_t n) const;
	bool isFIXED(size_t n) const;
	bool isLegal(size_t n, double val) const;
	bool isGlobal(size_t n) const;
	bool isLocal(size_t n) const;
};

#endif
