// Wrapper functions for linear algebra routines
// Kenji Sugase		12/13/2013

#include <iostream>
#include <cmath>
#include "LinearAlgebra.h"
#include "Alloc.h"

using namespace std;

int idxr(int c, int r, int order) { return 2*(c + order*r);     }
int idxi(int c, int r, int order) { return 2*(c + order*r) + 1; }
int idvr(int c) { return 2*c;     }
int idvi(int c) { return 2*c + 1; }

// copy a square matrix from one place to another
void zmat_cpy( int order, const double *A, double *B) {
	zlacpy_("A", &order, &order, A, &order, B, &order);
}

// C = A x B
void zmat_mlt(int order, const double *A, const double *B, double *C) {
	double alpha[2] = {1.0, 0.0};
	double beta[2]  = {0.0, 0.0};
	zgemm_("N", "N", &order, &order, &order, alpha, A, &order, B, &order, beta, C, &order);
}

// D = A * B * C
void zmat_mlt(int order, const double *A, const double *B, const double *C, double *D) {
	double tmp[2*order*order];
	zmat_mlt(order, A, B, tmp);
	zmat_mlt(order, tmp, C, D);

}

// E = A * B * C * D
void zmat_mlt(int order, const double *A, const double *B, const double *C, const double *D, double *E) {
	double tmp[2*order*order];
	zmat_mlt(order, A, B, C, tmp);
	zmat_mlt(order, tmp, D, E);
}

// Computes Y = A * X for matrix A and vector X
void zmat_vmlt(int order, const double *A, const double *X, double *Y) {
	double alpha[2] = {1.0,0.0};
	double beta[2] = {0.0,0.0};
	int incx = 1;
	int incy = 1;
	zgemv_("N", &order, &order, alpha, A, &order, X, &incx, beta, Y, &incy);
}

// Computes Y = A * B * X for matrices A, B and vector X
void zmat_vmlt(int order, const double *A, const double *B, const double *X, double *Y) {
	double tmp[2*order*order];
	zmat_mlt(order, A, B, tmp);
	zmat_vmlt(order, tmp, X, Y);
}

// C = A x B (A is a diagonal matrix)
void zdmat_mlt(int order, const double *A, const double *B, double *C) {
	for (int i = 0; i < order; i++) {
		for (int j = 0; j < order; j++) {
			C[2*i+2*order*j  ] = A[2*i] * B[2*i+2*order*j  ] - A[2*i+1] * B[2*i+2*order*j+1];
			C[2*i+2*order*j+1] = A[2*i] * B[2*i+2*order*j+1] + A[2*i+1] * B[2*i+2*order*j];
		}
	}
}

// D = A x B x C (B is a diagonal matrix)
void zdmat_mlt(int order, const double *A, const double *B, const double *C, double *D) {
	double tmp[2*order*order];
	zdmat_mlt(order, B, C, tmp);
	zmat_mlt(order, A, tmp, D);
}

// Compute B = exp(A*x)
void zmat_exp(int order, double x, const double *A, double *B) {
	for (int i = 0; i < order ; i++) {
		B[2*i]   = exp(x * A[2*i]) * cos(x * A[2*i+1]);
		B[2*i+1] = exp(x * A[2*i]) * sin(x * A[2*i+1]);
	}
}

// Compute a conjugate complex matrix: A*
void zmat_cnj(int order, const double *A, double *cnjA) {
	for (int i = 0; i < 2*order*order ; i += 2) {
		cnjA[i]   =  A[i];
		cnjA[i+1] = -A[i+1];
	}
}

// Compute an inverse matrix using LU decomposition: A-1
int zmat_inv(int order, const double *mat, double *imat) {
	int lwork = 2*order;
	int ipiv[order];
	double work[4*order];
	// Duplicate the matrix A
	zlacpy_("A", &order, &order, mat, &order, imat, &order);
	// LU decomposition
	int info;
	zgetrf_(&order, &order, imat, &order, ipiv, &info);
	if (info) {
		cerr << "WARNING: Relaxation matrix is singular. Check the input file.\n";
	} else {
		// Compute the inverse of A
		zgetri_(&order, imat, &order, ipiv, work, &lwork, &info);
		if (info) cerr << "WARNING: Relaxation matrix is singular. Check the input file.\n";
	}
	return info;
}

// Compute right eigenvectors and eigenvalues: A = (V-1) x E x V
int zmat_eig(int order, double *mat, double *evec, double *eval) {
	int lwork = 2*order;
	double rwork[2*order];
	double work[4*order];
	double *vl = NULL;
	int ldvl = 1;
	int info;
	zgeev_("N", "V", &order, mat, &order, eval, vl, &ldvl, evec, &order, work, &lwork, rwork, &info);
	if (info) {
		cerr << "\nERROR: Eigenvectors of the relaxation matrix cannot be computed.\n";
		exit(1);
	}
	return info;
}

// Compute right eigenvectors and eigenvalues: A = (V-1) x E x V
int zmat_eig(int order, const double *mat, double *evec, double *ievec, double *eval) {
	// Duplicate the matrix A
	double tmp[2*order*order];
	zlacpy_("A", &order, &order, mat, &order, tmp, &order);
	int lwork = 2*order;
	double rwork[2*order];
	double work[4*order];
	double *vl = NULL;
	int ldvl = 1;
	int info;
	zgeev_("N", "V", &order, tmp, &order, eval, vl, &ldvl, evec, &order, work, &lwork, rwork, &info);
	if (info) {
		cerr << "\nERROR: Eigenvectors of the relaxation matrix cannot be computed.\n";
		exit(1);
	}
	zmat_inv(order, evec, ievec);
	return info;
}

void zmat_pow(int order, double *A, int n, double *B) {
	double alpha[2] = {1.0, 0.0};
	double beta[2] = {0.0, 0.0};
	double *curr;
	double *last;
	double *tmp;
	double work[2*order*order];

	if (n == 0) {
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++) {
				B[idxr(i,j,order)] = B[idxi(i,j,order)] = 0.0;
			}
			B[idxr(i,i,order)] = 1.0;
		}
		return;
	}
	if (n % 2) {
		curr = B;
		last = work;
	} else {
		curr = work;
		last = B;
	}
	// Initialize last
	for (int i = 0; i < 2*order*order; i++) last[i] = 0.0;
	for (int i = 0; i < order; i++) last[2*i*(order+1)] = 1.0;
	for (int i = 0; i < n; i++) {
		// multiply by A
		zgemm_("N", "N", &order, &order, &order, alpha, last, &order, A, &order, beta, curr, &order);
		// Swap workspace pointers
		tmp = last; last = curr; curr = tmp;
	}
	return;
}

// E = (A * B * C * D)^n
void zmat_mltpow(int order, const double *A, const double *B, const double *C, const double *D, int n, double *E) {
	double tmp[2*order*order];
	zmat_mlt(order, A, B, C, D, tmp);
	zmat_pow(order, tmp, n, E);
}
/*
void zmat_powfast(int order, double *A, int n, double *B) {
        int nn = n;
	double alpha[2] = {1.0, 0.0};
	double beta[2] = {0.0, 0.0};
	double work[2*order*order];

	if (n == 0) {
		for (int i = 0; i < order; i++) {
			for (int j = 0; j < order; j++) {
				B[idxr(i,j,order)] = B[idxi(i,j,order)] = 0.0;
			}
			B[idxr(i,i,order)] = 1.0;
		}
		return;
	}
	// Initialize last (identity)
	for (int i = 0; i < 2*order*order; i++) B[i] = 0.0;
	for (int i = 0; i < order; i++) B[2*i*(order+1)] = 1.0;
        // Calculate power
	while (nn > 0) {
            if  (nn % 2 == 1) {
		// curr = last * A
		zgemm_("N", "N", &order, &order, &order, alpha, A, &order, B, &order, beta, work, &order);
		// Swap workspace pointers
		zdmat_cpy(order, work, B);
                nn = nn - 1;
            }
            zgemm_("N", "N", &order, &order, &order, alpha, A, &order, A, &order, beta, work, &order);
	    zdmat_cpy(order, work, A);
            nn = nn / 2;
	}
	return;
}

// E = (A * B * C * D)^n
void zmat_mltpowfast(int order, const double *A, const double *B, const double *C, const double *D, int n, double *E) {
	double tmp[2*order*order];
	zmat_mlt(order, A, B, C, D, tmp);
	zmat_powfast(order, tmp, n, E);
}
*/
/*
// Solve the normal equation using Cholesky decomposition: X = (A-1) x B
int LM_solv(int order, const d2vect &A, const dvect &B, dvect &X, double lambda) {
	int info, nrhs = 1;
	int k = 0;
	static int order0 = 0;
	static double *tmpA = NULL;
	static double *tmpB = NULL;
	// Allocate new matrices
	if (order0 < order) {
		order0 = order;
		make_new(tmpA, order * order, "LM_solv");
		make_new(tmpB, order, "LM_solv");
	}
	// Construct a lower triangular matrix
	for (int i = 0; i < order; i++) {
		for (int j = i; j < order; j++) tmpA[k++] = A[i][j] * (1 + (i == j) * lambda);
		tmpB[i] = B[i];
	}
	// Solve A * X = B using Cholesky decomposition
	dppsv_("L", &order, &nrhs, tmpA, tmpB, &order, &info);
	// Store the result
	if (!info) for (int i = 0; i < order; i++) X[i] = tmpB[i];
	else cerr << "WARNING (Levenberg-Marquardt Minimization): Design matrix is singular.\n";
	return info;
}
*/

// Solve the normal equation using LU decomposition: X = (A-1) x B
int LM_solv(int order, const d2vect &A, const dvect &B, dvect &X, double lambda) {
//	static int order0 = 0;
//	static double *tmpA = NULL;
//	static double *tmpB = NULL;
//	static int    *ipiv = NULL;
	double *tmpA = NULL;
	double *tmpB = NULL;
	int    *ipiv = NULL;
	// Allocate new matrices
//	if (order0 != order) {
//		order0 = order;
//		zap(tmpA);
//		zap(tmpB);
//		zap(ipiv);
		make_new(tmpA, order*order, "LM_solv");
		make_new(tmpB, order, "LM_solv");
		make_new(ipiv, order, "LM_solv");
//	}

	int k = 0;
	for(int i = 0; i < order; i++) {
		for (int j = 0; j < order; j++) {
			tmpA[k++] = A[i][j] * (1 + (i == j) * lambda);
		}
		tmpB[i] = B[i];
	}
	// Solve A * X = B using LU decomposition
	int info, nrhs = 1;
	dgesv_(&order, &nrhs, tmpA, &order, ipiv, tmpB, &order, &info);
	if (!info) {
		// Store the results
		for (int i = 0; i < order; i++) {
			X[i] = tmpB[i];
		}
	} else {
//		cerr << "WARNING (Levenberg-Marquardt Minimization): singular matrix.\n";
	}
	zap(tmpA);
	zap(tmpB);
	zap(ipiv);
	return info;
}

// Compute an inverse matrix using LU decomposition: A-1
int LM_inv(int order, d2vect &mat){
	double *tmpM = NULL;
	double *work = NULL;
	int *ipiv    = NULL;
	make_new(tmpM, order*order, "LM_inv");
	make_new(work, order, "LM_inv");
	make_new(ipiv, order, "LM_inv");
	int k = 0;
	for (int i = 0; i < order; i++) {
		for (int j = 0; j < order; j++) { tmpM[k++] = mat[i][j]; }
	}
	// LU decomposition
	int info;
	dgetrf_(&order, &order, tmpM, &order, ipiv, &info);
	if (!info) {
		// Compute the inverse matrix
		dgetri_(&order, tmpM, &order, ipiv, work, &order, &info);
		if (info) {

		} else {
			// Store the result
			k = 0;
			for (int i = 0; i < order; i++) {
				for (int j = 0; j < order; j++) {
					mat[i][j] = tmpM[k++];
				}
			}
		}
	}
	zap(tmpM);
	zap(work);
	zap(ipiv);
	return info;
}
