// func_cpmg2.cpp
// R2 dispersion experiment, 2-state exchange
// Kenji Sugase		12/20/2013

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Data.h"
#include "FuncLib.h"
#include "LinearAlgebra.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;

namespace {

void check_itcp(double itcp)
{
	if (itcp < LOW_ITCP) {
		cerr << "1/tcp is too short. Check the input file\n";
		exit(1);
	}
}

// Constant value function for CPMG data
// par[0]	R20			(/s)
double fcpmg_const(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	if (dy.size()) {
		// calculate partial derivatives
		dy[data.map[0]] = 1;
	}
	return par[0];
}

// Fast limit CPMG function, Luz & Meiboom approximation
// par[0]	kex			(/s)
// par[1] 	pdw = papb*dw^2		(rad2/s2)
// par[2]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	Temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_meiboom(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_kex = data.map[0];
	const int i_pdw = data.map[1];
	const int i_r20 = data.map[2];
	// copy parameters
	double itcp   = data.x[0];			// 1/tcp
	double field  = data.x[2];			// B0 Field / Base B0 Field
	double kex    = par[0];				// kex
	double pdw    = par[1] * sqr(field);		// papb * (dw * field)^2
	double r20    = par[2];				// R20
	// calculate parameters
	check_itcp(itcp);
	double ktcp   = 0.5 * kex / itcp;		// kex * tcp/2
	double tanhkt = 1 / ktcp * tanh(ktcp);		// 2 / (kex * tcp) * tanh(kex * tcp/2)
	double effr2  = r20 + pdw / kex * (1 - tanhkt);	// effective R2
	if (!dy.size()) {
		// no partial derivatives
		return effr2;
	}
	// calculate partial derivatives
	if (!fixed[i_kex]) {
		if (ktcp < COSH_MAX) {
			dy[i_kex] = pdw / sqr(kex) * (-1 + 2 * tanhkt - 1 / sqr(cosh(ktcp)));
		} else {
			dy[i_kex] = pdw / sqr(kex) * (-1 + 2 * tanhkt);
		}
	}
	if (!fixed[i_pdw]) { dy[i_pdw] = sqr(field) / kex * (1 - tanhkt); }
	if (!fixed[i_r20]) { dy[i_r20] = 1; }
	return effr2;
}

// Ishima & Torchia approximation
// par[0]	dw			(rad/s)
// par[1]	kex			(/s)
// par[2]	papb			(unitless)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	Temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_torchia(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kex  = data.map[1];
	const int i_papb = data.map[2];
	const int i_r20  = data.map[3];
	// copy parameters
	double itcp  = data.x[0];			// 1/tcp
	double field = data.x[2];			// B0 Field / Base B0 Field
	double dw    = par[0] * field;			// dw * field
	double kex   = par[1];				// kex
	double papb  = par[2];				// papb
	double r20   = par[3];				// R20
	// calculate parameters
	check_itcp(itcp);
	double dw2   = sqr(dw);				// dw^2
	double kex2  = sqr(kex);			// kex^2
	double pop   = sqrt(1 - 4 * papb);		// pa - pb
	double pa    = 0.5 * (1 + pop);			// pa
	double norm  = sqrt(sqr(pa * dw2) + 144 * biquad(itcp));
	double denom = 1 / (kex2 + norm);
	double Rex   = papb * dw2 * kex * denom;	// Rex
	double effr2 = r20 + Rex;			// effective R2
	if (!dy.size()) {	// no partial derivatives
		return effr2;
	}
	// calculate partial derivatives
	double dw4  = sqr(dw2);				// dw^4
	double dnm_inrm = denom / norm;
	if (!fixed[i_dw])   { dy[i_dw]   = 2 * Rex / dw * field * (1 - dw4 * sqr(pa) * dnm_inrm); }
	if (!fixed[i_kex])  { dy[i_kex]  = Rex / kex *(1 - 2 * kex2 * denom); }
	if (!fixed[i_papb]) { dy[i_papb] = Rex / papb * (1 + papb * dw4 * pa * dnm_inrm / pop); }
	if (!fixed[i_r20])  { dy[i_r20]  = 1; }
	return effr2;
}

// Partial derivatives. dX/dR20 is not necessary
// pdrv[X][0]: d(kex)/d(X)
// pdrv[X][1]: d(zeta)/d(X) * zeta
// pdrv[X][2]: d(psi)/d(X)
// pdrv[X][3]: inorm * {d(psi)/d(X) * psi + d(zeta)/d(X) * zeta}
// pdrv[X][4]: d(D)/d(X)
void dtmp_D(double *pdrv, double psi, double inorm, double pdw2, double sign = 1.0)
{
	pdrv[3] = inorm * (pdrv[1] + psi * pdrv[2]);
	pdrv[4] = 0.5 * inorm * (sign * pdrv[2] - pdw2 * pdrv[3]);
}

double calc_R2(const Data &data, dvect &dy, const bvect &fixed, double (*pdrv)[5], int i_r20
	, double kex, double r20, double psi, double D_p, double D_m, double norm)
{
	double itcp = data.x[0];
	check_itcp(itcp);
	// calculate effective R2
	double r2eff;
	double tcp = 1 / itcp;
	double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
	double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
	int sz = data.map.size() - 1;
	if (eta_p < COSH_MAX) {
		double coshp = cosh(eta_p);			// cosh(eta+)
		double cosm  = cos(eta_m);			// cos(eta-)
		r2eff = r20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
		// calculate partial derivatives of effective R2
		if (dy.size()) {
			double tmp_p = tcp / (sqrt(8 * ( psi + norm)) + TINY);
			double tmp_m = tcp / (sqrt(8 * (-psi + norm)) + TINY);
			double dnm   = 0.5 * itcp / (sqrt(sqr(D_p * coshp - D_m * cosm) - 1) + TINY);
			double Dps   = D_p * sinh(eta_p);
			double Dms   = D_m * sin(eta_m);
			double cshcs = coshp - cosm;
			for (int i = 0; i < sz; i++) {
				if (fixed[data.map[i]]) { continue; }
				double deta_p   = tmp_p*( pdrv[i][2]+pdrv[i][3]);
				double deta_m   = tmp_m*(-pdrv[i][2]+pdrv[i][3]);
				dy[data.map[i]] = 0.5*pdrv[i][0]-dnm*(pdrv[i][4]*cshcs+Dps*deta_p+Dms*deta_m);
			}
			dy[i_r20] = 1;			// d(effR2)/d(R20)
		}
	// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
	} else {
		r2eff = r20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
		// Calculate the partial derivatives of effective R2
		if (dy.size()) {
			double tmp_p = tcp / sqrt(8 * (psi + norm));
			for (int i = 0; i < sz; i++) {
				if (fixed[data.map[i]]) { continue; }
				dy[data.map[i]] = 0.5*(pdrv[i][0]-itcp*(pdrv[i][4]/(D_p+TINY)+tmp_p*(pdrv[i][2]+pdrv[i][3])));
			}
			dy[i_r20] = 1;			// d(effR2)/d(R20)
		}
	}
	return r2eff;
}

// Carver and Richards approximation
// par[0]	dw			(rad/s)
// par[1]	kex			(/s)
// par[2]	papb			(unitless)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K)
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_richards(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kex  = data.map[1];
	const int i_papb = data.map[2];
	const int i_r20  = data.map[3];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double temp, field, dw, papb, kex, psi, D_p, D_m, norm;
	static double pdrv[3][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (temp != data.x[1] || field != data.x[2]
	|| dw != par[0] || kex != par[1] || papb != par[2]){
		temp  = data.x[1];
		field = data.x[2];			// B0 / Base B0
		dw    = par[0];
		kex   = par[1];
		papb  = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = sqrt(1 - 4 * papb);	// pa - pb
		double zeta2 = sqr(2 * dwfld * kex * pop);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_kex]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta2 / kex;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_papb]) {
				//pdrv[2][0] = 0;
				pdrv[2][1] = -2 * zeta2 / sqr(pop);
				//pdrv[2][2] = 0;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// par[0]	dw			(rad/s)
// par[1]	kex			(/s)
// par[2]	pb			(unitless)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K)
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_richards_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw  = data.map[0];
	const int i_kex = data.map[1];
	const int i_pb  = data.map[2];
	const int i_r20 = data.map[3];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double temp, field, dw, pb, kex, psi, D_p, D_m, norm;
	static double pdrv[3][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (temp != data.x[1] || field != data.x[2]
	|| dw != par[0] || kex != par[1] || pb != par[2]){
		temp  = data.x[1];
		field = data.x[2];			// B0 / Base B0
		dw    = par[0];
		kex   = par[1];
		pb    = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = 1 - 2 * pb;	// pa - pb
		double zeta  = 2 * dwfld * kex * pop;
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_kex]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta2 / kex;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_pb]) {
				//pdrv[2][0] = 0;
				pdrv[2][1] = -2 * dwfld * kex * zeta;
				//pdrv[2][2] = 0;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// par[0]	dw			(rad/s)
// par[1]	kab			(/s)
// par[2]	kba			(/s)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K)
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_richards_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw  = data.map[0];
	const int i_kab = data.map[1];
	const int i_kba = data.map[2];
	const int i_r20 = data.map[3];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double temp, field, dw, kab, kba, kex, psi, D_p, D_m, norm;
	static double pdrv[3][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (temp != data.x[1] || field != data.x[2]
	|| dw != par[0] || kab != par[1] || kba != par[2]){
		temp  = data.x[1];
		field = data.x[2];			// B0 / Base B0
		dw    = par[0];
		kab   = par[1];
		kba   = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		       kex   = kab + kba;
		double kex2  = sqr(kex);
		double pb    = kab / kex;
		double pop   = 1 - 2 * pb;	// pa - pb
		double zeta  = 2 * dwfld * kex * pop;
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_kab]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = -2 * dwfld * zeta;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_kba]) {
				pdrv[2][0] = 1;
				pdrv[2][1] = 2 * dwfld * zeta;
				pdrv[2][2] = 2 * kex;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// par[0]	dw			(rad/s)
// par[1]	kex			(/s)
// par[2]	papb			(unitless)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K)
// data.x[2]	B0 Field/Base B0 Field	(unitless)
// data.x[3]	probe			(unitless)
double fcpmg_richards_probes(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kex  = data.map[1];
	const int i_papb = data.map[2];
	const int i_r20  = data.map[3];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double temp, field, probe, dw, papb, kex, psi, D_p, D_m, norm;
	static double pdrv[3][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (temp != data.x[1] || field != data.x[2] || probe != data.x[3]
	|| dw != par[0] || kex != par[1] || papb != par[2]){
		temp  = data.x[1];
		field = data.x[2];			// B0 / Base B0
		probe = data.x[3];
		dw    = par[0];
		kex   = par[1];
		papb  = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = sqrt(1 - 4 * papb);	// pa - pb
		double zeta2 = sqr(2 * dwfld * kex * pop);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_kex]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta2 / kex;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_papb]) {
				//pdrv[2][0] = 0;
				pdrv[2][1] = -2 * zeta2 / sqr(pop);
				//pdrv[2][2] = 0;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// RC/TROSY/anti_TROSY wriiten by Sung-Hun Bae 10/20/2008
//                    modified by Kenji Sugase 11/26/2013
// par[0]	dw			(rad/s)
// par[1]	dJ			(Hz)
// par[2]	kex			(/s)
// par[3]	papb			(unitless)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
// data.x[3]	experimental type	(unitless)
//		0:RC, 1:TROSY, 2:Anti-TROSY
double fcpmg_rdc(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_dJ   = data.map[1];
	const int i_kex  = data.map[2];
	const int i_papb = data.map[3];
	const int i_r20  = data.map[4];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double temp, field, dw, dJ, papb, kex, psi, D_p, D_m, norm;
	static double pdrv[4][5];	// partial derivatives. dX/dR20 is not necessary
	static int pseq;
	// copy parameters if they changed
	if (temp != data.x[1] || field != data.x[2] || pseq != (int)data.x[3]
	 || dw != par[0] || dJ != par[2] || kex != par[2] || papb != par[3]){
		temp  = data.x[1];
		field = data.x[2];			// B0 / Base B0
		pseq  = (int)data.x[3];
		dw    = par[0];
		dJ    = par[1];
		kex   = par[2];
		papb  = par[3];
		double factor = 0;
		// calculate parameters
		if (pseq == 0) factor =  0;		// RC
		if (pseq == 1) factor = -M_PI;		// TROSY
		if (pseq == 2) factor = +M_PI;		// anti-TROSY
		// dw is scaled to magnetic field, but dJ is not
		// for RC          dw = dw
		// for TROSY:      dw = dw - 0.5J*pi
		// for anti-TROSY: dw = dw + 0.5J*pi
		double dwfld = dw * field + factor * dJ;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = sqrt(1 - 4 * papb);	// pa - pb
		double zeta  = 2 * dwfld * kex * pop;
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_dJ]) {
				//pdrv[1][0] = 0;
				pdrv[1][1] = zeta * factor * kex * pop;
				pdrv[1][2] = -2 * factor * dwfld;
				dtmp_D(pdrv[1], psi, inorm, pdw2, -1);
			}
			if (!fixed[i_kex]) {
				pdrv[2][0] = 1;
				pdrv[2][1] = zeta2 / kex;
				pdrv[2][2] = 2 * kex;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
			if (!fixed[i_papb]) {
				//pdrv[3][0] = 0;
				pdrv[3][1] = -2 * zeta2 / sqr(pop);
				//pdrv[3][2] = 0;
				dtmp_D(pdrv[3], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[4];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// par[0]	dw			(rad/s)
// par[1]	dHab			(kcal/mol)
// par[2]	dSab			(kcal/T/mol)
// par[3]	dHba			(kcal/mol)
// par[4]	dSba			(kcal/T/mol)
// par[5]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K)
// data.x[2]	B0 Field/Base B0 Field	(unitless)
double fcpmg_temp2(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define constant
	const double k_h  = 2.0836644e10;		// Boltzman's constant / Planck's constant
	const double R    = 1.985878e-3;		// Gas constant (kcal/mol/K)
	// define index
	const int i_dw   = data.map[0];
	const int i_dHab = data.map[1];
	const int i_dHba = data.map[2];
	const int i_dSab = data.map[3];
	const int i_dSba = data.map[4];
	const int i_r20  = data.map[5];
	// Speed up calculations by keeping parameters independent of 1/tcp
	static double T, field, dw, dHab, dHba, dSab, dSba, kex, psi, D_p, D_m, norm;
	static double pdrv[5][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (T    != data.x[1] || field != data.x[2] || dw   != par[0] || dHab != par[1]
	 || dHba != par[2]    || dSab  != par[3]    || dSba != par[4] ) {
		T     = data.x[1];			// temperature
		field = data.x[2];			// B0 / Base B0
		dw    = par[0];
		dHab  = par[1];				// Delta H for kab (kcal/mol)
		dHba  = par[2];				// Delta H for kba (kcal/mol)
		dSab  = par[3];				// Delta S for kab (kcal/mol)
		dSba  = par[4];				// Delta S for kba (kcal/mol)
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double tfac  = k_h * T;
		double kab   = tfac * exp((-dHab / T + dSab) / R);
		double kba   = tfac * exp((-dHba / T + dSba) / R);
		       kex   = kab + kba;
		double kex2  = sqr(kex);
		double zeta  = 2 * dwfld * (kab - kba);
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / (norm + TINY);
		double pdw2  = (kex2 + dw2) * inorm;	// (psi + 2dw^2) / sqrt(psi^2 + zeta^2)
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// Calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			double zeta_2dw = 2 * dwfld * zeta;
			if (!fixed[i_dHab]) {
				pdrv[1][0] = -kab / (R * T);
				pdrv[1][1] = zeta_2dw * pdrv[1][0];
				pdrv[1][2] = 2 * kex * pdrv[1][0];
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_dHba]) {
				pdrv[2][0] = -kba / (R * T);
				pdrv[2][1] = -zeta_2dw * pdrv[2][0];
				pdrv[2][2] = 2 * kex * pdrv[2][0];
				dtmp_D(pdrv[2], psi, inorm, pdw2);
				}
			if (!fixed[i_dSab]) {
				pdrv[3][0] = kab / R;
				pdrv[3][1] = zeta_2dw * pdrv[3][0];
				pdrv[3][2] = 2 * kex * pdrv[3][0];
				dtmp_D(pdrv[3], psi, inorm, pdw2);
			}
			if (!fixed[i_dSba]) {
				pdrv[4][0] = kba / R;
				pdrv[4][1] = -zeta_2dw * pdrv[4][0];
				pdrv[4][2] = 2 * kex * pdrv[4][0];
				dtmp_D(pdrv[4], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[5];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// Parameterized with [B]kon and koff
// Specialized for ligand binding event
// No knowledge of the actual free ligand concentration assumed
// par[0]	dw			(rad/s)
// par[1]	Bkon			(/s)
// par[2]	koff			(/s)
// par[3]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
// data.x[3]	concentration ratio	(unitless)
double fcpmg_bind2(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_Bkon = data.map[1];
	const int i_koff = data.map[2];
	const int i_r20  = data.map[3];
	// Speed up calculations by keeping parameters independent of 1/tcp
	static double field, conc, dw, Bkon, koff, kex, psi, D_p, D_m, norm;
	static double pdrv[3][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (field != data.x[2] || conc != data.x[3]
	|| dw != par[0] || Bkon != par[1] || koff  != par[2]    ){
		field = data.x[2];			// B0 / Base B0
		conc  = data.x[3];			// Concentration ratio
		dw    = par[0];
		Bkon  = par[1];				// [B]kon
		koff  = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		       kex   = Bkon + koff;
		double kex2  = sqr(kex);
		double zeta  = 2 * dwfld * (Bkon - koff);
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / (norm + TINY);
		double pdw2  = (kex2 + dw2) * inorm;	// (psi + 2dw^2) / sqrt(psi^2 + zeta^2)
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-

		// Calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			double zeta_2dw = 2 * dwfld * zeta;
			if (!fixed[i_Bkon]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta_2dw;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_koff]) {
				pdrv[2][0] = 1;
				pdrv[2][1] = -zeta_2dw;
				pdrv[2][2] = 2 * kex;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}

	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

// Carver and Richards approximation
// Parameterized with kom, koff, A0 and B0
// Specialized for ligand binding event
// Free ligand concentration is calculated by Kd and concentrations of proteins
// par[0]	dw			(rad/s)
// par[1]	kon			(/M/s)
// par[2]	koff			(/s)
// par[3]	A0			(M)
// par[4]	B0			(M)
// par[5]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
// data.x[3]	concentration ratio	(unitless)
double fcpmg_bind2c(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kon  = data.map[1];
	const int i_koff = data.map[2];
	const int i_a0   = data.map[3];
	const int i_b0   = data.map[4];
	const int i_r20  = data.map[5];
	// Speed up calculations by keeping parameters independent of 1/tcp
	static double field, conc, dw, kon, koff, a0, b0, kex, psi, D_p, D_m, norm;
	static double pdrv[5][5];	// partial derivatives. dX/dR20 is not necessary
	// copy parameters if they changed
	if (field != data.x[2] || conc != data.x[3] || dw != par[0] || kon != par[1]
	 || koff  != par[2]    || a0   != par[3]    || b0 != par[4] ) {
		field = data.x[2];			// B0 / Base B0
		conc  = data.x[3];			// ligand concnetration ratio
		dw    = par[0];
		kon   = par[1];
		koff  = par[2];
		a0    = par[3];				// [A0]
		b0    = par[4];				// [B0]
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double Kd    = koff / kon * 1e-6;	// Kd
		double c_tmp = -Kd - a0 + conc * b0;
		double freeB = 0.5 * (c_tmp + sqrt(sqr(c_tmp) + 4 * conc * b0 * Kd));	// [B]
		double Bkon  = freeB * kon * 1e6;	// [B]kon
		       kex   = Bkon + koff;
		double kex2  = sqr(kex);
		double zeta  = 2 * dwfld * (Bkon - koff);
		double zeta2 = sqr(zeta);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / (norm + TINY);
		double pdw2  = (kex2 + dw2) * inorm;	// (psi + 2dw^2) / sqrt(psi^2 + zeta^2)
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// Calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				//pdrv[0][0] = 0;
				pdrv[0][1] = zeta2 / dw;
				pdrv[0][2] = -2 * dw * sqr(field);
				dtmp_D(pdrv[0], psi, inorm, pdw2, -1);
			}
			double c_dnm = 1 / (2 * freeB - c_tmp);
			double zeta_2dw = 2 * dwfld * zeta;
			if (!fixed[i_kon]) {
				pdrv[1][0] = 1e6 * (Kd * c_dnm * (freeB - conc * b0) + freeB);
				pdrv[1][1] = zeta_2dw * pdrv[1][0];
				pdrv[1][2] = 2 * kex * pdrv[1][0];
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_koff]) {
				pdrv[2][0] = -c_dnm * (freeB - conc * b0) + 1;
				pdrv[2][1] = zeta_2dw * (pdrv[2][0] - 2);
				pdrv[2][2] = 2 * kex * pdrv[2][0];
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
			if (!fixed[i_a0]) {
				pdrv[3][0] = -c_dnm * Bkon;
				pdrv[3][1] = zeta_2dw * pdrv[3][0];
				pdrv[3][2] = 2 * kex * pdrv[3][0];
				dtmp_D(pdrv[3], psi, inorm, pdw2);
			}
			if (!fixed[i_b0]) {
				pdrv[4][0] = c_dnm * conc * kex;
				pdrv[4][1] = zeta_2dw * pdrv[4][0];
				pdrv[4][2] = 2 * kex * pdrv[4][0];
				dtmp_D(pdrv[4], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[5];
	return calc_R2(data, dy, fixed, pdrv, i_r20, kex, r20, psi, D_p, D_m, norm);
}

/*
// Carver and Richards approximation
// Parameterized with kom, koff, A0 and B0
// Specialized for ligand binding event
// Free ligand concentration is calculated by Kd and concentrations of proteins
// par[0]	dw			(rad/s)
// par[1]	kon			(/M/s)
// par[2]	koff			(/s)
// par[3]	A0			(M)
// par[4]	B0			(M)
// par[5]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	B0 Field/Base B0 Field	(unitless)
// data.x[3]	concentration ratio	(unitless)
double cpmg_bind2c(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// copy parameters if they changed
	double field = data.x[2];			// B0 / Base B0
	double conc  = data.x[3];			// ligand concnetration ratio
	double dw    = par[0];
	double kon   = par[1];
	double koff  = par[2];
	double a0    = par[3];				// [A0]
	double b0    = par[4];				// [B0]
	double r20   = par[5];
	// calculate parameters
	double dwfld = dw * field;
	double dw2   = sqr(dwfld);
	double Kd    = koff / kon ;	// Kd
	double c_tmp = -Kd - a0 + conc * b0;
	double freeB = 0.5 * (c_tmp + sqrt(sqr(c_tmp) + 4 * conc * b0 * Kd));	// [B]
	double Bkon  = freeB * kon;	// [B]kon
	double kex   = Bkon + koff;
	double kex2  = sqr(kex);
	double zeta  = 2 * dwfld * (Bkon - koff);
	double zeta2 = sqr(zeta);
	double psi   = kex2 - dw2;
	double norm  = sqrt(sqr(psi) + zeta2);
	double inorm = 1 / norm;
	double pdw2  = (kex2 + dw2) * inorm;	// (psi + 2dw^2) / sqrt(psi^2 + zeta^2)
	double D_p   = 0.5 * ( 1 + pdw2);	// D+
	double D_m   = 0.5 * (-1 + pdw2);	// D-
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double itcp = data.x[0];
	if (itcp < LOW_ITCP) {
		cerr << "1/tcp is too short. Check the input file\n";
		exit(1);
	}
	double tcp = 1 / itcp;
	double r2eff;
	// calculate effective R2
	double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
	double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
	if (eta_p < COSH_MAX) {
		double coshp = cosh(eta_p);			// cosh(eta+)
		double cosm  = cos(eta_m);			// cos(eta-)
		r2eff = r20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
	// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
	} else {
		r2eff = r20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
	}
	return r2eff;
}

double fcpmg_bind2c(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 5;
	return calc_deriv(data, par, dy, fixed, cpmg_bind2c, i_R20);
}
*/
}
int Funclib::Read_CPMG2_Funcs(int idx)
{
	// Constant R2
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_const, "CPMG_const"
		, "f(R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "R20"  ,  0.01,    50, "11" );

	// CPMG Luz & Meiboom approximation (kex >> dw)
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_meiboom, "CPMG_Meiboom"
		, "f(R20,pdw,kex)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "kex"  ,   0.1, 12000, "10" );
	Define_Parm(idx, "pdw"  , 1e-30, 3.6e7, "10" );
	Define_Parm(idx, "R20"  ,   0.1,    50, "11" );

	// CPMG Ishima & Torchia approximation (pb/pa < 0.18)
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_torchia, "CPMG_Torchia"
		, "f(dw, kex, papb, R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kex"  ,     1, 12000, "10" );
	Define_Parm(idx, "papb" , 0.001, 0.113, "10" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "11" );

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_richards, "CPMG_Richards"
		, "f(dw, kex, papb, R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kex"  ,     1, 12000, "10" );
	Define_Parm(idx, "papb" , 0.001, 0.249, "10" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "11" );

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_richards_p, "CPMG_Richards_p"
		, "f(dw, kex, pb, R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kex"  ,     1, 12000, "10" );
	Define_Parm(idx, "pb"   , 0.005,  0.49, "10" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "11" );

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_richards_k, "CPMG_Richards_k"
		, "f(dw, kab, kba, R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kab"  ,     1, 12000, "10" );
	Define_Parm(idx, "kba"  ,     1, 12000, "10" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "11" );

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	// index2: probe
	idx = Define_Func(&fcpmg_richards_probes, "CPMG_Richards_probes"
		, "f(dw, kex, papb, R20)", "CT-CPMG", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "001" );
	Define_Parm(idx, "kex"  ,     1, 12000, "100" );
	Define_Parm(idx, "papb" , 0.001, 0.249, "100" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "111" );

	// CPMG Carver & Richards approximation
	// RC / TROSY / anti-TROSY by Sung-Hun Bae 10/20/2008
	//                modified by Kenji Sugase 11/26/2013
	// index0: temperature
	// index1: magnetic field
	// index2: pulse sequence dependent
	//         0:RC, 1:TROSY, 2:anti-TROSY
	idx = Define_Func(&fcpmg_rdc,"CPMG_RDC"
		, "f(dw, dJ, kex, papb, R20)", "CT-CPMG", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "000" );
	Define_Parm(idx, "dJ"   ,  -200,   200,	"000" );
	Define_Parm(idx, "kex"  ,     1, 12000, "100" );
	Define_Parm(idx, "papb" , 0.001, 0.249,	"100" );
	Define_Parm(idx, "R20"  ,   0.1,   100,	"111" );

	// CPMG Carver & Richards approximation
	// Using Arrhenius's equation to describe dH & dS
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_temp2, "CPMG_TEMP2"
		, "f(dw, dHab, dHba, dSab, dSba, R20)", "CT-CPMG", XDIM3, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "dHab" ,   -50,    50, "00" );
	Define_Parm(idx, "dHba" ,   -50,    50, "00" );
	Define_Parm(idx, "dSab" ,    -1,     1, "00" );
	Define_Parm(idx, "dSba" ,    -1,     1, "00" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "11" );

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind2, "CPMG_Bind2"
		, "f(dw, koff, [B]kon, R20)", "CT-CPMG", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "000");
	Define_Parm(idx, "Bkon" ,     1, 12000, "101");
	Define_Parm(idx, "koff" ,     1, 12000, "100");
	Define_Parm(idx, "R20"  ,   0.1,   100, "111");

	// CPMG Carver & Richards approximation
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind2c, "CPMG_Bind2c"
		, "f(dw, kon, koff, A0, B0, R20)", "CT-CPMG", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "000");
	Define_Parm(idx, "kon"  ,     1,  1e11, "100");
	Define_Parm(idx, "koff" ,     1, 12000, "100");
	Define_Parm(idx, "A0"   ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "B0"   ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "R20"  ,   0.1,   100, "111");

	return idx;
}
