// Fit class
// Kenji Sugase		4/1/2013

#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>
#include <cstdlib>
#include <cmath>
#include <vector>
#include <sys/stat.h>
#include <unistd.h>
#include "Alloc.h"
#include "Fit.h"
#include "Mathlib.h"

using namespace std;

// prepare all variables for fitting
Fit::Fit(const string &VERSION_in, Funclib &func_in, Method &method_in, ParDef &parm_in
	, Calc &calc_in, Xmgr &xmgr_in, const string &output_in, int verb_in, int prec_in) :
	VERSION(VERSION_in), func(func_in), method(method_in), parm(parm_in), calc(calc_in),
	xmgr(xmgr_in), output(output_in), verb(verb_in), prec(prec_in)
{
	int szPar = parm.PAR_size();
	int szSet = parm.SETsize();
	// resize parameter sets
	pTEMP.resize(szSet, szPar);
	pBEST.resize(szSet, szPar);
	pGBEST.resize(szSet, szPar);
	pMEAN.resize(szSet, szPar);
	pSTDEV.resize(szSet, szPar);
	make_new(ppar, parm.data[0].map.size(), "fit");
	// prepare the Levenberg-Marquardt minimization stuff
	marq0.resize(szPar);
	marq.resize(szPar);
	lmarq.resize(szPar);
	covar.resize(szPar);
	da.resize(szPar);
	dyda.resize(szPar);
	stdev.resize(szPar);
	for (int i = 0; i < szPar; i++) {
		covar[i].resize(szPar);
	}
	lMat.resize(szSet);
	for (int ns = 0; ns < szSet; ns++) {
		lMat[ns].resize(szPar);
		lMat[ns].reindex(parm.mask_local[ns]);
		if (!lMat[ns].size()) {
			cerr << "WARNING (" << parm.TITLE(ns) << "): All parameters are fixed\n";
		}
		lMat[ns].resize(lMat[ns].size());
	}
	gMat.resize(szPar);
	gMat.reindex(parm.mask_global);
	gMat.resize(gMat.size());
	// prepare CALC stuff
	calc.prep(szSet);

//	mt19937 engine(chrono::system_clock::now().time_since_epoch().count());
	uniform_real_distribution<double> unidist(0.0, 1.0);
	normal_distribution<double> gausdist(0.0, 1.0);
}

// start fit
void Fit::run_FIT(bool sim)
{
	// save the start time
	time(&start_t);
	// display the header of the fitting result
	display_chi_header();
	// loop over METHOD instructions
	int sz = method.size();
	for (int nm = 0; nm < sz; nm++) {
		switch (method.Type(nm)) {
		case METHOD_INIT:
			run_INIT(nm, sim);
			if (verb == 2 && !sim) { make_dir(nm, "INIT"); }
			break;

		case METHOD_ONE:
			run_ONE(nm);
			if (verb == 2 && !sim) { make_dir(nm, "ONE"); }
			break;

		case METHOD_ONEEX:
			run_ONEEX(nm);
			if (verb == 2 && !sim) { make_dir(nm, "ONEEX"); }
			break;

		case METHOD_GRID:
			run_GRID(nm);
			if (verb == 2 && !sim) { make_dir(nm, "GRID"); }
			break;

		case METHOD_RANDOM:
			run_RANDOM(nm, method.Iter(nm));
			if (verb == 2 && !sim) { make_dir(nm, "RANDOM"); }
			break;

		case METHOD_MCMIN:
			run_MCMIN(nm, method.Iter(nm), method.Scale(nm));
			if (verb == 2 && !sim) { make_dir(nm, "MCMIN"); }
			break;

		case METHOD_RMCMIN:
			run_RMCMIN(nm, method.Iter(nm), method.IterMC(nm), method.Scale(nm));
			if (verb == 2 && !sim) { make_dir(nm, "RMCMIN"); }
			break;

		default:
			exit(1);
		}
	}
	calc_covar(pBEST);
	// store the results in GBEST
	if (!sim) { pGBEST = pBEST; }
	time(&end_t);
}

void Fit::make_dir(int num, const string &method)
{
	// make a directory
	ostringstream dir;
	dir << "0" << num + 1 << "-" << method;
	mkdir(dir.str().c_str(), S_IRUSR | S_IWUSR | S_IXUSR |	// rwx
                                 S_IRGRP |           S_IXGRP |	// r-x
                                 S_IROTH |           S_IXOTH);	// r-x
	// change directory
	if (chdir(dir.str().c_str())) {
		cerr << "no directory found\n";
		exit(1);
	}
	// report the result in progress
	time(&end_t);
	calc_covar(pBEST);
	pGBEST = pBEST;
	report_fit(num+1);
	if (xmgr.isXmgr()) {
		plot_xmgr();
	}
	if (chdir("..")) {
		cerr << "no directory found\n";
		exit(1);
	}
}

// set the parameters to the starting values and evaluate them
void Fit::run_INIT(size_t num, bool sim)
{
	// JACK and MONTE start with the GBEST parameter set
	if (sim) { pBEST = pGBEST;          }
	else     { check_parm_range(pBEST); }
	// calculate local and global chi2
	calc_chi(pBEST);
	if (!std::isfinite(pBEST.chi)) {
		cerr << "WARNING: Bad initial parameters\n";
	}
	if (verb) {
		if (is_local()) {
			int szSet = parm.SETsize();
			display_chi(num, "INIT", szSet, 1, 1, pBEST.chi);
		} else {
			display_chi(num, "INIT", 0, 1, 1, pBEST.chi);
		}
		cerr << endl;
	}
	// store the result
	store_chi2("# INIT         X2/DoF = ", pBEST.chi/parm.gDoF());
}

// run Levenberg-Marquardt minimization routine separately on global and local parameters
void Fit::run_ONE(size_t num)
{
	check_chi(pBEST);
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localONE(num, ns);
		}
	} else {
		run_globalONE(num);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# ONE          X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localONE(size_t num, size_t ns)
{
	// save the initial chi2 value
	double chi0 = pBEST.chi;
	// copy initial parameter set
	pTEMP = pBEST;
	// optimize the parameter set
	pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
	if (std::isfinite(pTEMP.lchi[ns]) && pTEMP.lchi[ns] < pBEST.lchi[ns]) {
		pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
		pTEMP.recalc_chi();
	} else {
		pTEMP = pBEST;
	}
	display_chi(num, "ONE", ns+1, 1, 1, pBEST.chi, chi0, pTEMP.chi);
}

void Fit::run_globalONE(size_t num)
{
	// save the initial chi2 value
	double chi0 = pBEST.chi;
	// copy initial parameter set
	pTEMP = pBEST;
	pTEMP.chi = minimize_global(pTEMP);
	if (std::isfinite(pTEMP.chi) && pTEMP.chi < pBEST.chi) {
		pBEST = pTEMP;
	}
	display_chi(num, "ONE", 0, 1, 1, pBEST.chi, chi0, pTEMP.chi);
}

// run Levenberg-Marquardt minimization routine separately on global and local parameters
void Fit::run_ONEEX(size_t num)
{
	check_chi(pBEST);
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localONEEX(num, ns);
		}
	} else {
		run_globalONEEX(num);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# ONEEX        X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localONEEX(size_t num, size_t ns)
{
	// save the initial chi2 value
	double chi0 = pBEST.chi;
	// copy initial parameter set
	pTEMP = pBEST;
	double best_chi = HUGE_VAL;
	while (pTEMP.chi < best_chi) {
		best_chi = pTEMP.chi;
		// optimize the parameter set
		pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
		if (std::isfinite(pTEMP.lchi[ns]) && pTEMP.lchi[ns] < pBEST.lchi[ns]) {
			pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
			pTEMP.recalc_chi();
		} else {
			pTEMP = pBEST;
		}
		display_chi(num, "ONEEX", ns+1, 1, 1, pBEST.chi, chi0, pTEMP.chi);
	}
}

void Fit::run_globalONEEX(size_t num)
{
	// save the initial chi2 value
	double chi0 = pBEST.chi;
	// copy initial parameter set
	pTEMP = pBEST;
	double best_chi = HUGE_VAL;
	while (pTEMP.chi < best_chi) {
		best_chi = pTEMP.chi;
		// optimize the parameter set
		pTEMP.chi = minimize_global(pTEMP);
		if (std::isfinite(pTEMP.chi) && pTEMP.chi < pBEST.chi) {
			pBEST = pTEMP;
		}
		display_chi(num, "ONEEX", 0, 1, 1, pBEST.chi, chi0, pTEMP.chi);
	}
}

// run a grid search without minimization
void Fit::run_GRID(size_t num)
{
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localGRID(num, ns);
		}
	} else {
		run_globalGRID(num);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# GRID         X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localGRID(size_t num, size_t ns)
{
	// copy initial parameter set
	pTEMP = pBEST;
	// set up grid
	unsigned long total = 1;
	int sz    = parm.nParBySet[ns].size();
	int first = parm.nParBySet[ns][0];
	int last  = parm.nParBySet[ns][sz-1];
	for (int n = 0; n < sz; n++) {
		int np = parm.nParBySet[ns][n];
		total *= parm.grd_size(np);
	}
	int szPar = parm.PAR_size();
	ivect nGrid;		nGrid.resize(szPar, 0);
	unsigned long count = 0;
	while (nGrid[last] < parm.grd_size(last)) {
		++count;
		// set a value to each parameter in a dataset
		for (int n = 0; n < sz; n++) {
			int np = parm.nParBySet[ns][n];
			pTEMP.p[np] = set_grid(np, nGrid[np]);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		// optimize the parameter set
		pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
		// update parameter set
		if (std::isfinite(pTEMP.lchi[ns]) && pTEMP.lchi[ns] < pBEST.lchi[ns]) {
			pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
		}
		// move to a new grid point
		++nGrid[first];
		for (int n = 0; n < sz; n++) {
			int np = parm.nParBySet[ns][n];
			if (np != last && nGrid[np] == parm.grd_size(np)) {
				nGrid[np] = 0;
				++nGrid[ parm.nParBySet[ns][n+1] ];
			}
		}
		pTEMP.recalc_chi();
		display_chi(num, "GRID", ns+1, count, total, pBEST.chi, chi0, pTEMP.chi);
	}
}

void Fit::run_globalGRID(size_t num)
{
	// copy initial parameter set
	pTEMP  = pBEST;
	ParSet pLOCAL = pBEST;		// parameter set for temporal parameters
	unsigned long count = 0;
	unsigned long total = 1;
	int glast = 0;
	int szPar = parm.PAR_size();
	for (int np = 0; np < szPar; np++) {
		if (parm.isGlobal(np)) {
			total *= parm.grd_size(np);
			glast = np;
		}
	}
	unsigned long local_grid = 0;
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		unsigned long each_grid = 1;
		int szp = parm.nParBySet[ns].size();
		for (int p = 0; p < szp; p++) {
			int np = parm.nParBySet[ns][p];
			if (parm.isLocal(np)) {
				each_grid *= parm.grd_size(np);
			}
		}
		local_grid += each_grid;
	}
	total *= local_grid;
	int gfirst = 0;
	ivect nGrid;		nGrid.resize(szPar, 0);
	ivect ngGrid;		ngGrid.resize(glast+1, 0);
	while (ngGrid[glast] < parm.grd_size(glast)) {
		// grid search of the global parameters of pLOCAL
		// initialize local parameters of pLOCAL
		for (int i = 0; i < szPar; i++) {
			if (parm.isGlobal(i)) {
				pLOCAL.p[i] = set_grid(i, ngGrid[i]);
			} else {
				pLOCAL.p[i] = parm.grd_start(i);
			}
		}
		pLOCAL.chi = 0.0;
		for (int ns = 0; ns < szSet; ns++) {
			pLOCAL.lchi[ns] = calc_chi(pLOCAL, ns);
			pLOCAL.chi     += pLOCAL.lchi[ns];
		}
		pTEMP = pLOCAL;
		// save the initial chi2 value
		double chi0 = pLOCAL.chi;
		// set a value to each local parameter
		for (int ns = 0; ns < szSet; ns++) {
			int first = 0;
			int last = 0;
			int sz = parm.nParBySet[ns].size();
			// find the first and last local parameters in a parameter list
			bool find = false;
			for (int np = 0; np < sz; np++) {
				int i = parm.nParBySet[ns][np];
				if (parm.isLocal(i)) {
					if (find == false) {
						first = i;
						find = true;
					} else {
						last = i;
					}
				}
			}
			double best_lchi = HUGE_VAL;
			while (nGrid[last] < parm.grd_size(last)) {
				// set a value to each local parameter in a dataset
				for (int np = 0; np < sz; np++) {
					int i = parm.nParBySet[ns][np];
					if (parm.isLocal(i)) {
						pLOCAL.p[i] = set_grid(i, nGrid[i]);
					}
				}
				// optimize the parameter set
				pLOCAL.lchi[ns] = minimize_local(pLOCAL, ns);
				if (std::isfinite(pLOCAL.lchi[ns])) {
					if (pLOCAL.lchi[ns] < best_lchi) {
						pTEMP.copy(pLOCAL, parm.nParBySet[ns], ns);
						best_lchi = pLOCAL.lchi[ns];
					}
					if (pTEMP.chi < pBEST.chi) {
						pBEST.chi = pTEMP.chi;
					}
				}
				// move to a new grid point
				++nGrid[first];
				for (int np = 0; np < sz; np++) {
					int i = parm.nParBySet[ns][np];
					if (parm.isLocal(i) && i != last && nGrid[i] == parm.grd_size(i)) {
						nGrid[i] = 0;
						for (int j = 1; j < sz; j++) {
							int next = parm.nParBySet[ns][np+j];
							if (parm.isLocal(next)) {
								++nGrid[next];
								break;
							}
						}
					}
				}
				++count;
				pLOCAL.recalc_chi();
				display_chi(num, "GRID", 0, count, total, pBEST.chi, chi0, pLOCAL.chi);
			}
			for (int np = 0; np < szPar; np++) {
				nGrid[np] = 0;
			}
		}
		// run Levenberg-Marquardt minimization
		pTEMP.chi = minimize_global(pTEMP);
		// update parameter set
		if (std::isfinite(pTEMP.chi) && pTEMP.chi < pBEST.chi) {
			pBEST = pTEMP;
		}
		display_chi(num, "GRID", 0, count, total, pBEST.chi, chi0, pTEMP.chi);
		// move to a new grid point
		++ngGrid[gfirst];
		for (int np = 0; np <= glast; np++) {
			if (np != glast && ngGrid[np] == parm.grd_size(np)) {
				ngGrid[np] = 0;
				++ngGrid[np+1];
			}
		}
	}
}

void Fit::run_RANDOM(size_t num, int iter)
{
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localRANDOM(num, ns, iter);
		}
	} else {
		run_globalRANDOM(num, iter);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# RANDOM       X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localRANDOM(size_t num, size_t ns, int iter)
{
	// copy initial parameter set
	pTEMP = pBEST;
	for (int i = 0; i < iter; i++) {
		// pick new parameter values
		int szp = parm.nParBySet[ns].size();
		for (int n = 0; n < szp; n++) {
			int np = parm.nParBySet[ns][n];
			rand_parm(np, pTEMP);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		// optimize the parameter set
		pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
		if (!std::isfinite(pTEMP.lchi[ns])) { continue; }
		if (!std::isfinite(pBEST.chi) || pTEMP.lchi[ns] < pBEST.lchi[ns]) {
			pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
			pBEST.recalc_chi();
		}
		pTEMP.recalc_chi();
		display_chi(num, "RANDOM", ns+1, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
	}
}

void Fit::run_globalRANDOM(size_t num, int iter)
{
	// copy initial parameter set
	pTEMP = pBEST;
	for (int i = 0; i < iter; i++) {
		// pick new parameter values
		int szPar = parm.PAR_size();
		for (int np = 0; np < szPar; np++) {
			rand_parm(np, pTEMP);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		// optimize the parameter set
		pTEMP.chi = minimize_global(pTEMP);
		if (!std::isfinite(pTEMP.chi)) { continue; }
		if (!std::isfinite(pBEST.chi) || pTEMP.chi < pBEST.chi) {
			pBEST = pTEMP;
		}
		display_chi(num, "RANDOM", 0, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
	}
}

// run Monte Carlo with Minimization
void Fit::run_MCMIN(size_t num, int iter, double scale)
{
	check_chi(pBEST);
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localMCMIN(num, ns, iter, scale);
		}
	} else {
		run_globalMCMIN(num, iter, scale);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# MCMIN        X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localMCMIN(size_t num, size_t ns, int iter, double scale)
{
	// copy initial parameter set
	pTEMP = pBEST;
	for (int i = 0; i < iter; i++) {
		// reset parameter set
		pTEMP = pBEST;
		// pick new parameter values
		int szp = parm.nParBySet[ns].size();
		for (int n = 0; n < szp; n++) {
			int np = parm.nParBySet[ns][n];
			mc_parm(np, pTEMP, scale);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		if (!std::isfinite(chi0)) { continue; }
		// optimize the parameter set
		pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
		if (!std::isfinite(pTEMP.lchi[ns])) { continue; }
		if (pTEMP.lchi[ns] < pBEST.lchi[ns]) {
			pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
			i = 0;
		}
		pTEMP.recalc_chi();
		display_chi(num, "MCMIN", ns+1, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
	}
}

void Fit::run_globalMCMIN(size_t num, int iter, double scale)
{
	// copy initial parameter set
	pTEMP = pBEST;
	for (int i = 0; i < iter; i++) {
		// reset parameter set
		pTEMP = pBEST;
		// pick new parameter values
		int szPar = parm.PAR_size();
		for (int np = 0; np < szPar; np++) {
			mc_parm(np, pTEMP, scale);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		if (!std::isfinite(chi0)) { continue; }
		// optimize the parameter set
		pTEMP.chi = minimize_global(pTEMP);
		if (!std::isfinite(pTEMP.chi)) { continue; }
		// run Levenberg-Marquardt minimization
		if (pTEMP.chi < pBEST.chi) {
			pBEST = pTEMP;
			i = 0;
		}
		display_chi(num, "MCMIN", 0, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
	}
}

// run Monte Carlo with Minimization
void Fit::run_RMCMIN(size_t num, int iter, int iterMC, double scale)
{
	// copy initial parameter set
	pTEMP = pBEST;
	ParSet pB_MC;			// best parameter set in each MC calculation
	if (is_local()) {
		int szSet = parm.SETsize();
		for (int ns = 0; ns < szSet; ns++) {
			run_localRMCMIN(num, ns, iter, iterMC, scale);
		}
	} else {
		run_globalRMCMIN(num, iter, iterMC, scale);
	}
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# RMCMIN       X2/DoF = ", pBEST.chi/parm.gDoF());
}

void Fit::run_localRMCMIN(size_t num, size_t ns, int iter, int iterMC, double scale)
{
	// copy initial parameter set
	pTEMP = pBEST;
	ParSet pB_MC;
	for (int i = 0; i < iter; i++) {
		// pick new parameter values
		int szp = parm.nParBySet[ns].size();
		for (int n = 0; n < szp; n++) {
			int np = parm.nParBySet[ns][n];
			rand_parm(np, pTEMP);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		// optimize the parameter set
		pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
		if (!std::isfinite(pTEMP.lchi[ns])) { continue; }
		if (!std::isfinite(pBEST.chi) || pTEMP.lchi[ns] < pBEST.lchi[ns]) {
			pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
		}
		pTEMP.recalc_chi();
		display_chi(num, "RMCMIN", ns+1, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
		// start MCMIN
		pB_MC = pTEMP;
		for (int j = 0; j < iterMC; j++) {
			// reset parameter set
			pTEMP = pB_MC;
			// pick new parameter values
			int szp = parm.nParBySet[ns].size();
			for (int n = 0; n < szp; n++) {
				int np = parm.nParBySet[ns][n];
				mc_parm(np, pTEMP, scale);
			}
			// optimize the parameter set
			pTEMP.lchi[ns] = minimize_local(pTEMP, ns);
			if (!std::isfinite(pTEMP.lchi[ns])) { continue; }
			if (pTEMP.lchi[ns] < pB_MC.lchi[ns]) {
				pB_MC.copy(pTEMP, parm.nParBySet[ns], ns);
				j = 0;
			}
			if (pTEMP.lchi[ns] < pBEST.lchi[ns]) {
				pBEST.copy(pTEMP, parm.nParBySet[ns], ns);
			}
			pTEMP.recalc_chi();
			display_chi(num, "RMCMIN", ns+1, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
		}
	}
}

void Fit::run_globalRMCMIN(size_t num, int iter, int iterMC, double scale)
{
	// copy initial parameter set
	pTEMP = pBEST;
	ParSet pB_MC;
	for (int i = 0; i < iter; i++) {
		// pick new parameter values
		int szPar = parm.PAR_size();
		for (int np = 0; np < szPar; np++) {
			rand_parm(np, pTEMP);
		}
		// calculate the initial chi2 value
		double chi0 = calc_chi(pTEMP);
		// optimize the parameter set
		pTEMP.chi = minimize_global(pTEMP);
		if (!std::isfinite(pTEMP.chi)) { continue; }
		if (!std::isfinite(pBEST.chi) || pTEMP.chi < pBEST.chi) {
			pBEST = pTEMP;
		}
		pTEMP.recalc_chi();
		display_chi(num, "RMCMIN", 0, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
		// start MCMIN
		pB_MC = pTEMP;
		for (int j = 0; j < iterMC; j++) {
			// reset parameter set
			pTEMP = pB_MC;
			// pick new parameter values
			int szPar = parm.PAR_size();
			for (int np = 0; np < szPar; np++) {
				mc_parm(np, pTEMP, scale);
			}
			// optimize the parameter set
			pTEMP.chi = minimize_global(pTEMP);
			if (!std::isfinite(pTEMP.chi)) { continue; }
			if (pTEMP.chi < pB_MC.chi) {
				pB_MC = pTEMP;
				j = 0;
			}
			if (pTEMP.chi < pBEST.chi) {
				pBEST = pTEMP;
			}
			display_chi(num, "RMCMIN", 0, i+1, iter, pBEST.chi, chi0, pTEMP.chi);
		}
	}
}

// check the parameter range
void Fit::check_parm_range(ParSet &pr)
{
	// check the lower limits
	int szPar = parm.PAR_size();
	for (int np = 0; np < szPar; np++) {
		pBEST.p[np] = parm.grd_start(np);
	}
	check_parm(pBEST);
	// check the upper limits
	for (int np = 0; np < szPar; np++) {
		pBEST.p[np] = parm.grd_end(np);
	}
	check_parm(pBEST);
	// check the initial parameters
	for (int np = 0; np < szPar; np++) {
		pBEST.p[np] = parm.grd_initial(np);
	}
	check_parm(pBEST);
}

// check if the starting parameters are legal
void Fit::check_parm(const ParSet &pr)
{
	bool invalid = false;
	int szPar = parm.PAR_size();
	for (int np = 0; np < szPar; np++) {
		if (!parm.isLegal(np, pr.p[np])) {
			if (!invalid) {
				invalid = true;
				cerr << "ERROR: Illegal initial parameters\n";
			}
			cerr << parm.FullName(np) << " " << pr.p[np] << endl;
		}
	}
	if (invalid) { exit(1); }
}

void Fit::check_chi(const ParSet &pr) const
{
	if (!std::isfinite(pr.chi)) {
		cerr << "ERROR: Bad initial parameters\n";
		exit(1);
	}
}

// calculate a local chi2 value
double Fit::calc_chi(ParSet &pr, size_t ns)
{
	// calculate local chi2
	dvect dy0;
	pr.lchi[ns] = 0.0;
	int szd = parm.nDataBySet[ns].size();
	for (int np = 0; np < szd; np++) {
		int nd = parm.nDataBySet[ns][np];
		double y = calc_y_dy(pr.p, nd, dy0, bvect());
		if (!std::isfinite(y)) { return HUGE_VAL; }
		double dy = parm.data[nd].y - y;
		pr.lchi[ns] += sqr(dy * parm.data[nd].idy);
	}
	return pr.lchi[ns];
}

// calculate a global chi2 value
double Fit::calc_chi(ParSet &pr)
{
	pr.chi = 0;
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		pr.chi += calc_chi(pr, ns);
	}
	return pr.chi;
}

double Fit::calc_y_dy(const dvect &p, int nd, dvect &dy, const bvect &fixed)
{
	int sz = parm.data[nd].map.size();
	for (int i = 0; i < sz; i++) {
		ppar[i] = p[parm.data[nd].map[i]];
	}
	return  (*func.fx)(parm.data[nd], ppar, dy, fixed);
}

double Fit::set_grid(size_t i, int grid_pt)
{
	double ret = parm.grd_start(i);
	switch (parm.grd_interval(i)) {
	case GRID_NORMAL:
		ret = parm.grd_start(i) + grid_pt * parm.grd_step(i);
		break;
	case GRID_EXP:
		if (grid_pt != 0) {
			ret += exp(grid_pt * parm.grd_step(i));
		}
		break;
	case GRID_GAUSS:
		if (grid_pt != 0) {
			ret += exp(sqr(grid_pt * parm.grd_step(i)));
		}
		break;
	case GRID_TENTH:
		if (grid_pt != 0) {
			ret += pow(10, grid_pt * parm.grd_step(i));
		}
		break;
	case GRID_iEXP:
		if (grid_pt == 0) {
			ret = parm.grd_end(i);
		} else {
			ret = parm.grd_end(i) - exp(grid_pt * parm.grd_step(i));
		}
		break;
	case GRID_iGAUSS:
		if (grid_pt == 0) {
			ret = parm.grd_end(i);
		} else {
			ret = parm.grd_end(i) - exp(sqr(grid_pt * parm.grd_step(i)));
		}
		break;
	case GRID_iTENTH:
		if (grid_pt == 0) {
			ret = parm.grd_end(i);
		} else {
			ret = parm.grd_end(i) - pow(10, grid_pt * parm.grd_step(i));
		}
		break;
	}
	return ret;
}

void Fit::rand_parm(size_t np, ParSet &pr)
{
	if (parm.grd_range(np) > 0 && parm.isGRID(np)) {
		do {
			pr.p[np] = parm.grd_start(np) + unirand() * parm.grd_range(np);
		} while (!parm.isLegal(np, pr.p[np]));
	} else {
		pr.p[np] = parm.grd_start(np);
	}
}

void Fit::mc_parm(size_t np, ParSet &pr, double scale)
{
	if (parm.isGRID(np)){
		double new_p;
		do {
			new_p = pr.p[np] * (1 + gaussrand() * scale);
		} while (!parm.isLegal(np, new_p));
			pr.p[np] = new_p;
	}
}

void Fit::display_chi_header()
{
	if (!verb) { return; }
	if(is_local()) {
		cerr << "-------------------------------------------------------------------------\n"
		     << "  METHOD (set ,       #/  total)          X2/DoF           | best X2/DoF\n"
		     << "-------------------------------------------------------------------------\n";
	} else {
		cerr << "------------------------------------------------------------------\n"
		     << "  METHOD (      #/  total)         X2/DoF            | best X2/DoF\n"
		     << "------------------------------------------------------------------\n";
	}
}

// display chi2/DoF during a fit
void Fit::display_chi(int number, const char *method, int nset, int counter, int total
	, double chi_best, double chi0, double chi)
{
	if (!verb) { return; }
	if (number == 0) { chi = chi_best; }	// this is used only for INIT.
	cerr << "\r" << number << "." << left
	     << setw(6) << setfill(' ') << method
	     << " (" << right << setfill(' ');
	if (nset) {
		cerr << setw(4) << nset << ", ";
	}
	 cerr << setw(7) << counter << "/";
	if (number == 0) {
		cerr << setw(7) << total << ")                " << left;
	} else {
		cerr << setw(7) << total << ")  " << left;
		if ((chi0/parm.gDoF() > 1e-5 && chi0/parm.gDoF() < 1e5 )|| chi0/parm.gDoF() == 0) {
			cerr << setw(10) << setfill(' ') << chi0/parm.gDoF() << " -> ";
		} else {
			cerr << setprecision(5) << showpoint << chi0/parm.gDoF() << " -> ";
			cerr << resetiosflags(ios_base::floatfield) << noshowpoint;
		}
	}
	if ((chi/parm.gDoF() > 1e-5 && chi/parm.gDoF() < 1e5) || chi/parm.gDoF() == 0) {
		cerr << setw(10) << setfill(' ') << chi/parm.gDoF();
	} else {
		cerr << setprecision(5) << showpoint << chi/parm.gDoF();
		cerr << resetiosflags(ios_base::floatfield) << noshowpoint;
	}
	cerr << " | ";
	if ((chi_best/parm.gDoF() > 1e-5 && chi_best/parm.gDoF() < 1e5) || chi_best/parm.gDoF() == 0) {
		cerr << setw(10) << setfill(' ') << chi_best/parm.gDoF();
	} else {
		cerr << setprecision(5) << showpoint << chi_best/parm.gDoF();
		cerr << resetiosflags(ios_base::floatfield) << noshowpoint;
	}
}

double Fit::minimize_local(ParSet &pr, size_t ns)
{
	ParSet atry;
	return LevMarq(pr, atry, parm.nDataBySet[ns], parm.nParBySet[ns], lMat[ns], marq, &Fit::lMarq);
}

// Run Levenberg-Marquardt minimization routine separately on global and local parameters
double Fit::minimize_global(ParSet &pr)
{
	ParSet atry;
	return LevMarq(pr, atry, parm.allData, parm.allPAR, gMat, marq, &Fit::gMarq);
}

// calculate chi2 for a given data subset and construct alpha and beta
double Fit::lMarq(ParSet &pr, const ivect &dlist, const ivect &plist, const FitMat &mat, LevMar &levmar)
{
	// initialize alpha and beta
	// alpha and beta have the scrambled parameter order
	int szp = plist.size();
	for (int i = 0; i < szp; i++) {
		int np1 = mat.inv[plist[i]];
		levmar.beta[np1] = 0.0;
		for (int j = 0; j < szp; j++) {
			int np2 = mat.inv[plist[j]];
			levmar.alpha[np1][np2] = 0.0;
		}
	}
	// construct alpha and beta
	double chi2 = 0.0;
	int szd = dlist.size();
	for (int i = 0; i < szd; i++) {
		// reset partial derivatives
		// dyda is in the original order
		for (int j = 0; j < szp; j++) {
			dyda[plist[j]] = 0.0;
		}
		// calculate global chi2
		int nd = dlist[i];
		double y = calc_y_dy(pr.p, nd, dyda, parm.fixed);
		if (!std::isfinite(y)) { return HUGE_VAL; }
		double dy = parm.data[nd].y - y;
		double idy2 = sqr(parm.data[nd].idy);
		chi2 += sqr(dy) * idy2;
		// update matrix alpha and vector beta
		for (int j = 0; j < szp; j++) {
			int np1 = mat.inv[plist[j]];
			if (!std::isfinite(dyda[plist[j]])) { return HUGE_VAL; }
			double wt;		// temporal variable to calculate alpha and beta
			if ((wt = dyda[plist[j]] * idy2)) {
				levmar.beta[np1] += wt * dy;
				for (int k = 0; k < szp; k++) {
					int np2 = mat.inv[plist[k]];
					if (!std::isfinite(dyda[plist[k]])) { return HUGE_VAL; }
					levmar.alpha[np1][np2] += wt * dyda[plist[k]];
				}
			}
		}
	}
	return chi2;
}

// run Levenberg-Marquardt minimization routine separately on local parameters of each set
// separately to speed calculations
double Fit::gMarq(ParSet &pr, const ivect &dlist, const ivect &plist, const FitMat &mat, LevMar &levmar)
{
	int szPar = parm.PAR_size();
	int szSet = parm.SETsize();
	static ParSet atry;
	atry.resize(szSet, szPar);
	// initialize alpha, beta, and chi
	levmar.zero();
	pr.chi = 0.0;
	for (int ns = 0; ns < szSet; ns++) {
		pr.lchi[ns] = LevMarq(pr, atry, parm.nDataBySet[ns], parm.nParBySet[ns], lMat[ns]
			, lmarq, &Fit::lMarq);
		if (!std::isfinite(pr.lchi[ns])) { return HUGE_VAL; }
		pr.chi += pr.lchi[ns];
		// accumulate alpha and beta
		int szMat = mat.size();
		for (int i = 0; i < szMat; i++) {
			int np1 = lMat[ns].inv[mat.pidx[i]];
			levmar.beta[i] += lmarq.beta[np1];
			for (int j = 0; j < szMat; j++) {
				int np2 = lMat[ns].inv[mat.pidx[j]];
				levmar.alpha[i][j] += lmarq.alpha[np1][np2];
			}
		}
	}
	return pr.chi;
}

// Levenberg-Marquardt minimization routine
double Fit::LevMarq(ParSet &pr, ParSet &pTRY, const ivect &dlist, const ivect &plist
	, const FitMat &mat, LevMar &levmar, evalsub feval)
{
	const int MaxIter = 300;	// maximum number of iteration
	const double Tol  = 1e-12;
	const double Conv = 1e-15;
	const double Tau  = 1e-8;
	const double Up   = 2;
	const double Down = 4;
	double nu = 2;
	// calculate the initial chi2
	double chi2 = 0;
	if (!std::isfinite(chi2 = (this->*feval)(pr, dlist, plist, mat, levmar))) {
		return HUGE_VAL;
	}
	// set the initial lambda
	double lambda = 1;
	int szMat = mat.size();
	for (int nm = 0; nm < szMat; nm++) {
		int np = mat.inv[mat.pidx[nm]];
		if (lambda < levmar.alpha[np][np]) {
			lambda = levmar.alpha[np][np];
		}
	}
	lambda *= Tau;
	// prepare the initial alpha, beta, and parameter set
	d2vect alpha0;
	alpha0.resize(szMat);
	for (int nm = 0; nm < szMat; nm++) {
		alpha0[nm].resize(szMat);
	}
	dvect beta0;
	beta0.resize(szMat);
	pTRY = pr;
	// iterative search until convergence or iteration maximum reached
	for (int iter = 0; iter < MaxIter; iter++) {
		bool NoChange = true;
		for (int nm1 = 0; nm1 < szMat; nm1++) {
			int np1 = mat.inv[mat.pidx[nm1]];
			for (int nm2 = 0; nm2 < szMat; nm2++) {
				int np2 = mat.inv[mat.pidx[nm2]];
				alpha0[nm1][nm2] = levmar.alpha[np1][np2];
			}
			beta0[nm1] = levmar.beta[np1];
			if (fabs(beta0[nm1]) > Conv) {
				NoChange = false;
			}
		}
		if (NoChange) {
			return chi2;
		}
		// slove (alpha + lambda*I)*da = beta
		if (LM_solv(szMat, alpha0, beta0, da, lambda)) {
			lambda *= Up;
			continue;
		}
		// prevent needless iteration: quit when the step size vanishes
		// step size vanishing criteria: < Conv change in all parameters
		NoChange = true;
		for (int nm = 0; nm < szMat; nm++) {
			int np = mat.pidx[nm];
			if (fabs(da[nm]) > Conv*fabs(pr.p[np])) {
				NoChange = false;
				break;
			}
		}
		if (NoChange) {
			return chi2;
		}
		// check if parameters go out of bounds
		bool legal = true;
		for (int nm = 0; nm < szMat; nm++) {
			int np = mat.pidx[nm];
			pTRY.p[np] = pr.p[np] + da[nm];
			if (!parm.isLegal(np, pTRY.p[np])) {
				legal = false;
				break;
			}
		}
		if (!legal) {
			lambda *= Up;
			continue;
		}
		// evaluate chi2
		double chi2c;		// current chi2
		if (!std::isfinite(chi2c = (this->*feval)(pTRY, dlist, plist, mat, marq0))
		 || chi2c >= chi2) {
			// This fit is worse.  Increase lambda and try again.
			lambda *= nu;
			nu *= 2;
			continue;
		}
		if (chi2 - chi2c < Tol) {
			return chi2c;
		}
		// update chi2
		chi2  = chi2c;
		lambda /= Down;
		nu = Up;
		// update alpha and beta
		// alpha and beta must have the scrambled parameter order
		for (int nm1 = 0; nm1 < szMat; nm1++) {
			int np1 = mat.inv[mat.pidx[nm1]];
			for (int nm2 = 0; nm2 < szMat; nm2++) {
				int np2 = mat.inv[mat.pidx[nm2]];
				levmar.alpha[np1][np2] = marq0.alpha[np1][np2];
			}
			levmar.beta[np1] = marq0.beta[np1];
		}
		// update parameter set
		// pTRY and pr.p must have the parameters in the original order
		if (feval == &Fit::lMarq) {
			for (int nm = 0; nm < szMat; nm++) {
				int np = mat.pidx[nm];
				pr.p[np] = pTRY.p[np];
			}
		} else {
			pr.p = pTRY.p;
		}
	}
	return chi2;
}

// Calculate covariance matrix & estimate stdev
void Fit::calc_covar(const ParSet &pr)
{
	if(!std::isfinite(pr.chi)) { exit(1); }
	int szPar = parm.PAR_size();
	// initialize dyda and alpha
	fill(dyda.begin(), dyda.end(), 0.0);
	marq0.zero_alpha();
	int szSet = parm.SETsize();
	if (is_global()) {	// w/ GPAR
		for (int ns = 0; ns < szSet; ns++) {
			int szd = parm.nDataBySet[ns].size();
			int szp = parm.nParBySet[ns].size();
			for (int nd = 0; nd < szd; nd++) {
				// calculate dyda
				int n = parm.nDataBySet[ns][nd];
				calc_y_dy(pr.p, n, dyda, parm.fixed);
				// calculate alpha
				double idy2 = sqr(parm.data[n].idy);
				for (int n1 = 0; n1 < szp; n1++) {
					int np1 = parm.nParBySet[ns][n1];
					if (!parm.fixed[np1]) {
						double wt = dyda[np1] * idy2;
						for (int n2 = 0; n2 < szp; n2++) {
							int np2 = parm.nParBySet[ns][n2];
							if (!parm.fixed[np2]) {
								marq0.alpha[np1][np2] += wt * dyda[np2];
							}
						}
					}
				}
			}
		}
		if (LM_inv(gMat.size(), marq0.alpha)) {
			cerr << "ERROR: Standard deviations cannot be computed\n";
		}
	} else {			// w/o GPAR
		for (int ns = 0; ns < szSet; ns++) {
			int szd = parm.nDataBySet[ns].size();
			int szp = parm.nParBySet[ns].size();
			// initialize local alpha
			lmarq.resize(szp);
			lmarq.zero_alpha();
			for (int nd = 0; nd < szd; nd++) {
				// calculate dyda
				int n = parm.nDataBySet[ns][nd];
				calc_y_dy(pr.p, n, dyda, parm.fixed);
				// calculate local alpha
				double idy2 = sqr(parm.data[n].idy);
				for (int n1 = 0; n1 < szp; n1++) {
					int np1 = parm.nParBySet[ns][n1];
					if (!parm.fixed[np1]) {
						double wt = dyda[np1] * idy2;
						for (int n2 = 0; n2 < szp; n2++) {
							int np2 = parm.nParBySet[ns][n2];
							if (!parm.fixed[np2]) {
								lmarq.alpha[n1][n2] += wt * dyda[np2];
							}
						}
					}
				}
			}
			if (LM_inv(lMat[ns].size(), lmarq.alpha)) {
				cerr << "ERROR: Standard deviations cannot be computed"
				     << " for SET '" << parm.TITLE(ns) << "'" << endl;
			}
			for (int n1 = 0; n1 < szp; n1++) {
				int np1 = parm.nParBySet[ns][n1];
				for (int n2 = 0; n2 < szp; n2++) {
					int np2 = parm.nParBySet[ns][n2];
					marq0.alpha[np1][np2] = lmarq.alpha[n1][n2];
				}
			}
		}
	}
	for (int np = 0; np < szPar; np++) {
		stdev[np] = sqrt(marq0.alpha[np][np]);
	}
}

void Fit::report_fit(int num)
{
	ofstream rout;
	rout.open(output.c_str());
	check_file(rout, output);
	report_header(rout, num);
	report_time(rout);
	report_FitParm(rout);
	rout.close();
}

// report header
void Fit::report_header(ofstream &rout, int num)
{
	rout << "# GLOVE VERSION " << VERSION << endl;

	// output regular header
	rout << "MODEL  " << func.Name() << endl
	     << "EQN    " << func.Eqn()  << endl;
	for (int nm = 0; nm < num; nm++) {
		rout << method.report(nm) << endl;
	}
	rout << "\n# Fitting results\n"
	     << "# Total data points = " << parm.data.size() << endl
	     << "# Total degrees of freedom (DoF) = " << parm.gDoF() << endl;
	// output METHOD results
	for (int nm = 0; nm < num; nm++) {
		rout << report_chi2[nm] << endl;
	}
}

void Fit::report_time(ofstream &rout)
{
	int diff = static_cast<int>(difftime(end_t, start_t));
	int hour = diff / 3600;
	int min  = (diff % 3600) / 60;
	int sec  = diff % 60;
	rout << "# Fit duration        " << hour << "h" << min << "m" << sec << "s\n\n";
}

// Report the fit results
void Fit::report_FitParm(ofstream &rout)
{
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		rout << "SET\t" << parm.TITLE(ns) << endl;
		int sz = parm.nParBySet[ns].size();
		for (int np = 0; np < sz; np++) {
			int n = parm.nParBySet[ns][np];
			rout << "\t" << parm.ParNameBySet[ns][np] << "\t"
			     << setw(7) << setprecision(prec) << pGBEST.p[n];
			if (!parm.fixed[n]) {
				rout << " (" << stdev[n] << ")";
			}
			rout << endl;
		}
		int szc = calc.size();
		for (int nc = 0; nc < szc; nc++) {
			rout << "\t    " << calc.Label(nc)  << "\t" << run_CALC(ns, nc, pGBEST) << endl;
		}
		rout << "\tX2/DoF\t" << pGBEST.lchi[ns] / parm.lDoF(ns)
		     << " (DoF = " << parm.lDoF(ns) << ")\n\n";
	}

	rout << "Global X2/DoF\t" << pGBEST.chi / parm.gDoF()
	     << " (DoF = " << parm.gDoF() << ")\n";
}

// run jackknife error calculations
void Fit::run_JACK()
{
	// initialize parameter sets
	pMEAN.zero();
	pSTDEV.zero();
	calc.zero();
	// store the calculation results for future reference
	int szc = calc.size();
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		for (int nc = 0; nc < szc; nc++) {
			calc.storeResult(ns, nc, run_CALC(ns, nc, pGBEST));
		}
	}
	// open the output file
	ofstream rout;
	string filename = output + "-jack";
	rout.open(filename.c_str());
	check_file(rout, filename);
	rout << "Jackknife simulations (" << parm.data.size() << ")\n";

	int iter = parm.data.size();
	if (is_local()) {
		run_JACKlocal(rout);
	} else {
		for (int ns = 0; ns < szSet; ns++) {
			parm.setlDoF(ns, parm.lDoF(ns)-1);
		}
		for (int np = 0; np < iter; np++) {
			if (verb) {
				cerr << "\n# Jackknife simulation #" << np+1 << "/" << iter << endl;
			}
			// remove the 1st data point from the whole and local lists
			int hide_nd = parm.allData[0];
			int hide_ns = parm.data[hide_nd].ds;
			parm.allData.erase(parm.allData.begin());
			parm.nDataBySet[hide_ns].erase(parm.nDataBySet[hide_ns].begin());
			// fit the data
			run_FIT(true);
			// return the hiden data point to the end of the whole and local lists
			parm.allData.push_back(hide_nd);
			parm.nDataBySet[hide_ns].push_back(hide_nd);
			// accumulate the results
			accum_SimResult(SIM_JACK);
			// report the results
			report_SimResult(rout, np+1);
		}
		// recover the global and local degree of freedom
		parm.setgDoF(parm.gDoF()+1);
		for (int ns = 0; ns < szSet; ns++) {
			parm.setlDoF(ns, parm.lDoF(ns)+1);
		}
		// calculate means and standard deviations
		calc_SimResult(iter, SIM_JACK);
	}
	rout.close();
}

void Fit::run_JACKlocal(ofstream &rout)
{
	// save the start time
	time(&start_t);
	// reduce the global degree of freedom by 1
	parm.setgDoF(parm.gDoF()-1);
	int szSet = parm.SETsize();
	int iter = parm.data.size();
	int np = 0;
	for (int ns = 0; ns < szSet; ns++) {
		// reduce the local degree of freedom by 1
		parm.setlDoF(ns, parm.lDoF(ns)-1);
		// remove the 1st data point from the local data lists
		int szd = parm.nDataBySet[ns].size();
		for (int nd = 0; nd < szd; nd++) {
			if (verb) {
				cerr << "\n# Jackknife simulation #" << ++np << "/" << iter << " ("
				     << parm.TITLE(ns) << " #" << nd+1 << "/" << szd << ")\n";
			}
			int hide_nd = parm.nDataBySet[ns][nd];
			parm.nDataBySet[ns].erase(parm.nDataBySet[ns].begin());
			// display the header of the fitting result
			display_chi_header();
			// loop over METHOD instructions
			int sz = method.size();
			for (int nm = 0; nm < sz; nm++) {
				switch (method.Type(nm)) {
				case METHOD_INIT:
					run_JACKlocalINIT(nm, ns);
					break;

				case METHOD_ONE:
					run_JACKlocalONE(nm, ns);
					break;

				case METHOD_ONEEX:
					run_JACKlocalONEEX(nm, ns);
					break;

				case METHOD_GRID:
					run_JACKlocalGRID(nm, ns);
					break;

				case METHOD_RANDOM:
					run_JACKlocalRANDOM(nm, ns, method.Iter(nm));
					break;

				case METHOD_MCMIN:
					run_JACKlocalMCMIN(nm, ns, method.Iter(nm), method.Scale(nm));
					break;

				case METHOD_RMCMIN:
					run_JACKlocalRMCMIN(nm, ns, method.Iter(nm), method.IterMC(nm), method.Scale(nm));
					break;
				}
			}
			calc_covar(pBEST);
			// accumulate the results
			accum_LocalJackResult(ns);
			// report the results
			report_localJackResult(rout, ns, nd+1);
			// return the hiden data point to the end of the local data list
			parm.nDataBySet[ns].push_back(hide_nd);
		}
		parm.setlDoF(ns, parm.lDoF(ns)+1);
	}
	calc_localJackResult();
	// recover the global degree of freedom
	parm.setgDoF(parm.gDoF()+1);
	time(&end_t);
}

// set the parameters to the starting values and evaluate them
void Fit::run_JACKlocalINIT(size_t num, size_t ns)
{
	// error estimation starts with the GBEST parameter set
	pBEST = pGBEST;
	calc_chi(pBEST);
	if (!std::isfinite(pBEST.chi)) {
		cerr << "WARNING: Bad initial parameters\n";
	}
	display_chi(num, "INIT", ns+1, 1, 1, pBEST.chi);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# INIT       X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using ONE
void Fit::run_JACKlocalONE(size_t num, size_t ns)
{
	run_localONE(num, ns);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# ONE          X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using ONEEX
void Fit::run_JACKlocalONEEX(size_t num, size_t ns)
{
	run_localONEEX(num, ns);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# ONEEX        X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using GRID
void Fit::run_JACKlocalGRID(size_t num, size_t ns)
{
	run_localGRID(num, ns);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# GRID        X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using RANDOM
void Fit::run_JACKlocalRANDOM(size_t num, size_t ns, int iter)
{
	run_localRANDOM(num, ns, iter);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# RANDOM      X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using MCMIN
void Fit::run_JACKlocalMCMIN(size_t num, size_t ns, int iter, double scale)
{
	run_localMCMIN(num, ns, iter, scale);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# MCMIN       X2/DoF = ", pBEST.chi/parm.gDoF());
}

// Jackknife calulation of a local fit using RMCMIN
void Fit::run_JACKlocalRMCMIN(size_t num, size_t ns, int iter, int iterMC, double scale)
{
	run_localRMCMIN(num, ns, iter, iterMC, scale);
	if (verb) { cerr << endl; }
	// store the result
	store_chi2("# RMCMIN      X2/DoF = ", pBEST.chi/parm.gDoF());
}

// accumulate results
void Fit::accum_LocalJackResult(size_t ns)
{
	// accumulate mean values of global chi2, local chi2 and each parameter
	pMEAN.chi  += pBEST.chi;
	pMEAN.lchi[ns]  += pBEST.lchi[ns];
	int sz = parm.nParBySet[ns].size();
	for (int n = 0; n < sz; n++) {
		int np = parm.nParBySet[ns][n];
		pMEAN.p[np]  += pBEST.p[np];
	}

	// accumulate standard deviations of global chi2, local chi2 and each parameter
	pSTDEV.chi += sqr(pBEST.chi - pGBEST.chi);
	pSTDEV.lchi[ns] += sqr(pBEST.lchi[ns] - pGBEST.lchi[ns]);
	for (int n = 0; n < sz; n++) {
		int np = parm.nParBySet[ns][n];
		pSTDEV.p[np] += sqr(pBEST.p[np] - pGBEST.p[np]);
	}
}

// report the results
void Fit::report_localJackResult(ofstream &rout, size_t ns, size_t nd)
{
	rout << "#" << nd << ": " << parm.TITLE(ns) << endl;
	int szp = parm.nParBySet[ns].size();
	for (int np = 0; np < szp; np++) {
		rout << "#" << nd << ":\t" << parm.ParNameBySet[ns][np] << "\t"
		     << pBEST.p[ parm.nParBySet[ns][np] ] << endl;
	}
	int szc = calc.size();
	for (int nc = 0; nc < szc; nc++) {
		double val = run_CALC(ns, nc, pBEST);
		rout << "#" << nd<< ":\t\t" << calc.Label(nc) << "\t" << val << endl;
		if (calc.getJackType()) {
			calc.addJack(ns, nc, val);
		} else {
			calc.addVal(ns, nc, val);
		}
	}
	rout << "#" << nd << ":\t" << "X2/DoF\t" << pBEST.lchi[ns] / parm.lDoF(ns)
	     << "\t(DoF = " << parm.lDoF(ns) << ")\n";

	rout << "#" << nd << ": GLOBAL X2/DoF\t" << pBEST.chi / parm.gDoF()
	     << "\t(DoF = " << parm.gDoF() << ")\n\n";
}

// calculate the results of the error estimation
void Fit::calc_localJackResult()
{
	int szSet = parm.SETsize();
	int szd = parm.data.size();
	double Jackfactor = static_cast<double>(szd - 1) / szd;
	// calculate the mean values and standard deviations
	// of global chi2, local chi2 and each parameter
	pMEAN.chi /= szd;
	pSTDEV.chi = sqrt(Jackfactor * pSTDEV.chi);
	for (int ns = 0; ns < szSet; ns++) {
		int sz = parm.nDataBySet[ns].size();
		Jackfactor = static_cast<double>(sz - 1) / sz;
		pMEAN.lchi[ns] /= sz;
		pSTDEV.lchi[ns] = sqrt(Jackfactor * pSTDEV.lchi[ns]);
		int szp = parm.nParBySet[ns].size();
		for (int n = 0; n < szp; n++) {
			int np = parm.nParBySet[ns][n];
			pMEAN.p[np] /= sz;
			pSTDEV.p[np] = sqrt(Jackfactor * pSTDEV.p[np]);
		}
	}
}

// Do Monte Carlo error calculations
void Fit::run_MONTE()
{
	const double BIG = 9.999e99;	// very big number
	// initialize parameter sets
	pMEAN.zero();
	pSTDEV.zero();
	calc.zero();
	// Monte Carlo chi2
	int szSet = parm.SETsize();
	mcX2.resize(szSet);
	// Save the original data
	vector<Data> data_org;		// coopy of original data
	vector<Data> data_synth;	// synthetic data
	data_org = data_synth = parm.data;

	// generate synthetic data from the best fit
	dvect dy0;
	int szd = parm.data.size();
	for (int nd = 0; nd < szd; nd++) {
		data_synth[nd].y = calc_y_dy(pGBEST.p, nd, dy0, bvect());
	}

	// open the output file
	ofstream rout;
	string filename = output + "-monte";
	rout.open(filename.c_str());
	check_file(rout, filename);
	rout << "Monte Carlo simulations (" << calc.getMonteIter() << ")\n";

	int iter = calc.getMonteIter();
	int szPar = parm.PAR_size();
	for (int i = 0; i < iter; i++) {
		if (verb) {
			cerr << "\n# Monte Carlo simulation #" << i+1 << "/" << iter << endl;
		}
		// Add noise to synthetic data
		for (int nd = 0; nd < szd; nd++) {
			// If y is the log of a measurement with a normal distribution,
			// this is essential to get the appropriate effect of the noise
			// -- the noise distribution will be skewed to the left.
			// Taylor expansion fails for noise > 1, where the skew becomes significant
			double noise = gaussrand() * parm.data[nd].dy;
			if (func.Get_dist() == LOG) {
				double tmp = noise * parm.data[nd].x[func.posTau()];
				if (tmp <= -1.0) {
//					cerr << "ERROR (MonteCarlo): Noise value out of bounds.\n"
//					     << "Check input file for large dy values.\n"
//					     << "Error occured for set "
//					     << parm.TITLE(parm.data[nd].ds) << endl;
//					return;
					--nd;
					continue;
				}
				noise = -log(1 + tmp) / parm.data[nd].x[func.posTau()];
			}
			parm.data[nd].y = data_synth[nd].y + noise;
		}
		// Fit the set
		run_FIT(true);
		// Check for nan
		bool isfinite = true;
		for (int np = 0; np < szPar; np++) {
			if (!std::isfinite(pBEST.p[np])) {
				isfinite = false;
				break;
			}
		}
		if (!std::isfinite(pBEST.chi)) { isfinite = false; }

		if (isfinite) {
			// accumulate the results
			accum_SimResult(SIM_MONTE);
			for (int ns = 0; ns < szSet; ns++) {
				push_sort(mcX2[ns], pBEST.lchi[ns]);
			}
			// Report the results
			report_SimResult(rout, i+1);
		} else {
			--iter;
			cerr << "# Simulation hit nan - not storing result\n";
			for (int ns = 0; ns < szSet; ns++) {
				push_sort(mcX2[ns], BIG);
			}
		}
	}
	// calculate means and standard deviations
	calc_SimResult(iter, SIM_MONTE);
	// restore the original data
	parm.data = data_org;

	rout.close();
}

// accumulate results
void Fit::accum_SimResult(int SIM)
{
	int szPar = parm.PAR_size();
	int szSet = parm.SETsize();
	// accumulate mean values of global chi2, local chi2 and each parameter
	pMEAN.chi  += pBEST.chi;
	for (int ns = 0; ns < szSet; ns++) {
		pMEAN.lchi[ns]  += pBEST.lchi[ns];
	}
	for (int np = 0; np < szPar; np++) {
		pMEAN.p[np]  += pBEST.p[np];
	}

	// accumulate standard deviations of global chi2, local chi2 and each parameter
	if (SIM == SIM_JACK) {
		pSTDEV.chi += sqr(pBEST.chi - pGBEST.chi);
		for (int ns = 0; ns < szSet; ns++) {
			pSTDEV.lchi[ns] += sqr(pBEST.lchi[ns] - pGBEST.lchi[ns]);
		}
		for (int np = 0; np < szPar; np++) {
			pSTDEV.p[np] += sqr(pBEST.p[np] - pGBEST.p[np]);
		}
	} else if (SIM == SIM_MONTE) {
		pSTDEV.chi += sqr(pBEST.chi);
		// accumulate the local chi2 values
		for (int ns = 0; ns < szSet; ns++) {
			pSTDEV.lchi[ns] += sqr(pBEST.lchi[ns]);
		}
		// accumulate the parameter values
		for (int np = 0; np < szPar; np++) {
			pSTDEV.p[np] += sqr(pBEST.p[np]);
		}
	}
}

// calculate the results of the error estimation
void Fit::calc_SimResult(int szSim, int SIM)
{
	int szPar = parm.PAR_size();
	int szSet = parm.SETsize();
	// calculate the mean values of global chi2, local chi2 and each parameter
	pMEAN.chi /= szSim;
	for (int ns = 0; ns < szSet; ns++) {
		pMEAN.lchi[ns] /= szSim;
	}
	for (int np = 0; np < szPar; np++) {
		pMEAN.p[np] /= szSim;
	}

	// calculate standard deviations of global chi2, local chi2 and each parameter
	if (SIM == SIM_JACK) {
		double Jackfactor = static_cast<double>(szSim - 1) / szSim;
		pSTDEV.chi = sqrt(Jackfactor * pSTDEV.chi);
		for (int ns = 0; ns < szSet; ns++) {
			pSTDEV.lchi[ns] = sqrt(Jackfactor * pSTDEV.lchi[ns]);
		}
		for (int np = 0; np < szPar; np++) {
			pSTDEV.p[np] = sqrt(Jackfactor * pSTDEV.p[np]);
		}
	} else if (SIM == SIM_MONTE) {
		pSTDEV.chi = compute_stdev(pMEAN.chi, pSTDEV.chi, szSim);
		for (int ns = 0; ns < szSet; ns++) {
			pSTDEV.lchi[ns] = compute_stdev(pMEAN.lchi[ns], pSTDEV.lchi[ns], szSim);
		}
		for (int np = 0; np < szPar; np++) {
			pSTDEV.p[np] = compute_stdev(pMEAN.p[np], pSTDEV.p[np], szSim);
		}
	}
}

// report the results
void Fit::report_SimResult(ofstream &rout, int iter)
{
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		rout << "#" << iter << ": " << parm.TITLE(ns) << endl;
		int szp = parm.nParBySet[ns].size();
		for (int np = 0; np < szp; np++) {
			rout << "#" << iter << ":\t" << parm.ParNameBySet[ns][np] << "\t"
			     << pBEST.p[ parm.nParBySet[ns][np] ] << endl;
		}
		int szc = calc.size();
		for (int nc = 0; nc < szc; nc++) {
			double val = run_CALC(ns, nc, pBEST);
			rout << "#" << iter << ":\t\t" << calc.Label(nc) << "\t" << val << endl;
			if (calc.getJackType()) {
				calc.addJack(ns, nc, val);
			} else {
				calc.addVal(ns, nc, val);
			}
		}
		rout << "#" << iter << ":\t" << "X2/DoF\t" << pBEST.lchi[ns] / parm.lDoF(ns)
		     << "\t(DoF = " << parm.lDoF(ns) << ")\n";
	}
	rout << "#" << iter << ": GLOBAL X2/DoF\t" << pBEST.chi / parm.gDoF()
	     << "\t(DoF = " << parm.gDoF() << ")\n\n";
}

// report the simulation results
void Fit::report_final(int SIM)
{
	ofstream rout;
	string filename;
	if (SIM == SIM_JACK) {
		// summary of jackknife error estimation
		filename = output + "-jack.sum";
		rout.open(filename.c_str());
		rout << "# Jackknife error analysis results\n";
		report_SimParm(rout, SIM_JACK);
		rout.close();
		// final result
		filename = output + "-jack-final.out";
		rout.open(filename.c_str());
		report_header(rout, method.size());
		rout << "# Error Analysis: JACK\n\n";
		report_FinalParm(rout, SIM);
		rout.close();
	} else if (SIM == SIM_MONTE) {
		// summary of MonteCarlo error estimation
		filename = output + "-monte.sum";
		rout.open(filename.c_str());
		rout << "# Monte Carlo error analysis results\n";
		report_SimParm(rout, SIM_MONTE);
		rout.close();
		// final result
		filename = output + "-monte-final.out";
		rout.open(filename.c_str());
		report_header(rout, method.size());
		rout << "# Error Analysis: MONTE " << calc.getMonteIter() << "\n\n";
		report_FinalParm(rout, SIM);
		rout.close();
	}
}

void Fit::store_chi2(const string &method, double rChi2)
{
	ostringstream method_chi2;
	method_chi2 << method << rChi2;
	report_chi2.push_back(method_chi2.str());
}

void Fit::report_SimParm(ofstream &rout, int SIM)
{
	double perc;			// Percentile
	double comp;			// Temporary variable
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		rout << "SET\t" << parm.TITLE(ns) << endl;
		int sz = parm.nParBySet[ns].size();
		for (int np = 0; np < sz; np++) {
			int n = parm.nParBySet[ns][np];
			rout << "\t" << parm.ParNameBySet[ns][np] << "\t"
			     << setw(7) << pMEAN.p[n];
			if (!parm.fixed[n]) {
				rout << " (" << pSTDEV.p[n] << ")";
			}
			rout << endl;
		}
		int szc = calc.size();
		for (int nc = 0; nc < szc; nc++) {
			rout << "\t\t" << calc.Label(nc) << "\t"
			     << calc.calcMean(ns, nc) << "\t(";
			if (SIM == SIM_JACK) {
				rout << calc.calcJackStdev(ns, nc) << ")\n";
			} else if (SIM == SIM_MONTE) {
				rout << calc.calcStdev(ns, nc) << ")\n";
			}
		}
		rout << "\tX2/DoF\t" << pMEAN.lchi[ns] / parm.lDoF(ns)
		     << " (" << pSTDEV.lchi[ns] / parm.lDoF(ns) << ") "
		     << "(DoF = " << parm.lDoF(ns) << ")\n\n";

		if (SIM == SIM_MONTE) {
			if (calc.getMonteIter() >= 10) {
				rout << "\t\tPercentile\tX2/DoF\n";
				for (perc = 0.05; perc < 1.0; perc += 0.05) {
					comp = pchi(ns, perc);
					rout << "\t\t" << perc << "\t"
					     << setw(7) << comp / parm.lDoF(ns);
					if (comp > pGBEST.lchi[ns]) {
						rout << "\t***";
					}
					rout << endl;
				}
			}
		}
	}
	if (SIM == SIM_JACK) { parm.setgDoF(parm.gDoF()-1); }
	rout << "Global X2/DoF = " << pMEAN.chi / parm.gDoF()
	     << "  (" << pSTDEV.chi / parm.gDoF() << ")\n";
	if (SIM == SIM_JACK) { parm.setgDoF(parm.gDoF()+1); }
}

void Fit::report_FinalParm(ofstream &rout, int SIM)
{
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		rout << "SET\t" << parm.TITLE(ns) << endl;
		int sz = parm.nParBySet[ns].size();
		for (int np = 0; np < sz; np++) {
			int n = parm.nParBySet[ns][np];
			rout << "\t" << parm.ParNameBySet[ns][np] << "\t"
			     << setw(7) << pGBEST.p[n];
			if (!parm.fixed[n]) {
				rout << " (" << pSTDEV.p[n] << ")";
			}
			rout << endl;
		}
		int szc = calc.size();
		for (int nc = 0; nc < szc; nc++) {
			rout << "\t    " << calc.Label(nc)  << "\t" << run_CALC(ns, nc, pGBEST) << "\t(";
			if (SIM == SIM_JACK) {
				rout << calc.calcJackStdev(ns, nc) << ")\n";
			} else if (SIM == SIM_MONTE) {
				rout << calc.calcStdev(ns, nc) << ")\n";
			}
		}
		rout << "\tX2/DoF\t" << pGBEST.lchi[ns] / parm.lDoF(ns)
		     << " (DoF = " << parm.lDoF(ns) << ")\n\n";
	}
	rout << "Global X2/DoF\t" << pGBEST.chi / parm.gDoF()
	     << " (DoF = " << parm.gDoF() << ")\n";
}

// Plot the results
void Fit::plot_xmgr(int nsim, const string &label)
{
	ofstream xout;
	int szSet = parm.SETsize();
	for (int ns = 0; ns < szSet; ns++) {
		string filename = parm.TITLE(ns) + ".xmgr";
		if (!nsim) {
			xout.open(filename.c_str());
			check_file(xout, filename);

			xmgr.xplot(xout, pGBEST, ns, parm , func);
			if (xmgr.isError()) {
				xmgr.annot_plot(xout, pGBEST, ns, parm, &stdev);
			} else {
				xmgr.annot_plot(xout, pGBEST, ns, parm);
			}
		} else {
			xout.open(filename.c_str(), std::ios::app | std::ios::out);
			check_file(xout, filename);
			xmgr.annot_sims(xout, pMEAN, pSTDEV, ns, parm, nsim, label);
		}
		xout.close();
	}
}

// calculate the standard deviation
double Fit::compute_stdev(double mean, double sumsq, int n)
{
	// Stdev < 0 is possible due to roundoff error
	double std = (sumsq - n * sqr(mean)) / (n-1);
	if (std > 0.0) { return sqrt(std); }
	else           { return 0.0;       }
}

// add a value to a sorted list
void Fit::push_sort(list<double> &MC_chi2, double v)
{
	list<double>::iterator p = MC_chi2.begin();
	while (p != MC_chi2.end() && (*p) < v) { ++p; }
	MC_chi2.insert(p, v);
}

// return the chi value of a requested percentile
double Fit::pchi(size_t ns, double perc)
{
	int n = int(mcX2[ns].size() * perc + 0.5) - 1;
	list<double>::iterator p = mcX2[ns].begin();
	for (int i = 0; i < n; i++) { ++p; }
	return *p;
}

double Fit::run_CALC(size_t ns, size_t nc, ParSet &pr)
{
	parm.Get_datamap(calc.getData(nc), ns);		// create map
	return calc.run(nc, pr.p, func.fx, parm.indexBySet[ns]);
}

void Fit::check_file(const ofstream &rout, const string &filename)
{
	if (!rout.is_open()) {
		cerr << "Error: Opening file \'" << filename << "\'\n";
		exit(1);
	}
}

bool Fit::is_local() const
{
	if (parm.allGPAR.size()) { return false; }
	return true;
}

bool Fit::is_global() const
{
	if (parm.allGPAR.size()) { return true; }
	return false;
}

double Fit::unirand()
{
	mt19937 engine(chrono::system_clock::now().time_since_epoch().count());
	return unidist(engine);
}

double Fit::gaussrand()
{
	mt19937 engine(chrono::system_clock::now().time_since_epoch().count());
	return gausdist(engine);
}
