// func_exp.cpp
// Kenji Sugase		2/21/2012

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Data.h"
#include "FuncLib.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;

// CONSTANT: f(x) = C
// par[0]	C	(unitless)	y intercept
double fconst(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_c = data.map[0];
	
	double C = par[0];
	if (dy.size()) {
		dy[i_c] = 1;
	}
	return C;
}

// LINEAR: f(x) = A + Bx
// par[0]	A	(unitless)	y intercept
// par[1]	B	(unitless)	slope
double fline(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a = data.map[0];
	int i_b = data.map[1];
	
	double x  = data.x[0];
	double A  = par[0];
	double Bx  = x * par[1];
	if (dy.size()) {
		dy[i_a] = 1;
		dy[i_b] = x;
	}
	return (A + Bx);
}

// EXP2: f(x) = Aexp(Bx)
// par[0]	A	(unitless)
// par[1]	B	(unitless)
double fexp2(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a = data.map[0];
	int i_b = data.map[1];
	
	double x  = data.x[0];
	double A  = par[0];
	double Bx = x * par[1];
	if (Bx < EXP_MIN) { return -HUGE_VAL; }
	if (Bx > EXP_MAX) { return  HUGE_VAL; }
	double expBx = exp(Bx);
	if (dy.size()) {
		dy[i_a] = expBx;
		dy[i_b] = A * x * expBx;
	}              
	return A * expBx;
}

// EXP3: f(x) = Aexp(Bx) + C
// par[0]	A	(unitless)
// par[1]	B	(unitless)
// par[2]	C	(unitless)
double fexp3(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a = data.map[0];
	int i_b = data.map[1];
	int i_c = data.map[2];
	
	double x, A, Bx, expBx, C;
	x  = data.x[0];
	A  = par[0];
	Bx = x * par[1];
	C  = par[2];
	if (Bx < EXP_MIN) { return -HUGE_VAL; }
	if (Bx > EXP_MAX) { return  HUGE_VAL; }
	expBx = exp(Bx);
	if (dy.size()) {
		dy[i_a] = exp(Bx);
		dy[i_b] = A * x * exp(Bx);
		dy[i_c] = 1;
	}
	return A * expBx + C;
}

// GAUSS: f(x) = Aexp(-Bx^2) + C
// par[0]	A	(unitless)
// par[1]	B	(unitless)
// par[2]	C	(unitless)
double fgauss(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a = data.map[0];
	int i_b = data.map[1];
	int i_c = data.map[2];
	
	double x2, A, Bx2, expBx2, C;
	x2  = sqr(data.x[0]);
	A   = par[0];
	Bx2 = -x2 * par[1];
	C  = par[2];
	if (Bx2 < EXP_MIN) { return -HUGE_VAL; }
	if (Bx2 > EXP_MAX)  { return  HUGE_VAL; }
	expBx2 = exp(Bx2);
	if (dy.size()) {
		dy[i_a] = expBx2;
		dy[i_b] = -A * x2 * expBx2;
		dy[i_c] = 1;
	}
	return A * expBx2 + C;
}

// GAUSS2: f(x) = Aexp(-Bx^2) + Cexp(-Dx^2)
// par[0]	A	(unitless)
// par[1]	B	(unitless)
// par[2]	C	(unitless)
// par[3]	D	(unitless)
double fgauss2(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a = data.map[0];
	int i_b = data.map[1];
	int i_c = data.map[2];
	int i_d = data.map[3];
	
	double x2, A, Bx2, C, Dx2, expBx2, expDx2;
	x2  = sqr(data.x[0]);
	A   = par[0];
	Bx2 = -x2 * par[1];
	C   = par[2];
	Dx2 = -x2 * par[3];
	if (Bx2 < EXP_MIN || Dx2 < EXP_MIN) { return -HUGE_VAL; }
	if (Bx2 > EXP_MAX || Dx2 > EXP_MAX) { return  HUGE_VAL; }
	expBx2 = exp(Bx2);
	expDx2 = exp(Dx2);
	if (dy.size()) {
		dy[i_a] = expBx2;
		dy[i_b] = -A * x2 * expBx2;
		dy[i_c] = expDx2;
		dy[i_d] = -C * x2 * expDx2;
	}
	return A * expBx2 + C * expDx2;
}


int Funclib::Read_Exp_Funcs(int idx)
{
	// Constant function
	idx = Define_Func(&fconst, "Constant", "f(x) = C", "EXP");
	Define_Parm(idx, "c"    ,-1.0e9, 1.0e9);

	// Linear function
	idx = Define_Func(&fline, "Linear", "f(x) = A + Bx", "EXP");
	Define_Parm(idx, "a"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "b"    ,-1.0e6, 1.0e6);

	// Exponential function (2 parameters)
	idx = Define_Func(&fexp2, "Exp2", "f(x) = Aexp(Bx)", "EXP");
	Define_Parm(idx, "a"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "b"    ,-1.0e6, 1.0e6);

	// Exponential function with a constant offset (3 parameters)
	idx = Define_Func(&fexp3, "Exp3", "f(x) = Aexp(Bx) + C", "EXP");
	Define_Parm(idx, "a"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "b"    ,-1.0e6, 1.0e6);
	Define_Parm(idx, "c"    ,-1.0e9, 1.0e9);

	// Simple Gaussian
	idx = Define_Func(&fgauss, "Gauss", "f(x) = Aexp(-Bx^2) + C", "EXP");
	Define_Parm(idx, "a"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "b"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "c"    ,-1.0e9, 1.0e9);
	
	// Sum of two Gaussians
	idx = Define_Func(&fgauss2, "Gauss2", "f(x) = Aexp(-Bx^2) + Cexp(-Dx^2)", "EXP");
	Define_Parm(idx, "a"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "b"    ,-1.0e6, 1.0e6);
	Define_Parm(idx, "c"    ,-1.0e9, 1.0e9);
	Define_Parm(idx, "d"    ,-1.0e6, 1.0e6);
	
	return idx;
}
