// func_R1rho.cpp
// Kenji Sugase		2015/12/7

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Data.h"
#include "FuncLib.h"
#include "LinearAlgebra.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;

// no Rex function for off-resonance R1rho dispersion
// par[0]	R20		(/s)
// par[0]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_const(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_R20 = data.map[0];
	const int i_R1  = data.map[1];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double offset = data.x[1]*2*M_PI;
	double R20    = par[0];
	double R1     = par[1];
	// calculate parameters
	double theta  = atan2(w1, offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	// calculate effective R1rho
	double R1rho = R1*cos2 + R20*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}
	return R1rho;
}

// no Rex function for on-resonance R1rho dispersion
// par[0]	R20		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_const_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_R20 = data.map[0];
	// copy parameters
	double R20   = par[0];
	// calculate effective R1rho
	double R1rho = R20;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_R20]) {
		dy[i_R20] = 1;
	}
	return R1rho;
}

// no Rex function for off-resonance R1rho dispersion
// par[0]	R20		(/s)
// par[0]	R1		(/s)
// data.x[0]	offset		(Hz)
// data.x[1]	w1/2pi		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_const_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_R20 = data.map[0];
	const int i_R1  = data.map[1];
	// copy parameters
	double offset = data.x[0]*2*M_PI;
	double w1     = data.x[1]*2*M_PI;
	double R20    = par[0];
	double R1     = par[1];
	// calculate parameters
	double theta  = atan2(w1, offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	// calculate effective R1rho
	double R1rho = R1*cos2 + R20*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}
	return R1rho;
}

// off-resonance R1rho dispersion
// J Magn Reson B104, 266-275 (1994)
// par[0]	phi=pApB*dw^2	(rad^2/s)
// par[1]	kex		(/s)
// par[2]	R20		(/s)
// par[3]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_London(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_phi = data.map[0];
	const int i_kex = data.map[1];
	const int i_R20 = data.map[2];
	const int i_R1  = data.map[3];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double offset = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double phi    = par[0] * sqr(field);
	double kex    = par[1];
	double R20    = par[2];
	double R1     = par[3];
	// calculate parameters
	double kex2   = sqr(kex);
	double we2    = sqr(w1) + sqr(offset);
	double theta  = atan2(w1, offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(kex2 + we2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + kex*phi*inorm)*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_phi]) {
		dy[i_phi] = sin2*sqr(field)*kex*inorm;
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = sin2*phi*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}

	return R1rho;
}

// on-resonance R1rho dispersion
// J Magn Reson B104, 266-275 (1994)
// par[0]	phi=pApB*dw^2	(rad^2/s)
// par[1]	kex		(/s)
// par[2]	R20		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_London_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_phi = data.map[0];
	const int i_kex = data.map[1];
	const int i_R20 = data.map[2];
	// copy parameters
	double w1    = data.x[0]*2*M_PI;
	double field = data.x[3];
	double phi   = par[0] * sqr(field);
	double kex   = par[1];
	double R20   = par[2];
	// calculate parameters
	double we2   = sqr(w1);
	double kex2  = sqr(kex);
	double inorm = 1/(kex2 + we2);
	// calculate effective R1rho
	double R1rho = R20 + kex*phi*inorm;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_phi]) {
		dy[i_phi] = sqr(field)*kex*inorm;
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = phi*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = 1;
	}
	return R1rho;
}

// off-resonance R1rho dispersion
// J Magn Reson B104, 266-275 (1994)
// par[0]	phi=pApB*dw^2	(rad^2/s)
// par[1]	kex		(/s)
// par[2]	R20		(/s)
// par[3]	R1		(/s)
// data.x[0]	offset		(Hz)
// data.x[1]	w1/2pi		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_London_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_phi = data.map[0];
	const int i_kex = data.map[1];
	const int i_R20 = data.map[2];
	const int i_R1  = data.map[3];
	// copy parameters
	double offset = data.x[0]*2*M_PI;
	double w1     = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double phi    = par[0] * sqr(field);
	double kex    = par[1];
	double R20    = par[2];
	double R1     = par[3];
	// calculate parameters
	double kex2   = sqr(kex);
	double we2    = sqr(w1) + sqr(offset);
	double theta  = atan2(w1, offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(kex2 + we2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + kex*phi*inorm)*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_phi]) {
		dy[i_phi] = sin2*sqr(field)*kex*inorm;
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = sin2*phi*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}

	return R1rho;
}

// off-resonance R1rho dispersion  
// J Am Chem Soc 127, 713-721 (2005)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_Kay(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw  = data.map[0];
	const int i_kex = data.map[1];
	const int i_pB  = data.map[2];
	const int i_R20 = data.map[3];
	const int i_R1  = data.map[4];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double offset = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double d_B    = pA*dwfld - offset;
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Be2 + kex2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + pB*dw2*kex*inorm)*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_dw]) {
		dy[i_dw] = 2*sin2*pB*field*dwfld*kex*inorm*(1 - d_B*pA*dwfld*inorm);
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = sin2*pB*dw2*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_pB]) {
		dy[i_pB] = sin2*dw2*kex*inorm;
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}
	return R1rho;
}

// on-resonance R1rho dispersion  
// J Am Chem Soc 127, 713-721 (2005)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_Kay_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw  = data.map[0];
	const int i_kex = data.map[1];
	const int i_pB  = data.map[2];
	const int i_R20 = data.map[3];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double d_B    = pA*dwfld;
	double w_Be2  = sqr(d_B) + sqr(w1);
	double inorm  = 1/(w_Be2 + kex2);
	// calculate effective R1rho
	double R1rho = R20 + pB*dw2*kex*inorm;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_dw]) {
		dy[i_dw] = 2*pB*field*dwfld*kex*inorm*(1 - d_B*pA*dwfld*inorm);
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = pB*dw2*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_pB]) {
		dy[i_pB] = dw2*kex*inorm;
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = 1;
	}
	return R1rho;
}

// off-resonance R1rho dispersion  
// J Am Chem Soc 127, 713-721 (2005)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	offset		(Hz)
// data.x[1]	w1/2pi		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double fR1rho_Kay_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw  = data.map[0];
	const int i_kex = data.map[1];
	const int i_pB  = data.map[2];
	const int i_R20 = data.map[3];
	const int i_R1  = data.map[4];
	// copy parameters
	double offset = data.x[0]*2*M_PI;
	double w1     = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double d_B    = pA*dwfld - offset;
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -offset);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Be2 + kex2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + pB*dw2*kex*inorm)*sin2;
	if (!dy.size()) {
		// no partial derivatives
		return R1rho;
	}
	// calculate partial derivatives
	if (!fixed[i_dw]) {
		dy[i_dw] = 2*sin2*pB*field*dwfld*kex*inorm*(1 - d_B*pA*dwfld*inorm);
	}
	if (!fixed[i_kex]) {
		dy[i_kex] = sin2*pB*dw2*inorm*(1 - 2*kex2*inorm);
	}
	if (!fixed[i_pB]) {
		dy[i_pB] = sin2*dw2*kex*inorm;
	}
	if (!fixed[i_R20]) {
		dy[i_R20] = sin2;
	}
	if (!fixed[i_R1]) {
		dy[i_R1] = cos2;
	}
	return R1rho;
}

// off-resonance R1rho dispersion  
// J Magn Reson 154, 157-160 (2002)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer02(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
//	const int i_R1  = data.map[4];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double offset = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double W_B    = pA*dwfld;
	double w_rf   = offset + W_A;
	double d_A    = W_A - w_rf;
	double d_B    = W_B - w_rf;
	double w_e2   = sqr(w_rf) + sqr(w1);
	double w_Ae2  = sqr(d_A) + sqr(w1);
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -w_rf);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + phi*kex*inorm)*sin2;
	return R1rho;
}

double fR1rho_Palmer02(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = -1;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer02, i_R20);
}

// on-resonance R1rho dispersion  
// J Magn Reson 154, 157-160 (2002)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer02_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double w_e2   = sqr(W_A) + sqr(w1);
	double w_Ae2  = sqr(w1);
	double w_Be2  = dw2 + sqr(w1);
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2);
	// calculate effective R1rho
	double R1rho = R20 + phi*kex*inorm;
	return R1rho;
}

double fR1rho_Palmer02_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 3;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer02_on, i_R20);
}

// off-resonance R1rho dispersion  
// J Magn Reson 154, 157-160 (2002)
// par[0]	dw		(rad^2/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	offset		(Hz)
// data.x[1]	w1/2pi		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer02_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
//	const int i_R1  = data.map[4];
	// copy parameters
	double offset = data.x[0]*2*M_PI;
	double w1     = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double W_B    = pA*dwfld;
	double w_rf   = offset + W_A;
	double d_A    = W_A - w_rf;
	double d_B    = W_B - w_rf;
	double w_e2   = sqr(w_rf) + sqr(w1);
	double w_Ae2  = sqr(d_A) + sqr(w1);
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -w_rf);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2);
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + phi*kex*inorm)*sin2;
	return R1rho;
}

double fR1rho_Palmer02_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = -1;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer02_off, i_R20);
}

// off-resonance R1rho dispersion
// J Magn Reson 177, 2212-227 (2005)
// par[0]	dw		(rad/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer05(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
//	const int i_R1  = data.map[4];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double offset = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double W_B    = pA*dwfld;
	double w_rf   = offset + W_A;
	double d_A    = W_A - w_rf;
	double d_B    = W_B - w_rf;
	double w_e2   = sqr(w_rf) + sqr(w1);
	double w_Ae2  = sqr(d_A) + sqr(w1);
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -w_rf);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2 - sin2*phi
			*(1 + 2*kex2*(pA*w_Ae2 + pB*w_Be2)/(w_Ae2*w_Be2 + w_e2*kex2)));
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + phi*kex*inorm)*sin2;
	return R1rho;
}

double fR1rho_Palmer05(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = -1;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer05, i_R20);
}

// on-resonance R1rho dispersion
// J Magn Reson 177, 2212-227 (2005)
// par[0]	dw		(rad/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// data.x[0]	w1/2pi		(Hz)
// data.x[1]	offset		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer05_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
	// copy parameters
	double w1     = data.x[0]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double w_e2   = sqr(W_A) + sqr(w1);
	double w_Ae2  = sqr(w1);
	double w_Be2  = dw2 + sqr(w1);
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2 - phi
			*(1 + 2*kex2*(pA*w_Ae2 + pB*w_Be2)/(w_Ae2*w_Be2 + w_e2*kex2)));
	// calculate effective R1rho
	double R1rho = R20 + phi*kex*inorm;
	return R1rho;
}

double fR1rho_Palmer05_on(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = -1;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer05_on, i_R20);
}

// off-resonance R1rho dispersion
// J Magn Reson 177, 2212-227 (2005)
// par[0]	dw		(rad/s)
// par[1]	kex		(/s)
// par[2]	pB		(unitless)
// par[3]	R20		(/s)
// par[4]	R1		(/s)
// data.x[0]	offset		(Hz)
// data.x[1]	w1/2pi		(Hz)
// data.x[2]	temperature	(K)
// data.x[3]	B0 Field / Base B0 Field
double R1rho_Palmer05_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
//	const int i_dw  = data.map[0];
//	const int i_kex = data.map[1];
//	const int i_pB  = data.map[2];
//	const int i_R20 = data.map[3];
//	const int i_R1  = data.map[4];
	// copy parameters
	double offset = data.x[0]*2*M_PI;
	double w1     = data.x[1]*2*M_PI;
	double field  = data.x[3];
	double dw     = par[0];
	double kex    = par[1];
	double pB     = par[2];
	double R20    = par[3];
	double R1     = par[4];
	// calculate parameters
	double dwfld  = dw*field;
	double dw2    = sqr(dwfld);
	double kex2   = sqr(kex);
	double pA     = 1 - pB;
	double phi    = pA*pB*dw2;
	double W_A    = -pB*dwfld;
	double W_B    = pA*dwfld;
	double w_rf   = offset + W_A;
	double d_A    = W_A - w_rf;
	double d_B    = W_B - w_rf;
	double w_e2   = sqr(w_rf) + sqr(w1);
	double w_Ae2  = sqr(d_A) + sqr(w1);
	double w_Be2  = sqr(d_B) + sqr(w1);
	double theta  = atan2(w1, -w_rf);
	double sin2   = sqr(sin(theta));
	double cos2   = 1 - sin2;
	double inorm  = 1/(w_Ae2*w_Be2/w_e2 + kex2 - sin2*phi
			*(1 + 2*kex2*(pA*w_Ae2 + pB*w_Be2)/(w_Ae2*w_Be2 + w_e2*kex2)));
	// calculate effective R1rho
	double R1rho = R1*cos2 + (R20 + phi*kex*inorm)*sin2;
	return R1rho;
}

double fR1rho_Palmer05_off(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = -1;
	return calc_deriv(data, par, dy, fixed, R1rho_Palmer05_off, i_R20);
}

int Funclib::Read_R1RHO_Funcs(int idx)
{
	// index0: temperature
	// index1: magnetic field
	// index2: offset

	idx = Define_Func(&fR1rho_const, "R1RHO_CONST"
		, "f(R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "R20" ,  0.01,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_const_on, "R1RHO_CONST_ON"
		, "f(R20)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "R20" ,  0.01,   100, "110");

	// index2: w1
	idx = Define_Func(&fR1rho_const_off, "R1RHO_CONST_OFF"
		, "f(R20,R1)", "R1rho_off", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "R20" ,  0.01,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_London, "R1RHO_LONDON"
		, "f(phi,kex,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "phi" ,  1e-5,  1e+8, "100");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	// index2: w1
	idx = Define_Func(&fR1rho_London_off, "R1RHO_LONDON_OFF"
		, "f(phi,kex,R20,R1)", "R1rho_off", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "phi" ,  1e-5,  1e+8, "100");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_London_on, "R1RHO_LONDON_ON"
		, "f(phi,kex,R20)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "phi" ,  1e-5,  1e+9, "100");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "R20" ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Kay, "R1RHO_KAY"
		, "f(dw,kex,pB,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,   0.1, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Kay_on, "R1RHO_KAY_ON"
		, "f(dw,kex,pB,R20)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,   0.1, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");

	// index2: w1
	idx = Define_Func(&fR1rho_Kay_off, "R1RHO_KAY_OFF"
		, "f(dw,kex,pB,R20,R1)", "R1rho_off", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,   0.1, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Palmer02, "R1RHO_PALMER02"
		, "f(dw,kex,pB,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,  0.49, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Palmer02_on, "R1RHO_PALMER02_ON"
		, "f(dw,kex,pB,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,  0.49, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");

	// index2: w1
	idx = Define_Func(&fR1rho_Palmer02_off, "R1RHO_PALMER02_OFF"
		, "f(dw,kex,pB,R20,R1)", "R1rho_off", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001,  0.49, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Palmer05, "R1RHO_PALMER05"
		, "f(dw,kex,pB,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001, 0.499, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	idx = Define_Func(&fR1rho_Palmer05_on, "R1RHO_PALMER05_ON"
		, "f(dw,kex,pB,R20,R1)", "R1rho", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001, 0.499, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");

	// index2: w1
	idx = Define_Func(&fR1rho_Palmer05_off, "R1RHO_PALMER05_OFF"
		, "f(dw,kex,pB,R20,R1)", "R1rho_off", XDIM4, IDIM3, GAUSS);
	Define_Parm(idx, "dw"  ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 80000, "100");
	Define_Parm(idx, "pB"  , 0.001, 0.499, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "110");
	Define_Parm(idx, "R1"  ,  0.01,   100, "110");

	return idx;
}
