// func_exp.cpp
// Kenji Sugase		2/21/2012

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Data.h"
#include "FuncLib.h"
#include "Type.h"
#include "Mathlib.h"


using namespace std;

// SIG: f(x) = (bn+bd*exp(dh/r)(1/tm - 1/x))/(1+exp(dh/r)(1/tm-1/x)) 
// par[0]	bd	(unitless)
// par[1]	bn	(unitless)
// par[2]	tm	(kelvin)
// par[3]	dh	(joules/mole)
double fsig(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_bd = data.map[0];
	int i_bn = data.map[1];
	int i_tm = data.map[2];
	int i_dh = data.map[3];
	
	double v, bd, bn, dh, tm;
        double denom, expon;
        double numh, numa, du, A, r, f;
        const double TINY=1.0e-30;

	v   = data.x[0] + TINY;
        f = 1.0 / v;
        r = 1.987;

	bd  = par[0];
	bn  = par[1];
	tm  = par[2];
	dh  = par[3];

        expon = exp(dh/(r*tm)-dh*f/r);
        denom = 1 + expon;
        A = (1/(tm*r)) - f/r;

	if (v < EXP_MIN) { return -HUGE_VAL; }
	if (v > EXP_MAX) { return  HUGE_VAL; }

	if (dy.size()) {

		dy[i_bn] = 1/(denom + TINY);
		dy[i_bd] = expon/(denom + TINY);

                numh = bd*A*expon*(denom) - A*expon*(bn+bd*expon);
		dy[i_dh] = numh/(denom*denom);

 	        du= (-1/(tm*tm));
                numa = bd*dh*du*expon*(denom)-bd*du*expon*(bn+bd*expon);
		dy[i_tm] = numa/(denom*denom) + TINY;
	}
	//return (bn + bd * exp(dh/tm) * exp(-1/v)) / (1 + bd * exp(dh/tm)  * exp(-1/v)) ;
	return ((bn + bd*expon)/denom) ;
}

// SIG_Boltz: f(x) = (a2 - (a1-a1)/[1+exp(x-tm/m)]) 
// par[0]	a1	(unitless)
// par[1]	a2	(unitless)
// par[2]	tm	(kelvin)
// par[3]	m 	(slope factor)
double fsig_boltz(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int i_a1 = data.map[0];
	int i_a2 = data.map[1];
	int i_tm = data.map[2];
	int i_m  = data.map[3];
	
	double v, a1, a2, tm, m;
        double U, V, V2, G, m2, expG, tmx ;
        const double TINY=1.0e-30;

	v   = data.x[0] + TINY;

	a1  = par[0];
	a2  = par[1];
	tm  = par[2];
	m   = par[3];

	G = (v - tm)/m ;
	expG = exp(G);
	U = a1 - a2 ;
        V = 1 + expG;
        V2 = V*V; 
	m2 = m*m;
        tmx = (v-tm)/m2;

	if (V < EXP_MIN) { return -HUGE_VAL; }
	if (V > EXP_MAX) { return  HUGE_VAL; }

	if (dy.size()) {

		dy[i_a1] = 1/V ;
		dy[i_a2] = 1 - (1/V) ;

		dy[i_tm]  = U*expG*(1/m)/V2 ;
		dy[i_m]  = U*expG*tmx/V2 ;

	}

	return ( a2 + (U/V) ) ;
}


int Funclib::Read_Sig_Funcs(int idx)
{

	// Sig function 
	//idx = Define_Func(&fsig, "Sig", "f(x) = (bn+bd*exp(dh/r)(1/tm - 1/x))/(1+exp(dh/r)(1/tm-1/x))", "SIG");
	idx = Define_Func(&fsig, "Sig", "f(x) = (bn+bd*exp(dh/R*tm - dh/R*x))/(1+exp(dh/R*tm-dh/R*x))", "SIG");
	Define_Parm(idx, "bd"    ,-10, 0);
	Define_Parm(idx, "bn"    , 0, 10);
	Define_Parm(idx, "tm"    , 1, 500);
	Define_Parm(idx, "dh"    , 1, 5.0e5);

	// Sig_BOLTZ function 
	//idx = Define_Func(&fsig, "Sig", "f(x) = (bn+bd*exp(dh/r)(1/tm - 1/x))/(1+exp(dh/r)(1/tm-1/x))", "SIG");
	idx = Define_Func(&fsig_boltz, "Sig_boltz", "f(x) = a2 + (a1-a2)/[1+exp(x-tm/m)]", "SIG_BOLTZ");
	Define_Parm(idx, "a1"    , 0, 1);
	Define_Parm(idx, "a2"    ,-1, 0);
	Define_Parm(idx, "tm"    , 1, 300);
	Define_Parm(idx, "m"     , 1, 100);
	
	return idx;
}
