// Fit class
// Kenji Sugase		4/1/2013

#ifndef __FIT_H__
#define __FIT_H__

#include <list>
#include <random>
#include <chrono>
#include "FuncLib.h"
#include "Calc.h"
#include "Method.h"
#include "Xmgr.h"
#include "Data.h"
#include "LinearAlgebra.h"
#include "ParDef.h"
#include "ParSet.h"
#include "Type.h"

struct FitMat
{
	int szMat;			// matrix size used for fitting
	ivect pidx;			// parameter numbers
	ivect inv;			// inverse index

	void resize(size_t sz)
	{
		pidx.resize(sz);
		inv.resize(sz);
	}

	// Reindex columns so that fixed parameters show up last. This allows for
	// diagonalization of a smaller matrix (mfit x mfit instead of szPAR x szPAR)
	void reindex(const bvect &pmask)
	{
		int mfit = 0;		// size of reduced matrix
		int sz = pmask.size();
		for (int np = 0; np < sz; np++) {
			if (pmask[np]) {
				inv[np] = mfit;
				pidx[mfit++] = np;
			} else {
				inv[np] = sz + mfit - np - 1;
				pidx[inv[np]] = np;
			}
		}
		szMat = mfit;
	}

	int size() const
	{
		return szMat;
	}

};

struct LevMar
{
	d2vect alpha;
	dvect beta;
	void resize(int sz)
	{
		alpha.resize(sz);
		beta.resize(sz);
		for (int i = 0; i < sz; i++) { alpha[i].resize(sz); }
	}

	void zero_alpha()
	{
		int sz = alpha.size();
		for (int i = 0; i < sz; i++) {
			fill(alpha[i].begin(), alpha[i].end(), 0.0);
		}
	}

	void zero_beta()
	{
		fill(beta.begin(), beta.end(), 0.0);
	}

	void zero()
	{
		zero_alpha();
		zero_beta();
	}
};

class Fit
{
private:
	const string &VERSION;
	Funclib &func;
	Method &method;
	ParDef &parm;
	Calc &calc;
	Xmgr &xmgr;
	const string output;
	int verb;
	int prec;

//	mt19937 engine;
	uniform_real_distribution<double> unidist;
	normal_distribution<double> gausdist;

	ParSet pTEMP;
	ParSet pBEST;
	ParSet pGBEST;
	ParSet pMEAN;
	ParSet pSTDEV;
	double *ppar;			// pointer for parameters used for calc_y_dy

	LevMar marq;			// LevMar object
	LevMar lmarq;			// local LevMar object
	LevMar marq0;			// temporal LevMar object
	dvect da;
	d2vect covar;			// Covariance matrix
	dvect dyda;
	dvect stdev;			// Standard deviation / PAR
	vector<FitMat> lMat;
	FitMat gMat;

	vector<list<double> > mcX2;	// monte Carlo chi2

	time_t start_t;			// start time
	time_t end_t;			// end time
	svect report_chi2;
	typedef double (Fit::*evalsub)(ParSet &, const ivect &, const ivect &, const FitMat &, LevMar &);

	void make_dir(int num, const string &method);
	void run_INIT(size_t num, bool sim = false);
	void run_ONE(size_t num);
	void run_localONE(size_t num, size_t ns);
	void run_globalONE(size_t num);
	void run_ONEEX(size_t num);
	void run_localONEEX(size_t num, size_t ns);
	void run_globalONEEX(size_t num);
	void run_GRID(size_t num);
	void run_localGRID(size_t num, size_t ns);
	void run_globalGRID(size_t num);
	void run_RANDOM(size_t num, int iter);
	void run_localRANDOM(size_t num, size_t ns, int iter);
	void run_globalRANDOM(size_t num, int iter);
	void run_MCMIN(size_t num, int iter, double scale);
	void run_localMCMIN(size_t num, size_t ns, int iter, double scale);
	void run_globalMCMIN(size_t num, int iter, double scale);
	void run_RMCMIN(size_t num, int iter, int iterMC, double scale);
	void run_localRMCMIN(size_t num, size_t ns, int iter, int iterMC, double scale);
	void run_globalRMCMIN(size_t num, int iter, int iterMC, double scale);
	double minimize_local(ParSet &pr, size_t ns);
	double minimize_global(ParSet &pr);
	double LevMarq(ParSet &pr, ParSet &pTRY, const ivect &dlist, const ivect &plist
		, const FitMat &mat, LevMar &levmar, evalsub feval);
	double lMarq(ParSet &pr, const ivect &dlist, const ivect &plist, const FitMat &mat
		, LevMar &levmar);
	double gMarq(ParSet &pr, const ivect &dlist, const ivect &plist, const FitMat &mat
		, LevMar &levmar);
	void run_JACKlocal(ofstream &rout);
	void run_JACKlocalINIT(size_t num, size_t ns);
	void run_JACKlocalONE(size_t num, size_t ns);
	void run_JACKlocalONEEX(size_t num, size_t ns);
	void run_JACKlocalGRID(size_t num, size_t ns);
	void run_JACKlocalRANDOM(size_t num, size_t ns, int iter);
	void run_JACKlocalMCMIN(size_t num, size_t ns, int iter, double scale);
	void run_JACKlocalRMCMIN(size_t num, size_t ns, int iter, int iterMC, double scale);
	double set_grid(size_t i, int grid_pt);
	void rand_parm(size_t np, ParSet &pr);
	void mc_parm(size_t np, ParSet &pr, double scale);
	void display_chi_header();
	void display_chi(int number, const char *method, int nset, int counter, int total
		, double chi_best, double chi0 = 0, double chi = 0);
	void calc_covar(const ParSet &pr);
	void check_parm_range(ParSet &pr);
	void check_parm(const ParSet &pr);
	void check_chi(const ParSet &pr) const;
	double calc_chi(ParSet &pr, size_t ns);
	double calc_chi(ParSet &pr);
	double calc_y_dy(const dvect &p, int nd, dvect &dy, const bvect &fixed);
	void store_chi2(const string &method, double rChi2);
	void report_header(ofstream &rout, int num);
	void report_time(ofstream &rout);
	void report_FitParm(ofstream &rout);
	void report_SimParm(ofstream &rout, int SIM);
	void report_FinalParm(ofstream &rout, int SIM);
	void accum_SimResult(int SIM);
	void accum_LocalJackResult(size_t ns);
	void report_SimResult(ofstream &rout, int iter);
	void report_localJackResult(ofstream &rout, size_t ns, size_t nd);
	void calc_SimResult(int szSim, int SIM);
	void calc_localJackResult();
	double compute_stdev(double mean, double sumsq, int ni);
	void push_sort(list<double> &MC_chi2, double v);
	double pchi(size_t ns, double perc);
	double run_CALC(size_t ns, size_t nc, ParSet &pr);
	void check_file(const ofstream &rout, const string &filename);
	bool is_local() const;
	bool is_global() const;
	double unirand();
	double gaussrand();

public:
	Fit(const string &VERSION_in, Funclib &func_in, Method &method_in, ParDef &parm_in
		, Calc &calc_in, Xmgr &xmgr_in, const string &output_in, int verb_in, int prec_in);
	void run_FIT(bool sim = false);
	void report_fit(int num);
	void plot_xmgr(int nsim = 0, const string &label ="");
	void run_JACK();
	void run_MONTE();
	void report_final(int SIM);
};

#endif
