// Funclib class
//	Library of functions for use in data fitting
// Kenji Sugase			11/25/2013

#ifndef __FUNCLIB_H__
#define __FUNCLIB_H__

#include <iostream>
#include <string>
#include <cfloat>
#include <cmath>
#include <omp.h>
#include "Data.h"
#include "Rule.h"
#include "Type.h"
#include "Alloc.h"

namespace{
	const int GAUSS = 0;
	const int LOG   = 1;
	typedef double (*pFunc)(const Data &data, double *par, dvect &dy, const bvect &fixed);

	const double LOW_ITCP = 10;			// lower limit for 1/tcp
	const double TINY     = 1e-30;			// small offset to prevent division by zero
	const double DELTA    = 1e-9;			// small value for approximated derivatives

	// Finite difference approximation of derivatives
	// If analytical partial deriatives cannot be obtained or unstable, use this function
	double calc_deriv(const Data &data, double *par, dvect &dyda, const bvect &fixed, pFunc fx, int skip)
	{
		dvect fakedyda;
		double ret = (*fx)(data, par, fakedyda, fixed);
		if (!dyda.size()) { return ret; }
		// copy parameter set
		// finite difference approximation of derivateves
		int sz = data.map.size();
		#ifdef _OPENMP
		#pragma omp parallel for firstprivate(par)
		#endif
		for (int i = 0; i < sz; i++) {
			int np = data.map[i];
			if (!fixed[np]) {
				if (i == skip) {
					dyda[np] = 1.0;
				} else {
					double delta_perc = abs(par[i]) * DELTA;
					par[i] += delta_perc;
					dyda[np] = ((*fx)(data, par, dyda, fixed) - ret) / delta_perc;
					par[i] -= delta_perc;
				}
			}
		}
		return ret;
	}
}

class Funclib {
private:
	static const int XDIM1 = 1;
	static const int XDIM2 = 2;
	static const int XDIM3 = 3;
	static const int XDIM4 = 4;
	static const int XDIM5 = 5;
	static const int XDIM6 = 6;
	static const int IDIM1 = 1;
	static const int IDIM2 = 2;
	static const int IDIM3 = 3;
	static const int IDIM4 = 4;

	int num;			// function number
	vector<pFunc> pFuncs;		// pointer to the function
	svect func_names;		// function names
	svect func_eqns;		// description of the function
	svect xmgr_headers;		// header of the xmgr file
	ivect func_xdims;		// number of independent variables
	ivect func_idims;		// number of index for experimental conditions
	ivect func_rnd_dists;		// distribution type flag (GAUSS or LOG)
	ivect func_rnd_tau;		// used by log dist calcs
	vector<Rules> func_rules;	// parameter rules

	int Read_Sig_Funcs(int idx);
	int Read_Exp_Funcs(int idx);
	int Read_Bio_Funcs(int idx);
	int Read_CPMG2_Funcs(int idx);
	int Read_CPMG3_Funcs(int idx);
	int Read_MQCPMG_Funcs(int idx);
	int Read_R1RHO_Funcs(int idx);
	int Read_CLEANEX_Funcs(int idx);
	void ERROR_invalid_func(const string &name);

public:
	pFunc fx;

	Funclib();
	~Funclib();

	int Define_Func(pFunc pfx, const char *name, const char *desc, const char *xmgr
		, int xdim = 1, int idim = 0, int dist = GAUSS);
	void Define_Parm(int num, const char *name, double min, double max, const char *index = "");
	void Pick_Func(const string &name);
	const Rules & Get_Rules() const;
	const string & Name() const;
	const string & Eqn() const;
	const string & XmgrHeader() const;
	int Get_dimX() const;
	int Get_dimIndex() const;
	int Get_dist() const;
	int posTau() const;
};

#endif
