// func_cpmg3.cpp
// R2 dispersion experiment, 3-state exchange
// Kenji Sugase		12/16/2013

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <complex>
#include "Data.h"
#include "FuncLib.h"
#include "LinearAlgebra.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;

// prepare the relaxation matrix
void prep_R2mat(double *M0, double *D, double *U, double *Ui, double dwab, double dwac
	, double kab, double kba, double kbc, double kcb, double kca, double kac
	, double R20a = 0, double R20b = 0, double R20c = 0)
{
	double denom = 1 / ((kba + kbc)*(kac + kca) + (kac + kba)*kcb + kab*(kbc + kca + kcb));
	M0[0] = (kbc * kca + kba * (kca + kcb)) * denom;	// pa
	M0[1] = (kac * kcb + kab * (kca + kcb)) * denom;	// pb
	M0[2] = 1 - M0[0] - M0[1];				// pc
	// construct Kubo-Sack matrix R
	double R[18];
	R[0]  = R20a + kab + kac;	R[1]  = 0.0;		// (1, 1)
	R[2]  =       -kab      ;	R[3]  = 0.0;		// (2, 1)
	R[4]  =             -kac;	R[5]  = 0.0;		// (3, 1)

	R[6]  =       -kba      ;	R[7]  = 0.0;		// (1, 2)
	R[8]  = R20b + kba + kbc;	R[9]  = -dwab;		// (2, 2)
	R[10] =             -kbc;	R[11] = 0.0;		// (3, 2)

	R[12] =             -kca;	R[13] = 0.0;		// (1, 3)
	R[14] =       -kcb      ;	R[15] = 0.0;		// (2, 3)
	R[16] = R20c + kcb + kca;	R[17] = -dwac;		// (3, 3)
	// get eigenvectors and eigenvalues of R
	if (zmat_eig(3, R, U, Ui, D)) {
		cerr << "ERROR: Relaxation matrix is singular.";
		exit(1);
	}
}

// calculate MQ R2
double calc_R2(const Data &data, double dwab, double dwac, double dwab2, double dwac2, double kab
	, double kba, double kbc, double kcb, double kca, double kac, double R20)
{
	double M0[3];
	double denom = 1 / ((kba + kbc)*(kac + kca) + (kac + kba)*kcb + kab*(kbc + kca + kcb));
	M0[0] = (kbc * kca + kba * (kca + kcb)) * denom;	// pa
	M0[1] = (kac * kcb + kab * (kca + kcb)) * denom;	// pb
	M0[2] = 1 - M0[0] - M0[1];				// pc
	// Construct Kubo-Sack matrix R
	double R[18], DD[6], UD[18], iUD[18], DZ[6], UZ[18], iUZ[18];
	R[0]  = -R20 - kab - kac;	R[1]  = 0.0;		// (1, 1)
	R[2]  =        kab      ;	R[3]  = 0.0;		// (2, 1)
	R[4]  =              kac;	R[5]  = 0.0;		// (3, 1)

	R[6]  =        kba      ;	R[7]  = 0.0;		// (1, 2)
	R[8]  = -R20 - kba - kbc;	R[9]  = dwab2 + dwab;	// (2, 2)
	R[10] =              kbc;	R[11] = 0.0;		// (3, 2)

	R[12] =              kca;	R[13] = 0.0;		// (1, 3)
	R[14] =        kcb      ;	R[15] = 0.0;		// (2, 3)
	R[16] = -R20 - kcb - kca;	R[17] = dwac2 + dwac;	// (3, 3)
	// Get eigenvectors and eigenvalues of R
	zmat_eig(3, R, UD, iUD, DD);
	// Construct relaxation matrix of Z+
	R[9]  = dwab2 - dwab;	// (2,2)
	R[17] = dwac2 - dwac;	// (3,3)
	// Get eigenvectors and eigenvalues of R
	zmat_eig(3, R, UZ, iUZ, DZ);
	double DtD[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};		// exp(-D*t)
	double DtZ[] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};		// exp(-D*t)
	double eDpt[18], eZpt[18], eDmt[18], eZmt[18];
	double A[18], B[18], IDt[6], IZt[6];
	double itcp  = data.x[0];		// 1/tcp
	double Tcpmg = data.x[3];
	double t = 0.5/itcp;			// tcp/2
	int n = 0.25*Tcpmg*itcp;
	// Raise eigenvectors to a power
	zmat_exp(3, t, DD, DtD);
	zmat_mlt(3, UD, DtD, iUD, eDpt);
	zmat_cnj(3, eDpt, eDmt);

	zmat_exp(3, t, DZ, DtZ);
	zmat_mlt(3, UZ, DtZ, iUZ, eZpt);
	zmat_cnj(3, eZpt, eZmt);

	// Set R = A * cA * cA * A
	zmat_mltpow(3, eDpt, eZmt, eZmt, eDpt, n, A);
	zmat_mltpow(3, eZpt, eDmt, eDmt, eZpt, n, B);

	zmat_vmlt(3, A, B, M0, IDt);
	zmat_vmlt(3, B, A, M0, IZt);

	return log(M0[0]/(IDt[0]+IZt[0]))/Tcpmg;
}

double calc_R2(double itcp, const double *M0, const double *D, double *U, const double *Ui)
{
	double half_tcp, pa, pb, pc, pa2, pb2;
	double b11r, b12r, b12i, b13r, b13i, b22r, b23r, b23i, b33r;
	double b11r2, b12r2, b12i2, b13r2, b13i2, b22r2, b23r2, b23i2, b33r2;
	double B12p, B12m, B13p, B13m, B23p, B23m, M1, M2, M3, P, Q, S, T;
	double Dt[6], A[18], cA[18], tmp[18];
	half_tcp = -0.5 / itcp;		// -tcp/2
	zmat_exp(3, half_tcp, D, Dt);	// exp(-D*tcp/2)
	zdmat_mlt(3, U, Dt, Ui, A);	// U*Dt*U(-1)
	zmat_cnj(3, A, cA);
	zmat_mlt(3, A, cA, tmp);		// A*cA

	pa    = M0[0];		pa2   = sqr(pa);
	pb    = M0[1];		pb2   = sqr(pb);
	pc    = M0[2];

	b11r  = tmp[0];		b11r2 = sqr(b11r);
	b12r  = tmp[6];		b12r2 = sqr(b12r);	b12i  = tmp[7];		b12i2 = sqr(b12i);
	b13r  = tmp[12];	b13r2 = sqr(b13r);	b13i  = tmp[13];	b13i2 = sqr(b13i);
	b22r  = tmp[8];		b22r2 = sqr(b22r);
	b23r  = tmp[14];	b23r2 = sqr(b23r);	b23i  = tmp[15];	b23i2 = sqr(b23i);
	b33r  = tmp[16];	b33r2 = sqr(b33r);

	B12p = b12i2 + b12r2;	B12m = b12i2 - b12r2;
	B13p = b13i2 + b13r2;	B13m = b13i2 - b13r2;
	B23p = b23i2 + b23r2;	B23m = b23i2 - b23r2;

	M1 = (-b11r2 - b22r2 - b33r2 + 2 * ((pb*B12m + pc*B13m)/pa + pc/pb*B23m))/3;

	M2 = b11r2*b22r2 + b11r2*b33r2 + b22r2*b33r2 + (sqr(pb2*B12p) +sqr(pb*pc*B13p) +sqr(pa*pc*B23p)
		+2*pb*(pb2*(pc*(B12m*B13m +4*b12i*b12r*b13i*b13r) -pa*(b33r2*B12m +b11r*b22r*B12p))
		+pa*pc*(pc*(B13m*B23m +4*b13i*b13r*b23i*b23r) -pa*(b11r2*B23m +b22r*b33r*B23p)
		-pb*(b22r2*B13m +b11r*b33r*B13p -B12m*B23m +4*b12i*b12r*b23i*b23r
		+2*(b11r*(b13r*(b12r*b23r +b12i*b23i) -b13i*(b12r*b23i-b12i*b23r))
		-(b22r-b33r)*b23i*(b12r*b13i +b12i*b13r)
		+(b22r+b33r)*b23r*(b12r*b13r -b12i*b13i))))))/(pa2*pb2);

	M3 = -(sqr(b33r*pb*(b11r*b22r*pa -B12p*pb) -(b11r*B23p*pa+B13p*b22r*pb
		-2*(b12r*(b13i*b23i +b13r*b23r) -b12i*(b13r*b23i -b13i*b23r))*pb)*pc))/(pa2*pb2);

	P = -sqr(M1) + M2/3;
	Q = 2*cube(M1) - M1*M2 + M3;
	S = 0.5*Q/(P*sqrt(-P));
	T = acos(S);

	if (!std::isfinite(T)){
		if (S > 0) { return -0.5 * itcp * log( 2 * sqrt(-P) - M1); }	// cos(T/3) = 1
		else       { return -0.5 * itcp * log(-2 * sqrt(-P) - M1); }	// cos(T/3) = -1
	}
	return -0.5 * itcp * log(2 * sqrt(-P) * cos(T/3) - M1);
}

double cpmg_matrix3(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double field = data.x[3];
	double dwab  = par[0];
	double dwac  = par[1];
	double kab   = par[2];
	double kba   = par[3];
	double kbc   = par[4];
	double kca   = par[5];
	double kac   = par[6];
	double R20   = par[7];
	// Calculate derived parameters
	double kcb   = (kab * kbc * kca) / (kba * kac);
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_matrix3(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 7;
	return calc_deriv(data, par, dy, fixed, cpmg_matrix3, i_R20);
}

double cpmg_temp3(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const double k_h = 2.0836644e10;		// Boltzman's constant / Planck's constant
	const double R   = 1.985878e-3;			// Gas constant (kcal/mol/K)
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double T     = data.x[1];
	double field = data.x[2];
	double dwab  = par[0];
	double dwac  = par[1];
	double dHab  = par[2];
	double dHba  = par[3];
	double dHbc  = par[4];
	double dHcb  = par[5];
	double dHca  = par[6];
	double dHac  = par[7];
	double dSab  = par[8];
	double dSba  = par[9];
	double dSbc  = par[10];
	double dScb  = par[11];
	double dSca  = par[12];
	double dSac  = par[13];
	double R20   = par[14];
	// Calculate derived parameters
	double tfac  = k_h * T;
	double kab   = tfac * exp((-dHab / T + dSab) / R);
	double kba   = tfac * exp((-dHba / T + dSba) / R);
	double kbc   = tfac * exp((-dHbc / T + dSbc) / R);
	double kcb   = tfac * exp((-dHcb / T + dScb) / R);
	double kca   = tfac * exp((-dHca / T + dSca) / R);
	double kac   = tfac * exp((-dHac / T + dSac) / R);
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_temp3(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 14;
	return calc_deriv(data, par, dy, fixed, cpmg_temp3, i_R20);
}

double cpmg_temp3linear(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const double k_h = 2.0836644e10;		// Boltzman's constant / Planck's constant
	const double R   = 1.985878e-3;			// Gas constant (kcal/mol/K)
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double T     = data.x[1];
	double field = data.x[2];
	double dwab  = par[0];
	double dwac  = par[1];
	double dHab  = par[2];
	double dHba  = par[3];
	double dHbc  = par[4];
	double dHcb  = par[5];
	double dSab  = par[6];
	double dSba  = par[7];
	double dSbc  = par[8];
	double dScb  = par[9];
	double R20   = par[10];
	// Calculate derived parameters
	double tfac = k_h * T;
	double kab  = tfac * exp((-dHab / T + dSab) / R);
	double kba  = tfac * exp((-dHba / T + dSba) / R);
	double kbc  = tfac * exp((-dHbc / T + dSbc) / R);
	double kcb  = tfac * exp((-dHcb / T + dScb) / R);
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	// prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, 0, 0);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_temp3linear(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 10;
	return calc_deriv(data, par, dy, fixed, cpmg_temp3linear, i_R20);
}

double cpmg_temp3folked(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const double k_h = 2.0836644e10;		// Boltzman's constant / Planck's constant
	const double R   = 1.985878e-3;			// Gas constant (kcal/mol/K)
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double T     = data.x[1];
	double field = data.x[2];
	double dwab  = par[0];
	double dwac  = par[1];
	double dHab  = par[2];
	double dHba  = par[3];
	double dHca  = par[4];
	double dHac  = par[5];
	double dSab  = par[6];
	double dSba  = par[7];
	double dSca  = par[8];
	double dSac  = par[9];
	double R20   = par[10];
	// Calculate derived parameters
	double tfac  = k_h * T;
	double kab   = tfac * exp((-dHab / T + dSab) / R);
	double kba   = tfac * exp((-dHba / T + dSba) / R);
	double kca   = tfac * exp((-dHca / T + dSca) / R);
	double kac   = tfac * exp((-dHac / T + dSac) / R);
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	// prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, 0, 0, kca, kac);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_temp3folked(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 10;
	return calc_deriv(data, par, dy, fixed, cpmg_temp3folked, i_R20);
}

double cpmg_bind3a(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double field = data.x[2];
	double dwfb1 = par[0];
	double dwfb2 = par[1];
	double Bkon1 = par[2];
	double koff1 = par[3];
	double Bkon2 = par[4];
	double koff2 = par[5];
	double k12   = par[6];
	double k21   = par[7];
	double R20   = par[8];
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	// prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwfb1*field, dwfb2*field, Bkon1, koff1, k12, k21, koff2, Bkon2);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_bind3a(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 8;
	return calc_deriv(data, par, dy, fixed, cpmg_bind3a, i_R20);
}

double cpmg_bind3b(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double field = data.x[2];
	double dwfi  = par[0];
	double dwfb  = par[1];
	double kfi   = par[2];
	double kif   = par[3];
	double Bkon  = par[4];
	double koff  = par[5];
	double R20   = par[6];
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	// prep_R2mat(M0, D, U, Ui, dwab*field, dwac*field, kab, kba, kbc, kcb, kca, kac);
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwfi*field, dwfb*field, kfi, kif, Bkon, koff, 0, 0);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_bind3b(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 6;
	return calc_deriv(data, par, dy, fixed, cpmg_bind3b, i_R20);
}

double cpmg_bind3c_1F2B(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// Copy parameters
	double itcp  = data.x[0];		// 1/tcp
	double field = data.x[3];
	double c_sf  = data.x[4];		// ligand concnetration ratio between samples
	double dwfb1 = par[0];
	double dwfb2 = par[1];
	double kon1  = par[2] * 1e6;		// actual kon1 = kon1 x 1e6
	double koff1 = par[3];
	double kon2  = par[4] * 1e6;		// actual kon2 = kon2 x 1e6
	double koff2 = par[5];
	double k12   = par[6];
	double k21   = par[7];
	double A0    = par[8];
	double B0    = par[9];
	double R20   = par[10];
	// Calculate derived parameters
	double Kd    = (koff1*koff2+koff1*k21+koff2*k12) / (kon1*(koff2+k12+k21)+kon2*(koff1+k12+k21));
	double c_tmp = -Kd + B0 * c_sf - A0;
	double freeB = 0.5 * (c_tmp + sqrt(sqr(c_tmp) + 4*B0*c_sf*Kd));
	double Bkon1 = freeB * kon1;
	double Bkon2 = freeB * kon2;
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwfb1*field, dwfb2*field, Bkon1, koff1, k12, k21, koff2, Bkon2);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_bind3c_1F2B(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 10;
	return calc_deriv(data, par, dy, fixed, cpmg_bind3c_1F2B, i_R20);
}

double cpmg_bind3c_2F1B(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// copy parameters
	double itcp   = data.x[0];		// 1/tcp
	double field  = data.x[3];
	double c_sf   = data.x[4];		// ligand concnetration ratio between samples
	double dwf1f2 = par[0];
	double dwf1b  = par[1];
	double kon1   = par[2];		// actual kon1 = kon1 x 1e6
	double koff1  = par[3];
	double kon2   = par[4];		// actual kon2 = kon2 x 1e6
	double koff2  = par[5];
	double k12    = par[6];
	double k21    = par[7];
	double A0     = par[8];
	double B0     = par[9];
	double R20    = par[10];
	// calculate parameters
	double akon1 = kon1*1e6;
	double akon2 = kon2*1e6;
	double xB0 = B0*c_sf;
	double a = akon1*akon2;
	double b = akon1*akon2*(A0-xB0)+(koff1+k12)*akon2+(koff2+k21)*akon1;
	double c = -(koff1*akon2+koff2*akon1)*xB0+(k12*akon2+k21*akon1)*(A0-xB0)+(k12+k21)*(koff1+koff2);
	double d = -(k12+k21)*(koff1+koff2)*xB0;
	double freeB = 0;
	if (a == 0 && b == 0) {
		cerr << "illegal parameter specified\n";
		exit(1);
	} else if (a == 0) {
		freeB = (-c + sqrt(sqr(c)-4*b*d))/(2*b);
	} else {
		freeB = solveCubic(a, b, c, d);
	}
	if (freeB < 0 || !std::isfinite(freeB)) {
		return HUGE_VAL;
	}
	double Bkon1 = freeB * akon1;
	double Bkon2 = freeB * akon2;
	// Get relaxation Matrix R, eigenvectors U, eigenvalues D & inverse matrix U-1
	double M0[3], D[6], U[18], Ui[18];
	prep_R2mat(M0, D, U, Ui, dwf1f2*field, dwf1b*field, k12, k21, Bkon2, koff2, koff1, Bkon1);
	return R20 + calc_R2(itcp, M0, D, U, Ui);
}

double fcpmg_bind3c_2F1B(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 10;
	return calc_deriv(data, par, dy, fixed, cpmg_bind3c_2F1B, i_R20);
}

int Funclib::Read_CPMG3_Funcs(int idx)
{
	// CPMG Matrix 3-site exchange
	// index0: temperature, concentration ratio, or something
	// index1: magnetic field
	idx = Define_Func(&fcpmg_matrix3, "CPMG_matrix3"
		, "f(dwab, dwac, kab, kba, kbc, kca, kac, R20)"
		, "CT-CPMG", XDIM4, IDIM2, LOG);
	Define_Parm(idx, "dwab"  ,-12000, 12000, "00" );
	Define_Parm(idx, "dwac"  ,-12000, 12000, "00" );
	Define_Parm(idx, "kab"   ,     0, 12000, "10" );
	Define_Parm(idx, "kba"   ,     0, 12000, "10" );
	Define_Parm(idx, "kbc"   ,     0, 12000, "10" );
	Define_Parm(idx, "kca"   ,     0, 12000, "10" );
	Define_Parm(idx, "kac"   ,     0, 12000, "10" );
	Define_Parm(idx, "R20"   ,   0.1,   100, "11" );

	// CPMG Matrix 3-site exchange
	// Using Arrhenius's equation to describe dH & dS
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_temp3, "CPMG_temp3"
		, "f(dwab, dwac, dHab, dHba, dHbc, dHcb, dHca, dHac, dSab, dSba, dSbc, dScb, dSca, dSac, R20)"
		, "CT-CPMG", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "dwab"  ,-12000, 12000, "00");
	Define_Parm(idx, "dwac"  ,-12000, 12000, "00");
	Define_Parm(idx, "dHab"  ,   -50,    50, "00");
	Define_Parm(idx, "dHba"  ,   -50,    50, "00");
	Define_Parm(idx, "dHbc"  ,   -50,    50, "00");
	Define_Parm(idx, "dHcb"  ,   -50,    50, "00");
	Define_Parm(idx, "dHca"  ,   -50,    50, "00");
	Define_Parm(idx, "dHac"  ,   -50,    50, "00");
	Define_Parm(idx, "dSab"  ,    -1,     1, "00");
	Define_Parm(idx, "dSba"  ,    -1,     1, "00");
	Define_Parm(idx, "dSbc"  ,    -1,     1, "00");
	Define_Parm(idx, "dScb"  ,    -1,     1, "00");
	Define_Parm(idx, "dSca"  ,    -1,     1, "00");
	Define_Parm(idx, "dSac"  ,    -1,     1, "00");
	Define_Parm(idx, "R20"   ,   0.1,   100, "11");

	// CPMG Matrix linear 3-site exchange
	// Using Arrhenius's equation to describe dH & dS
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_temp3linear, "CPMG_temp3linear"
		, "f(dwab, dwac, dHab, dHba, dHbc, dHcb, dSab, dSba, dSbc, dScb, R20)"
		, "CT-CPMG", XDIM4, IDIM2, LOG);
	Define_Parm(idx, "dwab"  ,-12000, 12000, "00");
	Define_Parm(idx, "dwac"  ,-12000, 12000, "00");
	Define_Parm(idx, "dHab"  ,   -50,    50, "00");
	Define_Parm(idx, "dHba"  ,   -50,    50, "00");
	Define_Parm(idx, "dHbc"  ,   -50,    50, "00");
	Define_Parm(idx, "dHcb"  ,   -50,    50, "00");
	Define_Parm(idx, "dSab"  ,    -1,     1, "00");
	Define_Parm(idx, "dSba"  ,    -1,     1, "00");
	Define_Parm(idx, "dSbc"  ,    -1,     1, "00");
	Define_Parm(idx, "dScb"  ,    -1,     1, "00");
	Define_Parm(idx, "R20"   ,   0.1,   100, "11");

	// CPMG Matrix folked 3-site exchange
	// Using Arrhenius's equation to describe dH & dS
	// index0: temperature
	// index1: magnetic field
	idx = Define_Func(&fcpmg_temp3folked, "CPMG_temp3folked"
		, "f(dwab, dwac, dHab, dHba, dHca, dHac, dSab, dSba, dSca, dSac, R20)"
		, "CT-CPMG", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "dwab"  ,-12000, 12000, "00");
	Define_Parm(idx, "dwac"  ,-12000, 12000, "00");
	Define_Parm(idx, "dHab"  ,   -50,    50, "00");
	Define_Parm(idx, "dHba"  ,   -50,    50, "00");
	Define_Parm(idx, "dHca"  ,   -50,    50, "00");
	Define_Parm(idx, "dHac"  ,   -50,    50, "00");
	Define_Parm(idx, "dSab"  ,    -1,     1, "00");
	Define_Parm(idx, "dSba"  ,    -1,     1, "00");
	Define_Parm(idx, "dSca"  ,    -1,     1, "00");
	Define_Parm(idx, "dSac"  ,    -1,     1, "00");
	Define_Parm(idx, "R20"   ,   0.1,   100, "11");

	// CPMG Matrix 3-site exchange (R20f = R20i = R20b)
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind3a, "CPMG_bind3a"
		, "f(dwfb1, dwfb2, Bkon1, koff1, Bkon2, koff2, k12, k21, R20)"
		, "CT-CPMG", XDIM5, IDIM3, LOG);
	Define_Parm(idx, "dwfb1" ,-12000, 12000, "000");
	Define_Parm(idx, "dwfb2" ,-12000, 12000, "000");
	Define_Parm(idx, "Bkon1" ,     0, 12000, "101");
	Define_Parm(idx, "koff1" ,     0, 12000, "100");
	Define_Parm(idx, "Bkon2" ,     0, 12000, "101");
	Define_Parm(idx, "koff2" ,     0, 12000, "100");
	Define_Parm(idx, "k12"   ,     0, 12000, "100");
	Define_Parm(idx, "k21"   ,     0, 12000, "100");
	Define_Parm(idx, "R20"   ,   0.1,   100, "111");

	// CPMG Matrix 3-site exchange (R20f = R20i = R20b)
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind3b, "CPMG_bind3b"
		, "f(dwfi, dwfb, kfi, kif, Bkon, koff, R20)"
		, "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwfi"  ,-12000, 12000, "000");
	Define_Parm(idx, "dwfb"  ,-12000, 12000, "000");
	Define_Parm(idx, "kfi"   ,     0, 12000, "100");
	Define_Parm(idx, "kif"   ,     0, 12000, "100");
	Define_Parm(idx, "Bkon"  ,     0, 12000, "101");
	Define_Parm(idx, "koff"  ,     0, 12000, "100");
	Define_Parm(idx, "R20"   ,   0.1,   100, "111");

	// CPMG Matrix 3-site exchange (R20f = R20i = R20b)
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind3c_1F2B, "CPMG_bind3c_1F2B"
		, "f(dwfb1, dwfb2, kon1, koff1, k12, k21, A0, B0, R20)"
		, "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwfb1" ,-12000, 12000, "000");
	Define_Parm(idx, "dwfb2" ,-12000, 12000, "000");
	Define_Parm(idx, "kon1"  ,     0, 10000, "100");
	Define_Parm(idx, "koff1" ,     0, 12000, "100");
	Define_Parm(idx, "kon2"  ,     0, 10000, "100");
	Define_Parm(idx, "koff2" ,     0, 12000, "100");
	Define_Parm(idx, "k12"   ,     0, 12000, "100");
	Define_Parm(idx, "k21"   ,     0, 12000, "100");
	Define_Parm(idx, "A0"    ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "B0"    ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "R20"   ,     1,   100, "111");

	// CPMG Matrix 3-site exchange (R20f = R20i = R20b)
	// index0: temperature
	// index1: magnetic field
	// index2: concentration ratio
	idx = Define_Func(&fcpmg_bind3c_2F1B, "CPMG_bind3c_2F1B"
		, "f(dwf1f2, dwf1b, kon1, koff1, k12, k21, A0, B0, R20)"
		, "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwf1f2",-12000, 12000, "000");
	Define_Parm(idx, "dwf1b" ,-12000, 12000, "000");
	Define_Parm(idx, "kon1"  ,     0, 10000, "100");
	Define_Parm(idx, "koff1" ,     0, 12000, "100");
	Define_Parm(idx, "kon2"  ,     0, 10000, "100");
	Define_Parm(idx, "koff2" ,     0, 12000, "100");
	Define_Parm(idx, "k12"   ,     0, 12000, "100");
	Define_Parm(idx, "k21"   ,     0, 12000, "100");
	Define_Parm(idx, "A0"    ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "B0"    ,  1e-9,  1e-2, "000");
	Define_Parm(idx, "R20"   ,     1,   100, "111");

	return idx;
}
