// func_mqcpmg.cpp
// Bryn Fenwick			01/21/2013
// Kenji Sugase			12/17/2013
//
// J Am Chem Soc 126, 3964-3973 (2004)
// J Am Chem Soc 126, 7320-7329 (2004)

#include <iostream>
#include <cmath>
#include <cstdlib>
#include "Data.h"
#include "FuncLib.h"
#include "LinearAlgebra.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;
namespace {

void check_itcp(int itcp)
{
	if (itcp < LOW_ITCP) {
		cerr << "1/tcp is too short. Check the input file\n";
		exit(1);
	}
}

void check_spin(int spin)
{
	if (spin < 0 || spin > 5) {
		cerr << "ERROR: Unknown spin state.\n"
		     << "       0:15N DQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ\n";
		exit(1);
	}
}

void calc_dw(int spin, double &dw_d, double &dw_z, double dwN, double dwH)
{
	if (spin == 0) {		// 15N SQ
		dw_d = dwN;
		dw_z = dwN;
	} else if (spin == 1) {		// 1H SQ
		dw_d = dwH;
		dw_z = dwH;
	} else if (spin == 2) {		// ZQ
		dw_d = dwH - dwN;
		dw_z = dwH - dwN;
	} else if (spin == 3) {		// DQ
		dw_d = dwH + dwN;
		dw_z = dwH + dwN;
	} else if (spin == 4) {		// 15N MQ
		dw_d = dwN + dwH;
		dw_z = dwN - dwH;
	} else if (spin == 5) {		// 1H MQ
		dw_d = dwH + dwN;
		dw_z = dwH - dwN;
	}
}

void calc_dw(int spin, double &dwab_d, double &dwac_d, double &dwab_z, double &dwac_z, double dwabN
	, double dwacN, double dwabH, double dwacH)
{
	if (spin == 0) {		// 15N SQ
		dwab_d = dwabN;
		dwac_d = dwacN;
		dwab_z = dwabN;
		dwac_z = dwacN;
	} else if (spin == 1) {		// 1H SQ
		dwab_d = dwabH;
		dwac_d = dwacH;
		dwab_z = dwabH;
		dwac_z = dwacH;
	} else if (spin == 2) {		// ZQ
		dwab_d = dwabH - dwabN;
		dwac_d = dwacH - dwacN;
		dwab_z = dwabH - dwabN;
		dwac_z = dwacH - dwacN;
	} else if (spin == 3) {		// DQ
		dwab_d = dwabH + dwabN;
		dwac_d = dwacH + dwacN;
		dwab_z = dwabH + dwabN;
		dwac_z = dwacH + dwacN;
	} else if (spin == 4) {		// 15N MQ
		dwab_d = dwabN + dwabH;
		dwac_d = dwacN + dwacH;
		dwab_z = dwabN - dwabH;
		dwac_z = dwacN - dwacH;
	} else if (spin == 5) {		// 1H MQ
		dwab_d = dwabN + dwabH;
		dwac_d = dwacN + dwacH;
		dwab_z = dwabN - dwabH;
		dwac_z = dwacN - dwacH;
	}
}

void prep_R2mat(double *R, double dw, double kab, double kba, double R20a = 0, double R20b = 0)
{
	R[0] = R20a + kab;		R[1] = 0.0;		// (1, 1)
	R[2] =       -kab;		R[3] = 0.0;		// (2, 1)

	R[4] =       -kba;		R[5] = 0.0;		// (1, 2)
	R[6] = R20b + kba;		R[7] = -dw;		// (2, 2)
}

void prep_R2mat(double *R, double dwab, double dwac, double kab, double kba, double kac, double kca
	, double kbc, double kcb, double R20a = 0, double R20b = 0, double R20c = 0)
{
	R[0]  = R20a + kab + kac;	R[1]  = 0.0;		// (1, 1)
	R[2]  =       -kab      ;	R[3]  = 0.0;		// (2, 1)
	R[4]  =             -kac;	R[5]  = 0.0;		// (3, 1)

	R[6]  =       -kba      ;	R[7]  = 0.0;		// (1, 2)
	R[8]  = R20b + kba + kbc;	R[9]  = -dwab;		// (2, 2)
	R[10] =             -kbc;	R[11] = 0.0;		// (3, 2)

	R[12] =             -kca;	R[13] = 0.0;		// (1, 3)
	R[14] =       -kcb      ;	R[15] = 0.0;		// (2, 3)
	R[16] = R20c + kcb + kca;	R[17] = -dwac;		// (3, 3)
}

void calc_UDtUi(int order, const double *R, double *A, double *cA, double t)
{
	int sz = 2 * sqr(order);
	// get eigenvectors and eigenvalues of R
	double D[2*order], U[sz], Ui[sz], Dt[sz];
	if (zmat_eig(order, R, U, Ui, D)) {
		cerr << "ERROR: Relaxation matrix is singular.";
		exit(1);
	}
	zmat_exp(order, t, D, Dt);		// exp(-D*tcp/2)
	zdmat_mlt(order, U, Dt, Ui, A);		// A = U * Dt * U(-1)
	zmat_cnj(order, A, cA);
}

int find_ground_state(int order, const double *M0)
{
	double gpop = 0;
	int gstate = 0;
	for (int i = 0; i < order; i++) {
		if(M0[2*i] > gpop) {
			gpop = M0[2*i];
			gstate = 2*i;
		}
	}
	return gstate;
}

double calc_mqR2(int order, double *R_d, double *R_z, double *M0, double itcp, double Tcpmg)
{
	int sz = 2 * sqr(order);
	int n = static_cast<int>(itcp*Tcpmg)/2;
	// check that n is valid
	if (n < 1) {
		cerr << "ERROR: N is less than 2.\n";
		exit(1);
	}
	// calculate U * exp(-D*tcp/2) * U(-1)
	double half_tcp = -0.5 / itcp;			// -tcp/2
	double Dp[sz], Dm[sz], Zp[sz], Zm[sz];		// D+, D-, Z+, Z-
	calc_UDtUi(order, R_d, Dp, Dm, half_tcp);
	calc_UDtUi(order, R_z, Zp, Zm, half_tcp);
	// combine DQ and ZQ
	double DZZD[sz], ZDDZ[sz], DpZm[sz], DmZp[sz];
	zmat_mlt(order, Dp, Zm, Zm, Dp, DZZD);		// DZZD = D+ * Z- * Z- * D+
	if (n % 2 == 0) {	// even
		zmat_mlt(order, Zp, Dm, Dm, Zp, ZDDZ);	// ZDDZ = Z+ * D- * D- * Z+
	} else {		// odd
		zmat_mlt(order, Dm, Zp, Zp, Dm, ZDDZ);	// ZDDZ = D- * Z+ * Z+ * D-  <- this is correct.
		zmat_mlt(order, Dp, Zm, DpZm);		// DpZm = D+ * Z-
		zmat_mlt(order, Dm, Zp, DmZp);		// DmZp = D- * Z+
	}
	double A[sz], B[sz];
	zmat_cpy(order, DZZD, A);	// n = 2
	zmat_cpy(order, ZDDZ, B);	// n = 2
	double tmp[sz];
	for (int i = 0; i <= n - 4; i = i + 2) {
		zmat_mlt(order, DZZD, A, tmp);
		zmat_cpy(order, tmp, A);
		zmat_mlt(order, ZDDZ, B, tmp);
		zmat_cpy(order, tmp, B);
	}
	if (n != 1 && n % 2 == 1){
		zmat_mlt(order, A, DpZm, tmp);
		zmat_cpy(order, tmp, A);
		zmat_mlt(order, B, DmZp, tmp);
		zmat_cpy(order, tmp, B);
	}
	if (n == 1) {
		zmat_cpy(order, DmZp, A);
		zmat_cpy(order, DpZm, B);
	}
	// calculate the DQ and ZQ components
	double AB[sz], BA[sz], DQ[sz], cDQ[sz], ZQ[sz], cZQ[sz];
	zmat_mlt(order, A, B, AB);	// AB = A x B
	zmat_mlt(order, B, A, BA);	// BA = B x A
	zmat_vmlt(order, AB, M0, DQ);	// DQ = AB x M0
	zmat_vmlt(order, BA, M0, ZQ);	// ZQ = BA x M0
	zmat_cnj(order, DQ, cDQ);
	zmat_cnj(order, ZQ, cZQ);
	int gs = find_ground_state(order, M0);
	return log(4*M0[2*gs]/(DQ[2*gs]+ZQ[2*gs]+cDQ[2*gs]+cZQ[2*gs]))/Tcpmg;
}

/////////////////////////////////////////////////////////////////////////////////////////////


// Carver and Richards approximation
// par[0]	dwN			(rad/s)
// par[1]	dwH			(rad/s)
// par[2]	kex			(/s)
// par[3]	papb			(unitless)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	Tcpmg			(s)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
// data.x[4]	spin			(unitless)
double mqcpmg_richards(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	double itcp  = data.x[0];
	double Tcpmg = data.x[2];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwN   = par[0] * field;
	double dwH   = par[1] * field;
	double kex   = par[2];
	double papb  = par[3];
	double R20   = par[4];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate parameters
	double tcp = 1 / itcp;
	double kex2  = sqr(kex);
	double dwfld  = 0;
	double dwfld2 = 0;
	if (spin == 0) {
		dwfld = dwN;
	} else if (spin == 1) {
		dwfld = dwH;
	} else if (spin == 2) {
		dwfld = dwH - dwN;
	} else if (spin == 3) {
		dwfld = dwH + dwN;
	} else if (spin == 4) {
		dwfld  = dwN;
		dwfld2 = dwH;
	} else if (spin == 5) {
		dwfld  = dwH;
		dwfld2 = dwN;
	}
	double dw2 = sqr(dwfld);
	double pop = sqrt(1 - 4*papb);	// pa - pb
	if (spin == 0 || spin == 1 || spin == 2 || spin == 3) {
		double zeta  = 2*dwfld*kex*pop;
		double zeta2 = sqr(zeta);
		double psi   = kex2 - dw2;
		double norm  = sqrt(sqr(psi) + zeta2);
		double pdw2  = (kex2 + dw2) / norm;
		double D_p   = 0.5 * ( 1 + pdw2);	// D+
		double D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate effective R2
		double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
		double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
		if (eta_p < COSH_MAX) {
			double coshp = cosh(eta_p);			// cosh(eta+)
			double cosm  = cos(eta_m);			// cos(eta-)
			return R20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return R20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
		}
	} else {
		cdouble zeta  = -2*dwfld*(I*dwfld2+pop*kex);
		cdouble zeta2 = sqr(zeta);
		cdouble cpsi  = sqr(I*dwfld2+pop*kex)-dw2+4*papb*kex2;
		cdouble cnorm = sqrt(sqr(cpsi) + zeta2);
		cdouble cpdw2  = (cpsi + 2.0*dw2) / cnorm;
		cdouble cD_p  = 0.5 * ( 1.0 + cpdw2);
		cdouble cD_m  = 0.5 * (-1.0 + cpdw2);
		cdouble d_p   = (dwfld2 + dwfld) + kex*I;
		cdouble d_m   = (dwfld2 + dwfld) - kex*I;
		cdouble z_p   = (dwfld2 - dwfld) + kex*I;
		cdouble z_m   = (dwfld2 - dwfld) - kex*I;
		// calculate effective R2
		cdouble ceta_p = tcp * sqrt(0.5 * ( cpsi + cnorm));	// eta+
		cdouble ceta_m = tcp * sqrt(0.5 * (-cpsi + cnorm));	// eta-
		cdouble mD =  I*kex*sqrt(papb)/(d_p*z_p)*(z_p+2.0*dwfld*sin(z_p*tcp*0.5)/sin((d_p+z_p)*tcp*0.5));
		cdouble mZ = -I*kex*sqrt(papb)/(d_m*z_m)*(d_m-2.0*dwfld*sin(d_m*tcp*0.5)/sin((d_m+z_m)*tcp*0.5));
		double Q   = log(real(1.0-sqr(mD)+mD*mZ-sqr(mZ)+(mD+mZ)*0.5*sqrt((1-pop)/(1+pop))))/Tcpmg;
		if (real(ceta_p) < COSH_MAX) {
			cdouble coshp = cosh(ceta_p);			// cosh(eta+)
			cdouble cosm  = cos(ceta_m);			// cos(eta-)
			return real(R20 + 0.5 * (kex - itcp * acosh(cD_p * coshp - cD_m * cosm))) - Q;
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return real(R20 + 0.5 * (kex - itcp * (ceta_p + log(cD_p)))) - Q;
		}
	}
}

double fmqcpmg_richards(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 4;
	return calc_deriv(data, par, dy, fixed, mqcpmg_richards, i_R20);
}

// Carver and Richards approximation
// par[0]	dwN			(rad/s)
// par[1]	dwH			(rad/s)
// par[2]	kex			(/s)
// par[3]	pb			(unitless)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	Tcpmg			(s)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
// data.x[4]	spin			(unitless)
double mqcpmg_richards_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	double itcp  = data.x[0];
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwN = par[0] * field;
	double dwH = par[1] * field;
	double kex = par[2];
	double pb  = par[3];
	double R20 = par[4];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate parameters
	double tcp = 1 / itcp;
	double kex2  = sqr(kex);
	double dwfld  = 0;
	double dwfld2 = 0;
	if (spin == 0) {
		dwfld = dwN;
	} else if (spin == 1) {
		dwfld = dwH;
	} else if (spin == 2) {
		dwfld = dwH - dwN;
	} else if (spin == 3) {
		dwfld = dwH + dwN;
	} else if (spin == 4) {
		dwfld  = dwN;
		dwfld2 = dwH;
	} else if (spin == 5) {
		dwfld  = dwH;
		dwfld2 = dwN;
	}
	double dw2 = sqr(dwfld);
	double papb = (1 - pb) * pb;
	double pop = 1 - 2*pb;		// pa - pb
	if (spin == 0 || spin == 1 || spin == 2 || spin == 3) {
		double zeta  = 2*dwfld*kex*pop;
		double zeta2 = sqr(zeta);
		double psi   = kex2 - dw2;
		double norm  = sqrt(sqr(psi) + zeta2);
		double pdw2  = (kex2 + dw2) / norm;
		double D_p   = 0.5 * ( 1 + pdw2);	// D+
		double D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate effective R2
		double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
		double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
		if (eta_p < COSH_MAX) {
			double coshp = cosh(eta_p);			// cosh(eta+)
			double cosm  = cos(eta_m);			// cos(eta-)
			return R20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return R20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
		}
	} else {
		cdouble zeta  = -2*dwfld*(I*dwfld2+pop*kex);
		cdouble zeta2 = sqr(zeta);
		cdouble cpsi  = sqr(I*dwfld2+pop*kex)-dw2+4*papb*kex2;
		cdouble cnorm = sqrt(sqr(cpsi) + zeta2);
		cdouble cpdw2  = (cpsi + 2.0*dw2) / cnorm;
		cdouble cD_p  = 0.5 * ( 1.0 + cpdw2);
		cdouble cD_m  = 0.5 * (-1.0 + cpdw2);
		cdouble d_p   = (dwfld2 + dwfld) + kex*I;
		cdouble d_m   = (dwfld2 + dwfld) - kex*I;
		cdouble z_p   = (dwfld2 - dwfld) + kex*I;
		cdouble z_m   = (dwfld2 - dwfld) - kex*I;
		// calculate effective R2
		cdouble ceta_p = tcp * sqrt(0.5 * ( cpsi + cnorm));	// eta+
		cdouble ceta_m = tcp * sqrt(0.5 * (-cpsi + cnorm));	// eta-
		cdouble mD =  I*kex*sqrt(papb)/(d_p*z_p)*(z_p+2.0*dwfld*sin(z_p*tcp*0.5)/sin((d_p+z_p)*tcp*0.5));
		cdouble mZ = -I*kex*sqrt(papb)/(d_m*z_m)*(d_m-2.0*dwfld*sin(d_m*tcp*0.5)/sin((d_m+z_m)*tcp*0.5));
		double Q   = log(real(1.0-sqr(mD)+mD*mZ-sqr(mZ)+(mD+mZ)*0.5*sqrt((1-pop)/(1+pop))))/Tcpmg;
		if (real(ceta_p) < COSH_MAX) {
			cdouble coshp = cosh(ceta_p);			// cosh(eta+)
			cdouble cosm  = cos(ceta_m);			// cos(eta-)
			return real(R20 + 0.5 * (kex - itcp * acosh(cD_p * coshp - cD_m * cosm))) - Q;
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return real(R20 + 0.5 * (kex - itcp * (ceta_p + log(cD_p)))) - Q;
		}
	}
}

double fmqcpmg_richards_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 4;
	return calc_deriv(data, par, dy, fixed, mqcpmg_richards_p, i_R20);
}

// Carver and Richards approximation
// par[0]	dwN			(rad/s)
// par[1]	dwH			(rad/s)
// par[2]	kab			(/s)
// par[3]	kba			(/s)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	Tcpmg			(s)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
// data.x[4]	spin			(unitless)
double mqcpmg_richards_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	double itcp  = data.x[0];
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwN = par[0] * field;
	double dwH = par[1] * field;
	double kab = par[2];
	double kba = par[3];
	double R20 = par[4];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate parameters
	double tcp = 1 / itcp;
	double kex = kab + kba;
	double kex2  = sqr(kex);
	double dwfld  = 0;
	double dwfld2 = 0;
	if (spin == 0) {
		dwfld = dwN;
	} else if (spin == 1) {
		dwfld = dwH;
	} else if (spin == 2) {
		dwfld = dwH - dwN;
	} else if (spin == 3) {
		dwfld = dwH + dwN;
	} else if (spin == 4) {
		dwfld  = dwN;
		dwfld2 = dwH;
	} else if (spin == 5) {
		dwfld  = dwH;
		dwfld2 = dwN;
	}
	double dw2 = sqr(dwfld);
	double pb = kab / kex;
	double papb = (1 - pb) * pb;
	double pop = 1 - 2*pb;		// pa - pb
	if (spin == 0 || spin == 1 || spin == 2 || spin == 3) {
		double zeta  = 2*dwfld*kex*pop;
		double zeta2 = sqr(zeta);
		double psi   = kex2 - dw2;
		double norm  = sqrt(sqr(psi) + zeta2);
		double pdw2  = (kex2 + dw2) / norm;
		double D_p   = 0.5 * ( 1 + pdw2);	// D+
		double D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate effective R2
		double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
		double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
		if (eta_p < COSH_MAX) {
			double coshp = cosh(eta_p);			// cosh(eta+)
			double cosm  = cos(eta_m);			// cos(eta-)
			return R20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return R20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
		}
	} else {
		cdouble zeta  = -2*dwfld*(I*dwfld2+pop*kex);
		cdouble zeta2 = sqr(zeta);
		cdouble cpsi  = sqr(I*dwfld2+pop*kex)-dw2+4*papb*kex2;
		cdouble cnorm = sqrt(sqr(cpsi) + zeta2);
		cdouble cpdw2  = (cpsi + 2.0*dw2) / cnorm;
		cdouble cD_p  = 0.5 * ( 1.0 + cpdw2);
		cdouble cD_m  = 0.5 * (-1.0 + cpdw2);
		cdouble d_p   = (dwfld2 + dwfld) + kex*I;
		cdouble d_m   = (dwfld2 + dwfld) - kex*I;
		cdouble z_p   = (dwfld2 - dwfld) + kex*I;
		cdouble z_m   = (dwfld2 - dwfld) - kex*I;
		// calculate effective R2
		cdouble ceta_p = tcp * sqrt(0.5 * ( cpsi + cnorm));	// eta+
		cdouble ceta_m = tcp * sqrt(0.5 * (-cpsi + cnorm));	// eta-
		cdouble mD =  I*kex*sqrt(papb)/(d_p*z_p)*(z_p+2.0*dwfld*sin(z_p*tcp*0.5)/sin((d_p+z_p)*tcp*0.5));
		cdouble mZ = -I*kex*sqrt(papb)/(d_m*z_m)*(d_m-2.0*dwfld*sin(d_m*tcp*0.5)/sin((d_m+z_m)*tcp*0.5));
		double Q   = log(real(1.0-sqr(mD)+mD*mZ-sqr(mZ)+(mD+mZ)*0.5*sqrt((1-pop)/(1+pop))))/Tcpmg;
		if (real(ceta_p) < COSH_MAX) {
			cdouble coshp = cosh(ceta_p);			// cosh(eta+)
			cdouble cosm  = cos(ceta_m);			// cos(eta-)
			return real(R20 + 0.5 * (kex - itcp * acosh(cD_p * coshp - cD_m * cosm))) - Q;
		// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
		} else {
			return real(R20 + 0.5 * (kex - itcp * (ceta_p + log(cD_p)))) - Q;
		}
	}
}

double fmqcpmg_richards_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 4;
	return calc_deriv(data, par, dy, fixed, mqcpmg_richards_k, i_R20);
}

// Matrix2
// par[0]	dwN			(rad/s)
// par[1]	dwH			(rad/s)
// par[2]	kex			(/s)
// par[3]	pb			(unitless)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	Tcpmg			(s)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
// data.x[4]	spin			(unitless)
double mqcpmg_matrix2_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int order = 2;		// matrix size
	int sz = 2 * sqr(order);
	// copy parameters
	double itcp  = data.x[0];	// 1/tcp
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);
	double dwN   = par[0] * field;
	double dwH   = par[1] * field;
	double kex   = par[2];
	double pb    = par[3];
	double R20   = par[4];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate populations
	double M0[2*order];
	M0[0] = 1 - pb;		M0[1] = 0;	// pa
	M0[2] = pb    ;		M0[3] = 0;	// pb
	// calculate kinetic rates
	double kab = pb * kex;
	double kba = kex - kab;
	// calculate chemical shift differences
	double dw_d, dw_z;
	calc_dw(spin, dw_d, dw_z, dwN, dwH);
	// prepare relaxation Matrix R
	double R_d[sz], R_z[sz];
	prep_R2mat(R_d, dw_d, kab, kba);
	prep_R2mat(R_z, dw_z, kab, kba);
	// calculate effective R2 rate
	return R20 + calc_mqR2(order, R_d, R_z, M0, itcp, Tcpmg);
}

double fmqcpmg_matrix2_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 4;
	return calc_deriv(data, par, dy, fixed, mqcpmg_matrix2_p, i_R20);
}

// Matrix2
// par[0]	dwN			(rad/s)
// par[1]	dwH			(rad/s)
// par[2]	kab			(/s)
// par[3]	kba			(/s)
// par[4]	R20			(/s)
// data.x[0]	1/tcp			(/s)
// data.x[1]	temperature		(K) unused
// data.x[2]	Tcpmg			(s)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
// data.x[4]	spin			(unitless)
double mqcpmg_matrix2_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int order = 2;		// matrix size
	int sz = 2 * sqr(order);
	// copy parameters
	double itcp  = data.x[0];	// 1/tcp
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwN   = par[0] * field;
	double dwH   = par[1] * field;
	double kab   = par[2];
	double kba   = par[3];
	double R20   = par[4];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate populations
	double M0[2*order];
	M0[0] = kba/(kab + kba);	M0[1] = 0;	// pa
	M0[2] = 1 - M0[0]      ;	M0[3] = 0;	// pb
	// calculate chemical shift differences
	double dw_d, dw_z;
	calc_dw(spin, dw_d, dw_z, dwN, dwH);
	// prepare relaxation Matrix R
	double R_d[sz], R_z[sz];
	prep_R2mat(R_d, dw_d, kab, kba);
	prep_R2mat(R_z, dw_z, kab, kba);
	// calculate effective R2 rate
	return R20 + calc_mqR2(order, R_d, R_z, M0, itcp, Tcpmg);
}

double fmqcpmg_matrix2_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const  int i_R20 = 4;
	return calc_deriv(data, par, dy, fixed, mqcpmg_matrix2_k, i_R20);
}

// cyclic 3-state matrix  model for 6 types of expts 15N SQ, 1H SQ, ZQ, DQ, 15N MQ, 1H MQ
// set up to take two populations and three independent kinetic rates
double mqcpmg_matrix3_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int order = 3;		// matrix size
	int sz = 2 * sqr(order);
	// copy parameters
	double itcp  = data.x[0];	// 1/tcp
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwabN = par[0] * field;
	double dwabH = par[1] * field;
	double dwacN = par[2] * field;
	double dwacH = par[3] * field;
	double pb    = par[4];
	double pc    = par[5];
	double kex1  = par[6];
	double kex2  = par[7];
	double kex3  = par[8];
	double R20   = par[9];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate populations
	double M0[2*order];
	M0[0] = 1 - pb - pc;	M0[1] = 0;	// pa
	M0[2] = pb;		M0[3] = 0;	// pb
	M0[4] = pc;		M0[5] = 0;	// pc
	// calculate kinetic rates
	double kba = (kex1 * (1 - pb - pc))/(1 - pc);
	double kab = (kex1 * pb)/(1 - pc);
	double kbc = (kex2 * pc)/(pb + pc);
	double kcb = (kex2 * pb)/(pb + pc);
	double kca = (kex3 * (1 -pc - pb))/(1 - pb);
	double kac = (kex3 * pc)/(1 - pb);
	// calculate chemical shift differences
	double dwab_d, dwac_d, dwab_z, dwac_z;
	calc_dw(spin, dwab_d, dwac_d, dwab_z, dwac_z, dwabN, dwacN, dwabH, dwacH);
	// prepare relaxation Matrix R
	double R_d[sz], R_z[sz];
	prep_R2mat(R_d, dwab_d, dwac_d, kab, kba, kac, kca, kbc, kcb);
	prep_R2mat(R_z, dwab_z, dwac_z, kab, kba, kac, kca, kbc, kcb);
	// calculate effective R2
	return R20 + calc_mqR2(order, R_d, R_z, M0, itcp, Tcpmg);
}

double fmqcpmg_matrix3_p(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const int i_R20 = 9;
	return calc_deriv(data, par, dy, fixed, mqcpmg_matrix3_p, i_R20);
}

// cyclic 3-state matrix  model for 6 types of expts 15N-SQ, 1H-SQ, ZQ, DQ, 15N-MQ, 1H-MQ
// set up to take five independent kinetic rates
double mqcpmg_matrix3_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	int order = 3;			// matrix size
	int sz = 2 * sqr(order);
	// copy parameters
	double itcp  = data.x[0];	// 1/tcp
	double Tcpmg = data.x[1];
	double field = data.x[3];	// B0 / Base B0
	int spin = static_cast<int>(data.x[4]);		// 0:15N SQ, 1:1H SQ, 2:ZQ, 3:DQ, 4:15N MQ, 5:1H MQ
	double dwabN = par[0] * field;
	double dwabH = par[1] * field;
	double dwacN = par[2] * field;
	double dwacH = par[3] * field;
	double kab   = par[4];
	double kba   = par[5];
	double kbc   = par[6];
	double kca   = par[7];
	double kac   = par[8];
	double R20   = par[9];
	// check parameters
	check_itcp(itcp);
	check_spin(spin);
	// calculate kinetic rates
	double kcb   = (kab * kbc * kca) / (kba * kac);
	// calculate populations
	double M0[2*order];
	double denom = 1 / ((kba + kbc)*(kac + kca) + (kac + kba)*kcb + kab*(kbc + kca + kcb));
	M0[0] = (kbc * kca + kba * (kca + kcb)) * denom;	M0[1] = 0;	// pa
	M0[2] = (kac * kcb + kab * (kca + kcb)) * denom;	M0[3] = 0;	// pb
	M0[4] = (kab * kbc + kac * (kba + kbc)) * denom;	M0[5] = 0;	// pc
	// calculate chemical shift differences
	double dwab_d, dwac_d, dwab_z, dwac_z;
	calc_dw(spin, dwab_d, dwac_d, dwab_z, dwac_z, dwabN, dwacN, dwabH, dwacH);
	// prepare relaxation Matrix R
	double R_d[sz], R_z[sz];
	prep_R2mat(R_d, dwab_d, dwac_d, kab, kba, kac, kca, kbc, kcb);
	prep_R2mat(R_z, dwab_z, dwac_z, kab, kba, kac, kca, kbc, kcb);
	// calculate effective R2
	return R20 + calc_mqR2(order, R_d, R_z, M0, itcp, Tcpmg);
}

double fmqcpmg_matrix3_k(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	const  int i_R20 = 9;
	return calc_deriv(data, par, dy, fixed, mqcpmg_matrix3_k, i_R20);
}

}

int Funclib::Read_MQCPMG_Funcs(int idx)
{
	// MQ CPMG Carver & Richards approximation
	// index0: temperature
	// index1: Tcpmg
	// index2: magnetic field
	idx = Define_Func(&fmqcpmg_richards, "MQCPMG_Richards"
		, "f(dwN, dwH, kex, papb, R20)", "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwN" ,-12000, 12000, "000" );
	Define_Parm(idx, "dwH" ,-12000, 12000, "000" );
	Define_Parm(idx, "kex" ,     1, 12000, "100" );
	Define_Parm(idx, "papb", 0.001, 0.249, "100" );
	Define_Parm(idx, "R20" ,   0.1,   100, "111" );

	// MQ CPMG Carver & Richards approximation
	// index0: temperature
	// index1: Tcpmg
	// index2: magnetic field
	idx = Define_Func(&fmqcpmg_richards_p, "MQCPMG_Richards_p"
		, "f(dwN, dwH, kex, pb, R20)", "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwN" ,-12000, 12000, "000" );
	Define_Parm(idx, "dwH" ,-12000, 12000, "000" );
	Define_Parm(idx, "kex" ,     1, 12000, "100" );
	Define_Parm(idx, "pb"  , 0.005, 0.249, "100" );
	Define_Parm(idx, "R20" ,   0.1,   100, "111" );

	// MQ CPMG Carver & Richards approximation
	// index0: temperature
	// index1: Tcpmg
	// index2: magnetic field
	idx = Define_Func(&fmqcpmg_richards_k, "MQCPMG_Richards_k"
		, "f(dwN, dwH, kab, kba, R20)", "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwN" ,-12000, 12000, "000" );
	Define_Parm(idx, "dwH" ,-12000, 12000, "000" );
	Define_Parm(idx, "kab" ,     1, 12000, "100" );
	Define_Parm(idx, "kba" ,     1, 12000, "100" );
	Define_Parm(idx, "R20" ,   0.1,   100, "111" );

	// MQ CPMG Matrix 2-site exchange
	// index0: temperature
	// index1: magnetic field
	// index2: spin type
	idx = Define_Func(&fmqcpmg_matrix2_p, "MQCPMG_Matrix2_p"
		, "f(dwN, dwH, kex, pb, R20)", "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwN" ,-12000, 12000, "000");
	Define_Parm(idx, "dwH" ,-12000, 12000, "000");
	Define_Parm(idx, "kex" ,     1, 12000, "100");
	Define_Parm(idx, "pb"  , 0.005,  0.49, "100");
	Define_Parm(idx, "R20" ,   0.1,   100, "111");

	// MQ CPMG Matrix 2-site exchange
	// index0: temperature
	// index1: magnetic field
	// index2: spin type
	idx = Define_Func(&fmqcpmg_matrix2_k, "MQCPMG_Matrix2_k"
		, "f(dwN, dwH, kab, kba, R20)", "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwN" , -12000, 12000, "000");
	Define_Parm(idx, "dwH" , -12000, 12000, "000");
	Define_Parm(idx, "kab" ,      1, 12000, "100");
	Define_Parm(idx, "kba" ,      1, 12000, "100");
	Define_Parm(idx, "R20" ,    0.1,   100, "111");

	// MQ CPMG Matrix 3-site exchange
	// index0: temperature
	// index1: magnetic field
	// index2: spin type
	idx = Define_Func(&fmqcpmg_matrix3_p, "MQCPMG_Matrix3_p"
		, "f(dwabN, dwabH, dwacN, dwacH, pb, pc, kex1, kex2, kex3, R20)"
		, "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwabN", -12000, 12000, "000");
	Define_Parm(idx, "dwabH", -12000, 12000, "000");
	Define_Parm(idx, "dwacN", -12000, 12000, "000");
	Define_Parm(idx, "dwacH", -12000, 12000, "000");
	Define_Parm(idx, "pb"   ,  0.005,  0.49, "100");
	Define_Parm(idx, "pc"   ,  0.005,  0.49, "100");
	Define_Parm(idx, "kex1" ,      0, 12000, "100");
	Define_Parm(idx, "kex2" ,      0, 12000, "100");
	Define_Parm(idx, "kex3" ,      0, 12000, "100");
	Define_Parm(idx, "R20"  ,    0.1,   100, "111");

	// MQ CPMG Matrix 3-site exchange
	// index0: temperature
	// index1: magnetic field
	// index2: spin type
	idx = Define_Func(&fmqcpmg_matrix3_k, "MQCPMG_Matrix3_k"
		, "f(dwabN, dwabH, dwacN, dwacH, kab, kba, kca, kac, kbc, R20)"
		, "CT-CPMG", XDIM5, IDIM3, GAUSS);
	Define_Parm(idx, "dwabN", -12000, 12000, "000");
	Define_Parm(idx, "dwabH", -12000, 12000, "000");
	Define_Parm(idx, "dwacN", -12000, 12000, "000");
	Define_Parm(idx, "dwacH", -12000, 12000, "000");
	Define_Parm(idx, "kab"  ,      0, 12000, "100");
	Define_Parm(idx, "kba"  ,      0, 12000, "100");
	Define_Parm(idx, "kac"  ,      0, 12000, "100");
	Define_Parm(idx, "kca"  ,      0, 12000, "100");
	Define_Parm(idx, "kbc"  ,      0, 12000, "100");
	Define_Parm(idx, "R20"  ,    0.1,   100, "111");

	return idx;
}
