// func_rdd.cpp
// R2 dispersion difference experiment
// Kenji Sugase		3/19/2012

#include <iostream>
#include <cmath>
#include <cstdlib>
#include <omp.h>
#include "Data.h"
#include "FuncLib.h"
#include "LinearAlgebra.h"
#include "Type.h"
#include "Mathlib.h"

using namespace std;
namespace{

// Constant value function for RDD data
// par[0]	I0			(unitless)
// data.x[0]	1/tcp			(/sec)
// data.x[1]	reference 1/tcp		(/sec)
// data.x[2]	Tcpmg			(/sec)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
double frdd_const(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	if (dy.size()) {
		// calculate partial derivatives
		dy[data.map[0]] = 1;
	}
	return par[0];
}

// Fast limit CPMG function, Luz & Meiboom approximation
// par[0]	kex			(/sec)
// par[1] 	pdw = papb*dw^2		(rad2/sec2)
// par[2]	R20			(/sec)
// par[4]	I0			(unitless)
// data.x[0]	1/tcp			(/sec)
// data.x[1]	maximum 1/tcp		(/sec)
// data.x[2]	Tcpmg			(msec)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
double frdd_meiboom(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_kex = data.map[0];
	const int i_pdw = data.map[1];
	const int i_r20 = data.map[2];
	const int i_I0  = data.map[3];
	// copy parameters
	double itcp   = data.x[0];
	double itcp0  = data.x[1];
	double tcpmg  = data.x[2];
	double field  = data.x[3];				// B0 Field / Base B0 Field
	double kex    = par[0];
	double pdw    = par[1] * sqr(field);			// papb * (dw * field)^2
	double r20    = par[2];
	double I0     = par[3];
	// calculate parameters
	double ktcp    = 0.5 * kex / itcp;			// kex * tcp/2
	double tanhkt  = 1 / ktcp * tanh(ktcp);			// 2 / (kex * tcp) * tanh(kex * tcp/2)
	double r2eff   = r20 + pdw / kex * (1 - tanhkt);	// effective R2
	double ktcp0   = 0.5 * kex / itcp0;			// kex * tcp/2
	double tanhkt0 = 1 / ktcp0 * tanh(ktcp0);		// 2 / (kex * tcp) * tanh(kex * tcp/2)
	double r2eff0  = r20 + pdw / kex * (1 - tanhkt0);	// effective R2
	double dexp    = exp(-r2eff0*tcpmg) - exp(-r2eff*tcpmg);
	double dI      = I0 * dexp;
	if (!dy.size()) {
		// no partial derivatives
		return dI;
	}
	// calculate partial derivatives
	if (!fixed[i_kex]) {
		if (ktcp < COSH_MAX) {
			double dr2eff_kex  = pdw / sqr(kex) * (-1 + 2 * tanhkt  - 1 / sqr(cosh(ktcp)));
			double dr2eff0_kex = pdw / sqr(kex) * (-1 + 2 * tanhkt0 - 1 / sqr(cosh(ktcp0)));
			dy[i_kex] = -I0*tcpmg*(dr2eff0_kex*exp(-r2eff0*tcpmg)-dr2eff_kex*exp(-r2eff*tcpmg));
		} else {
			double dr2eff_kex  = pdw / sqr(kex) * (-1 + 2 * tanhkt);
			double dr2eff0_kex = pdw / sqr(kex) * (-1 + 2 * tanhkt0);
			dy[i_kex] = -I0*tcpmg*(dr2eff0_kex*exp(-r2eff0*tcpmg)-dr2eff_kex*exp(-r2eff*tcpmg));
		}
	}
	if (!fixed[i_pdw]) {
		double dr2eff_pdw  = sqr(field) / kex * (1 - tanhkt);
		double dr2eff0_pdw = sqr(field) / kex * (1 - tanhkt0);
		dy[i_pdw] = -I0*tcpmg*(dr2eff0_pdw*exp(-r2eff0*tcpmg)-dr2eff_pdw*exp(-r2eff*tcpmg));
		}
	if (!fixed[i_r20]) {
		dy[i_r20] = -dI * tcpmg;
	}
	if (!fixed[i_I0]) {
		dy[i_I0]  = dexp;
	}
	return dI;
}

// Partial derivatives. dX/dR20 is not necessary
// pdrv[X][0]: d(kex)/d(X)
// pdrv[X][1]: d(zeta)/d(X) * zeta
// pdrv[X][2]: d(psi)/d(X)
// pdrv[X][3]: inorm * {d(psi)/d(X) * psi + d(zeta)/d(X) * zeta}
// pdrv[X][4]: d(D)/d(X)
void dX_dw(double *pdrv, double field, double dw, double zeta2, double psi, double inorm, double pdw2)
{
	//pdrv[0] = 0;
	pdrv[1] = zeta2 / dw;
	pdrv[2] = -2 * dw * sqr(field);
	pdrv[3] = inorm * (pdrv[1] + psi * pdrv[2]);
	pdrv[4] = -0.5 * inorm * (pdrv[2] + pdw2 * pdrv[3]);
}

void dtmp_D(double *pdrv, double psi, double inorm, double pdw2)
{
	pdrv[3] = inorm * (pdrv[1] + psi * pdrv[2]);
	pdrv[4] = 0.5 * inorm * (pdrv[2] - pdw2 * pdrv[3]);
}

double dX_R2(double *pdrv, double tmp_p, double tmp_m, double dnm, double cshcs, double Dps, double Dms)
{
	double deta_p = tmp_p * ( pdrv[2] + pdrv[3]);
	double deta_m = tmp_m * (-pdrv[2] + pdrv[3]);
	return 0.5 * pdrv[0] - dnm * (pdrv[4] * cshcs + Dps * deta_p + Dms * deta_m);
}

double dX_R2(double *pdrv, double tmp_p, double itcp, double D_p)
{
	return 0.5 * (pdrv[0] - itcp * (pdrv[4] / (D_p + TINY) + tmp_p * (pdrv[2] + pdrv[3])));
}

double calc_RDD(const Data &data, dvect &dy, const bvect &fixed, double (*pdrv)[5], int i_r20
	, int i_I0, double kex, double r20, double I0, double psi, double D_p, double D_m, double norm)
{
	double itcp  = data.x[0];
	double itcp0 = data.x[1];
	double tcpmg = data.x[2];
	if (itcp < LOW_ITCP || itcp0 < LOW_ITCP) {
		cerr << "1/tcp is too short. Check the input file\n";
		exit(1);
	}
	// calculate effective RDD
	double tcp  = 1 / itcp;
	double tcp0 = 1 / itcp0;
	double eta_p  = tcp  * sqrt(0.5 * ( psi + norm));	// eta+
	double eta_m  = tcp  * sqrt(0.5 * (-psi + norm));	// eta-
	double eta_p0 = tcp0 * sqrt(0.5 * ( psi + norm));	// eta+
	double eta_m0 = tcp0 * sqrt(0.5 * (-psi + norm));	// eta-
	double dI = 0;
	int sz = data.map.size() - 2;
	if (eta_p < COSH_MAX) {
		double coshp  = cosh(eta_p);			// cosh(eta+)
		double cosm   = cos(eta_m);			// cos(eta-)
		double r2eff  = r20 + 0.5 * (kex - itcp * acosh(D_p * coshp  - D_m * cosm));
		double coshp0 = cosh(eta_p0);			// cosh(eta+)
		double cosm0  = cos(eta_m0);			// cos(eta-)
		double r2eff0 = r20 + 0.5 * (kex - itcp0 * acosh(D_p * coshp0 - D_m * cosm0));
		double dexp   = exp(-r2eff0*tcpmg) - exp(-r2eff*tcpmg);
		       dI     = I0*dexp;
		// calculate partial derivatives of effective R2
		if (dy.size()) {
			double tmp_p  = tcp /  sqrt(8 * ( psi + norm));
			double tmp_m  = tcp / (sqrt(8 * (-psi + norm)) + TINY);
			double dnm    = 0.5 * itcp / (sqrt(sqr(D_p * coshp - D_m * cosm) - 1) + TINY);
			double Dps    = D_p * sinh(eta_p);
			double Dms    = D_m * sin(eta_m);
			double cshcs  = coshp - cosm;
			double tmp_p0 = tcp0 /  sqrt(8 * ( psi + norm));
			double tmp_m0 = tcp0 / (sqrt(8 * (-psi + norm)) + TINY);
			double dnm0   = 0.5 * itcp0 / (sqrt(sqr(D_p * coshp0 - D_m * cosm0) - 1) + TINY);
			double Dps0   = D_p * sinh(eta_p0);
			double Dms0   = D_m * sin(eta_m0);
			double cshcs0 = coshp0 - cosm0;
			for (int i = 0; i < sz; i++) {
				int index = data.map[i];
				if (fixed[index]) { continue; }
				double deta_p  = tmp_p * ( pdrv[i][2] + pdrv[i][3]);
				double deta_m  = tmp_m * (-pdrv[i][2] + pdrv[i][3]);
				double dr2eff  = 0.5*pdrv[i][0]-dnm*(pdrv[i][4]*cshcs+Dps*deta_p+Dms*deta_m);
				double deta_p0 = tmp_p0 * ( pdrv[i][2] + pdrv[i][3]);
				double deta_m0 = tmp_m0 * (-pdrv[i][2] + pdrv[i][3]);
				double dr2eff0 = 0.5*pdrv[i][0]-dnm0*(pdrv[i][4]*cshcs0+Dps0*deta_p0+Dms0*deta_m0);
				dy[index] = -I0*tcpmg*(dr2eff0*exp(-r2eff0*tcpmg)-dr2eff*exp(-r2eff*tcpmg));
			}
			dy[i_r20] = -dI * tcpmg;		// d(dI)/d(R20)
			dy[i_I0]  = dexp;			// d(dI)/d(I0)

		}
	// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
	} else {
		double r2eff  = r20 + 0.5 * (kex - itcp  * (eta_p  + log(D_p)));
		double r2eff0 = r20 + 0.5 * (kex - itcp0 * (eta_p0 + log(D_p)));
		double dexp   = exp(-r2eff0*tcpmg) - exp(-r2eff*tcpmg);
		double dI     = I0*dexp;
		// Calculate the partial derivatives of effective R2
		if (dy.size()) {
			double tmp_p  = tcp  / sqrt(8 * (psi + norm));
			double tmp_p0 = tcp0 / sqrt(8 * (psi + norm));
			for (int i = 0; i < sz; i++) {
				int index = data.map[i];
				if (fixed[index]) { continue; }
				double dr2eff  = 0.5*(pdrv[i][0]-itcp*(pdrv[i][4]/(D_p+TINY)+tmp_p*(pdrv[i][2]+pdrv[i][3])));
				double dr2eff0 = 0.5*(pdrv[i][0]-itcp0*(pdrv[i][4]/(D_p+TINY)+tmp_p0*(pdrv[i][2]+pdrv[i][3])));
				dy[index] = -I0*tcpmg*(dr2eff0*exp(-r2eff0*tcpmg)-dr2eff*exp(-r2eff*tcpmg));
			}
			dy[i_r20] = -dI * tcpmg;		// d(dI)/d(R20)
			dy[i_I0]  = dexp;			// d(dI)/d(I0)
		}
	}
	return dI;
}

double calc_RD(const Data &data, dvect &dy, const bvect &fixed, double (*pdrv)[5], int i_r20, int i_I0
	, double kex, double r20, double psi, double D_p, double D_m, double norm)
{
	double itcp = data.x[0];
	if (itcp < LOW_ITCP) {
		cerr << "1/tcp is too short. Check the input file\n";
		exit(1);
	}
	// calculate effective R2
	double r2eff, tmp_p;
	double tcp = 1 / itcp;
	double eta_p = tcp * sqrt(0.5 * ( psi + norm));		// eta+
	double eta_m = tcp * sqrt(0.5 * (-psi + norm));		// eta-
	int sz = data.map.size() - 2;
	if (eta_p < COSH_MAX) {
		double coshp = cosh(eta_p);			// cosh(eta+)
		double cosm  = cos(eta_m);			// cos(eta-)
		r2eff = r20 + 0.5 * (kex - itcp * acosh(D_p * coshp - D_m * cosm));
		// calculate partial derivatives of effective R2
		if (dy.size()) {
			       tmp_p = tcp /  sqrt(8 * ( psi + norm));
			double tmp_m = tcp / (sqrt(8 * (-psi + norm)) + TINY);
			double dnm   = 0.5 * itcp / (sqrt(sqr(D_p * coshp - D_m * cosm) - 1) + TINY);
			double Dps   = D_p * sinh(eta_p);
			double Dms   = D_m * sin(eta_m);
			double cshcs = coshp - cosm;
			for (int i = 0; i < sz; i++) {
				if (fixed[data.map[i]]) { continue; }
				double deta_p   = tmp_p*( pdrv[i][2]+pdrv[i][3]);
				double deta_m   = tmp_m*(-pdrv[i][2]+pdrv[i][3]);
				dy[data.map[i]] = 0.5*pdrv[i][0]-dnm*(pdrv[i][4]*cshcs+Dps*deta_p+Dms*deta_m);
			}
			dy[i_r20] = 1;			// d(effR2)/d(R20)
			dy[i_I0]  = 0;			// d(effR2)/d(I0)

		}
	// Use approximation that D_p * cosh(eta_p) >> D_m * cos(eta_m)
	} else {
		r2eff = r20 + 0.5 * (kex - itcp * (eta_p + log(D_p)));
		// Calculate the partial derivatives of effective R2
		if (dy.size()) {
			tmp_p = tcp / sqrt(8 * (psi + norm));
			for (int i = 0; i < sz; i++) {
				if (fixed[data.map[i]]) { continue; }
				dy[data.map[i]] = 0.5*(pdrv[i][0]-itcp*(pdrv[i][4]/(D_p+TINY)+tmp_p*(pdrv[i][2]+pdrv[i][3])));
			}
			dy[i_r20] = 1;			// d(effR2)/d(R20)
			dy[i_I0]  = 0;			// d(effR2)/d(I0)
		}
	}
	return r2eff;
}

// Carver and Richards approximation (R20a = R20b)
// par[0]	dw			(rad/sec)
// par[1]	kex			(/sec)
// par[2]	papb			(unitless)
// par[3]	R20			(/sec)
// par[4]	I0			(unitless)
// data.x[0]	1/tcp			(/sec)
// data.x[1]	maximum 1/tcp		(/sec)
// data.x[2]	Tcpmg			(msec)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
double frdd_richards(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kex  = data.map[1];
	const int i_papb = data.map[2];
	const int i_r20  = data.map[3];
	const int i_I0   = data.map[4];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double field, dw, papb, kex;
	static double psi, D_p, D_m, norm;
	// partial derivatives. dX/dR20 and dX/dI0 are not necessary
	static double pdrv[3][5]  = { {0,0,0,0,0}, {0,0,0,0,0}, {0,0,0,0,0} };
	// copy parameters if they changed
	if (field != data.x[3] || dw != par[0] || kex != par[1] || papb != par[2]){
		field  = data.x[3];			// B0 / Base B0
		dw     = par[0];
		kex    = par[1];
		papb   = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = sqrt(1 - 4 * papb);	// pa - pb
		double zeta2 = sqr(2 * dwfld * kex * pop);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				dX_dw(pdrv[0], field, dw, zeta2, psi, inorm, pdw2);
			}
			if (!fixed[i_kex]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta2 / kex;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_papb]) {
				//pdrv[2][0] = 0;
				pdrv[2][1] = -2 * zeta2 / sqr(pop);
				//pdrv[2][2] = 0;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	double I0  = par[4];
	return calc_RDD(data, dy, fixed, pdrv, i_r20, i_I0, kex, r20, I0, psi, D_p, D_m, norm);
}

// Carver and Richards approximation (R20a = R20b)
// par[0]	dw			(rad/sec)
// par[1]	kex			(/sec)
// par[2]	papb			(unitless)
// par[3]	R20			(/sec)
// par[4]	I0			(unitless)
// data.x[0]	1/tcp			(/sec)
// data.x[1]	maximum 1/tcp		(/sec)
// data.x[2]	Tcpmg			(msec)
// data.x[3]	B0 Field/Base B0 Field	(unitless)
double frdd_rd(const Data &data, double *par, dvect &dy, const bvect &fixed)
{
	// define index
	const int i_dw   = data.map[0];
	const int i_kex  = data.map[1];
	const int i_papb = data.map[2];
	const int i_r20  = data.map[3];
	const int i_I0   = data.map[4];
	// speed up calculations by keeping parameters independent of 1/tcp
	static double field, dw, papb, kex;
	static double psi, D_p, D_m, norm;
	// partial derivatives. dX/dR20 and dX/dI0 are not necessary
	static double pdrv[3][5]  = { {0,0,0,0,0}, {0,0,0,0,0}, {0,0,0,0,0} };
	// copy parameters if they changed
	if (field != data.x[3] || dw != par[0] || kex != par[1] || papb != par[2]){
		field  = data.x[3];			// B0 / Base B0
		dw     = par[0];
		kex    = par[1];
		papb   = par[2];
		// calculate parameters
		double dwfld = dw * field;
		double dw2   = sqr(dwfld);
		double kex2  = sqr(kex);
		double pop   = sqrt(1 - 4 * papb);	// pa - pb
		double zeta2 = sqr(2 * dwfld * kex * pop);
		       psi   = kex2 - dw2;
		       norm  = sqrt(sqr(psi) + zeta2);
		double inorm = 1 / norm;
		double pdw2  = (kex2 + dw2) * inorm;
		       D_p   = 0.5 * ( 1 + pdw2);	// D+
		       D_m   = 0.5 * (-1 + pdw2);	// D-
		// calculate partial derivatives independent of 1/tcp
		if (dy.size()) {
			if (!fixed[i_dw]) {
				dX_dw(pdrv[0], field, dw, zeta2, psi, inorm, pdw2);
			}
			if (!fixed[i_kex]) {
				pdrv[1][0] = 1;
				pdrv[1][1] = zeta2 / kex;
				pdrv[1][2] = 2 * kex;
				dtmp_D(pdrv[1], psi, inorm, pdw2);
			}
			if (!fixed[i_papb]) {
				//pdrv[2][0] = 0;
				pdrv[2][1] = -2 * zeta2 / sqr(pop);
				//pdrv[2][2] = 0;
				dtmp_D(pdrv[2], psi, inorm, pdw2);
			}
		}
	}
	// Calculate effective R2 rate & partial derivatives dependent on 1/tcp
	double r20 = par[3];
	double I0  = par[4];
	if (I0 == 0) {
		return calc_RD(data, dy, fixed, pdrv, i_r20, i_I0, kex, r20, psi, D_p, D_m, norm);
	} else {
		return calc_RDD(data, dy, fixed, pdrv, i_r20, i_I0, kex, r20, I0, psi, D_p, D_m, norm);
	}
}

}

int Funclib::Read_rdd_Funcs(int idx)
{
	// Constant I0
	// index0: Tcpmg
	// index1: magnetic field
	idx = Define_Func(&frdd_const, "RDD_const"
		, "f(I0)", "RDD", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "I0"  ,     1,   1e9, "01" );

	// CPMG Luz & Meiboom approximation (kex >> dw)
	// index0: Tcpmg
	// index1: magnetic field
	idx = Define_Func(&frdd_meiboom, "RDD_Meiboom"
		, "f(R20, pdw, kex, I0)", "RDD", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "kex"  ,   0.1, 12000, "00" );
	Define_Parm(idx, "pdw"  , 1e-30, 3.6e7, "00" );
	Define_Parm(idx, "R20"  ,   0.1,    50, "01" );
	Define_Parm(idx, "I0"   ,     0,   1e9, "01" );

	// CPMG Carver & Richards approximation (R20a = R20b)
	// index0: Tcpmg
	// index1: magnetic field
	idx = Define_Func(&frdd_richards, "RDD_Richards"
		, "f(dw, kex, papb, R20, I0)", "RDD", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kex"  ,     1, 12000, "00" );
	Define_Parm(idx, "papb" , 0.001, 0.249, "00" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "01" );
	Define_Parm(idx, "I0"   ,     0,   1e9, "01" );

	// CPMG Carver & Richards approximation (R20a = R20b)
	// index0: Tcpmg
	// index1: magnetic field
	idx = Define_Func(&frdd_rd, "RDD_RD"
		, "f(dw, kex, papb, R20, I0)", "RDD", XDIM4, IDIM2, GAUSS);
	Define_Parm(idx, "dw"   ,     1, 12000, "00" );
	Define_Parm(idx, "kex"  ,     1, 12000, "00" );
	Define_Parm(idx, "papb" , 0.001, 0.249, "00" );
	Define_Parm(idx, "R20"  ,   0.1,   100, "01" );
	Define_Parm(idx, "I0"   ,     0,   1e9, "01" );

	return idx;
}
